//==========================================================================

//	This source code is part of the Agent Simulation Test Bed system which

//	was developed for simulating interaction among agents who are attempting 

//	to acquire information from each other and learn about each other. 

//

//	The system was created at North Carolina State Univeristy, 

//	Department of Computer Science between September and December 2001. 

//	It was implemented by Michael J. Suralik based on the specifications 

//	developed by Pinar Yolum and Bin Yu, who were conducting research under 

//   the guidance of Dr. Munindar Singh. 

//==========================================================================



package edu.ncsu.simulators;



import edu.ncsu.simulators.util.ErrorHandler;

import edu.ncsu.simulators.vectors.Answer;



import java.io.File;

import java.io.FileInputStream;

import java.util.Properties;

	

/**

 * Maintains the constants for the system.  Some constants are hard-coded

 * constants that should never need to be changed by users of the system.

 * Other constants are configurable (i.e. parameters) which are loaded from

 * the parameter file which was specified when the program was started.

 */

public class Constants

{

    private static Properties properties = new Properties();

	

    //***********************************************************************		

    //

    //	DECLARE THE "HARD" CONSTANTS & SET THEIR VALUES

    //	THESE CONSTANTS ARE NOT INTENDED TO BE CONFIGURABLE

    //	DO NOT ALTER THEM UNLESS YOU KNOW WHAT YOU ARE DOING

    //

    //***********************************************************************



    public static final String XML_TAG_AGENT = "Agent";		

    public static final String XML_TAG_SERVICE_AGENT = "ServiceAgent";		

    public static final String XML_TAG_AGENTS = "Agents";			

    public static final String XML_TAG_AGENT_MODEL = "AgentModel";	

    public static final String XML_TAG_DEFAULTS = "Defaults";			

    public static final String XML_TAG_DIMENSION = "Dimension";		

    public static final String XML_TAG_EXPERTISE = "Expertise";		

    public static final String XML_TAG_SOCIABILITY = "Sociability";		

    public static final String XML_TAG_GROUP = "Group";		

    public static final String XML_TAG_INTERESTS = "Interests";		

    public static final String XML_TAG_NEIGHBORS = "Neighbors";			

    public static final String XML_TAG_POLICIES = "Policies";	

    public static final String XML_TAG_POLICY_ASKING = "Asking";	

    public static final String XML_TAG_POLICY_ANSWERING = "Answering";	

    public static final String XML_TAG_POLICY_REFERRING = "Referring";	

    public static final String XML_TAG_POLICY_LEARNING = "Learning";	

    public static final String XML_TAG_POLICY_CHANGING_NEIGHBORS = "ChangingNeighbors";		
        
    public static final String XML_TAG_POLICY_CACHING = "Caching";		

    public static final String XML_TAG_POLICY_QUERY_GENERATION = "Query";		

    public static final String XML_TAG_SIMULATION_GENERATION = "SimulationGeneration";			

    public static final String XML_TAG_LATTICE_EXPERTISE = "LatticeExpertise";

    public static final String XML_TAG_WEIGHT_MATRIX = "WeightMatrix";

    public static final String XML_TAG_PROVIDER_LIST = "ProviderList";	

    public static final String XML_TAG_PROVIDER_LIST_NODE = "ProviderListNode";	

    public static final String XML_TAG_ROLE_LIST = "RoleList";	

    public static final String XML_ATTRIBUTE_ID = "id";		

		

    public static final String XML_ATTRIBUTE_AGENTS_PER_GROUP = "agents";		

    public static final String XML_ATTRIBUTE_CLUSTERING = "clustering";		

    public static final String XML_ATTRIBUTE_DIMENSIONS = "dimensions";	

    public static final String XML_ATTRIBUTE_HIGH = "high";		

    public static final String XML_ATTRIBUTE_INDIVIDUALS = "individuals";		

    public static final String XML_ATTRIBUTE_LOW = "low";		

    public static final String XML_ATTRIBUTE_PERCENTAGE_OF_AGENTS = "percentageOfAgents";			

    public static final String XML_ATTRIBUTE_PERCENT_OF_POPULATION = "percentOfPopulation";			

    public static final String XML_ATTRIBUTE_POPULATION_SIZE = "agents";	

    public static final String XML_ATTRIBUTE_SOCIABILITY = "sociability";	

    public static final String XML_ATTRIBUTE_DEFAULT = "-1";	

    public static final String XML_ATTRIBUTE_VALUE = "value";			

	

    // These constants are used to name fields in our messages

    public static final String MESSAGE_FIELD_TYPE = "MFT";	

    public static final String MESSAGE_FIELD_SENDER_ID = "MFSID";				

    public static final String MESSAGE_FIELD_QUERY_ID = "MFQID";		

    public static final String MESSAGE_FIELD_REFERRED_AGENT_ID = "MFRAID";		

    public static final String MESSAGE_FIELD_QUERY = "MFQ";		

    public static final String MESSAGE_FIELD_ANSWER = "MFA";		

    public static final String MESSAGE_FIELD_REFERRAL_CHAIN_ID = "MFRCID";	

	

    // These are the valid values for the MESSAGE_FIELD_TYPE field

    public static final String MESSAGE_ANSWER = "MA";

    public static final String MESSAGE_QUESTION = "MQ";

    public static final String MESSAGE_REFERRAL = "MR";



    public static final Answer NULL_ANSWER = new Answer();

	

    public static final String JNDI_QUEUE_CONNECTION_FACTORY = "QueueConnectionFactory";
    // public static final String JNDI_QUEUE_CONNECTION_FACTORY = "ConnectionFactory";
    

    // These are the possible message listening states of an agent.

    public static final int PRE_SIMULATION = 1;

    public static final int LISTEN_TO_ALL_MESSAGES = 2;

    public static final int IGNORE_MESSAGES_RESPONDING_TO_MY_QUESTION = 3;	

    public static final int IGNORE_ALL_MESSAGES = 4;	

	

    public static final String AGENT_LOG_FILE_NAME_PREFIX = "Agent";

    public static final String AGENT_LOG_FILE_NAME_SUFFIX = ".LOG";

    public static final String SIMULATION_STATE_FILE_NAME_PREFIX = "NeighborSetChanges";

    public static final String SIMULATION_STATE_FILE_NAME_SUFFIX = ".STATE";	



    //***********************************************************************		

    //

    //	DECLARE THE SIMULATION PARAMETERS THAT WILL BE LOADED

    //

    //***********************************************************************

	

    //----------------------------------------------------------------------			

    //	FUNDAMENTAL PARAMETERS

    //----------------------------------------------------------------------


    // Agent type to load

    //public static final String AGENT_TYPE;

    // The number of times that each agent will change its neighbors (or consider it)

    public static final long NUM_NEIGHBOR_SET_CHANGES;

    // The number of questions that each agent will ask between each change of its neighbors (or consideration)

    public static final long NUM_QUESTIONS_PER_NEIGHBOR_SET;

	

    // NOTE: 

    // The total number of questions that each agent will ask during the 

    // simulation = NUM_NEIGHBOR_SET_CHANGES * NUM_QUESTIONS_PER_NEIGHBOR_SET	

	

    // The minimum amount of system time (in milliseconds) that the simulator 

    // will allow to elapse BETWEEN the time when the last message was sent or 

    // received by an agent, AND the time when the simulator will stop agents

    // from waiting for any more messages and then move on to the next question.

    public static final long WAITING_LIMIT_FOR_MESSAGES;

	

    // The maximum number of referrals that an agent will allow in a referral chain.

    // Even though the first agent in a chain is asked directly 

    // (and is not referred by another agent), we still count that agent as

    // the first agent in the chain.

    public static final int MAXIMUM_REFERRAL_CHAIN_LENGTH;



    //----------------------------------------------------------------------			

    //	NAMES OF THE POLICY IMPLENTATION CLASSES

    //----------------------------------------------------------------------

	

    public static final String ASKING_POLICY_IMPLEMENTATION;

    public static final String ANSWERING_POLICY_IMPLEMENTATION;

    public static final String REFERRAL_POLICY_IMPLEMENTATION;

    public static final String LEARNING_POLICY_IMPLEMENTATION;

    public static final String CHANGING_NEIGHBORS_POLICY_IMPLEMENTATION;

	

    //----------------------------------------------------------------------			

    //	MATHEMATICAL CONSTANTS USED IN EQUATIONS

    //----------------------------------------------------------------------

    public static final double PROMOTION_THRESHOLD;    

    public static final int NUMBER_OF_INTERACTIONS;    

    // Determines the range of perturbation of the questioning agent's interest vector in order to generate a query vector 

    public static final double MAXIMUM_PERCENTAGE_PERTURBATION_FOR_GENERATING_QUERY;			

    // Determines the range of perturbation of the answering agent's expertise vector in order to generate an answer vector

    public static final double MAXIMUM_PERCENTAGE_PERTURBATION_FOR_GENERATING_ANSWER;



    // When an agent asks a question, the agent receiving the question 

    // typically modifies its model of the asking agent's interests.

    // The new interest is computed as a weighted average of the old 

    // interest model and the query that was just received.

    // This constant is the weighting factor that determines how much of the 

    // new estimate is based on the orginal estimate, and how much on the query.

    // The range of this constants is [0,1].

    // Examples:

    // 0.0 corresponds to a constant model that never changes.

    // 0.5 corresponds to an average of the two vectors.

    // 1.0 corresponds to using the most recent query as a model of the interests.

    public static final double RATE_OF_LEARNING_INTERESTS;

	

    // Determines how much an agent's expertise is increased from receiving

    // a correct answer.

    // The range of this constants is [0,1].

    // However, small values are recommended.

    public static final double RATE_OF_LEARNING_FROM_GOOD_ANSWERS;

	

    // Alpha is a constant in the equation for updating expertise

    public static final double ALPHA;

    // This constant determines how much the questioning agent will decay the 

    // model of expertise for an agent who never responded to a query.

    public static final double PERCENTAGE_DECAY_FOR_NO_ANSWER;

	

    // Beta is a constant in the equation for updating sociability

    public static final double BETA;

	

    // W is a constant in the equation for calculating relevance

    public static final double W;

	

    // The minimum similarity required between agent Q's query and agent Q's 

    // model of agent N's expertise in order for agent Q to decide to send the 

    // question to agent N.

    public static final double SIMILARITY_THRESHOLD_FOR_ASKING_NEIGHBOR;

    // The minimum similarity required between agent N's expertise and agent Q's 

    // question in order for agent N to decide to send an answer back to agent Q.

    public static final double SIMILARITY_THRESHOLD_FOR_ANSWERING_QUESTIONER;

    // The minimum similarity required between agent Q's query and agemt N's 

    // model of agent R's expertise  in order for agent N to decide to refer 

    // agent Q to agent R.

    public static final double SIMILARITY_THRESHOLD_FOR_REFERING_QUESTIONER_TO_NEIGHBOR;

    // The minimum similarity required between agent Q's query and agent A's 

    // answer, in order for that answer to be graded as a GOOD_ANSWER by 

    // the system.

    public static final double SIMILARITY_THRESHOLD_FOR_EVALUATING_ANSWER;

	

    // Observation:

    //

    // Usually you will want the following relationship to hold,

    //

    //	SIMILARITY_THRESHOLD_FOR_EVALUATING_ANSWER

    //		> SIMILARITY_THRESHOLD_FOR_ANSWERING_QUESTIONER

    //			> SIMILARITY_THRESHOLD_FOR_REFERING_QUESTIONER_TO_NEIGHBOR

    //				> SIMILARITY_THRESHOLD_FOR_ASKING_NEIGHBOR	

	

    //----------------------------------------------------------------------			

    //	INPUT / OUTPUT

    //----------------------------------------------------------------------



    // Recognizing that it is difficult to read the *.LOG and *.STATE files

    // when decimal numbers are written with full precision, 

    // we have created this constant which allows you to include only the 

    // first few leading characters in the decimal strings.

    // Keep in mind that if you repeatedly use the *.STATE files as input,

    // you will cause increasingly larger losses of precision.

    public static final int DECIMAL_STRING_LENGTH;

	

    // Determines whether or not the simulator writes the extra information

    // to the *.STATE files that enables them to be read in as agent input files. 

    public static final boolean OPTION_WRITE_READABLE_STATE_FILES;



    // Determines whether or not a log file is maintained for the decisions

    // and events related to each agent.

    public static final boolean OPTION_WRITE_AGENT_LOGS;	



    // Determines whether or not a log file is maintained for the decisions

    // and events related to each agent.

    public static final boolean OPTION_WRITE_AGENT_STATISTICS;	

	

    //***********************************************************************		

    //

    //	LOAD THE SIMULATION PARAMETERS

    //

    //***********************************************************************



    static

    {	

	String fileName = null;

		

	try

	    {

		//File fileProperties = new File( "H:/JBoss-2.4.1_Tomcat-3.2.3/jboss/ncsu/resources/simulation.conf" );

		fileName = SimulationContext.getParameterFileName();

		File fileProperties = new File( fileName );

		FileInputStream fis = new FileInputStream( fileProperties );

		properties.load( fis );

		System.out.println( "properties = " + properties );

	    }

	catch ( Exception e )

	    { 

		ErrorHandler.exit( e, "Could not initialize constants using file " + fileName );

	    }

		

	//----------------------------------------------------------------------			

	//	FUNDAMENTAL PARAMETERS

	//----------------------------------------------------------------------


        //AGENT_TYPE = properties.getProperty("AGENT_TYPE");

	NUM_NEIGHBOR_SET_CHANGES = Integer.parseInt(properties.getProperty("NUM_NEIGHBOR_SET_CHANGES"));

	NUM_QUESTIONS_PER_NEIGHBOR_SET = Integer.parseInt(properties.getProperty("NUM_QUESTIONS_PER_NEIGHBOR_SET"));

	WAITING_LIMIT_FOR_MESSAGES = Integer.parseInt(properties.getProperty("WAITING_LIMIT_FOR_MESSAGES"));

	MAXIMUM_REFERRAL_CHAIN_LENGTH = Integer.parseInt(properties.getProperty("MAXIMUM_REFERRAL_CHAIN_LENGTH"));

	PROMOTION_THRESHOLD = Double.parseDouble(properties.getProperty("PROMOTION_THRESHOLD", "0.25"));

	NUMBER_OF_INTERACTIONS = Integer.parseInt(properties.getProperty("NUMBER_OF_INTERACTIONS", "0"));

	//----------------------------------------------------------------------			

	//	NAMES OF THE POLICY IMPLENTATION CLASSES

	//----------------------------------------------------------------------



	ASKING_POLICY_IMPLEMENTATION = properties.getProperty("ASKING_POLICY_IMPLEMENTATION");

	ANSWERING_POLICY_IMPLEMENTATION = properties.getProperty("ANSWERING_POLICY_IMPLEMENTATION");;

	REFERRAL_POLICY_IMPLEMENTATION = properties.getProperty("REFERRAL_POLICY_IMPLEMENTATION");;

	LEARNING_POLICY_IMPLEMENTATION = properties.getProperty("LEARNING_POLICY_IMPLEMENTATION");;

	CHANGING_NEIGHBORS_POLICY_IMPLEMENTATION = properties.getProperty("CHANGING_NEIGHBORS_POLICY_IMPLEMENTATION");;

	

	//----------------------------------------------------------------------			

	//	MATHEMATICAL CONSTANTS USED IN EQUATIONS

	//----------------------------------------------------------------------



	MAXIMUM_PERCENTAGE_PERTURBATION_FOR_GENERATING_QUERY = Double.parseDouble(properties.getProperty("MAXIMUM_PERCENTAGE_PERTURBATION_FOR_GENERATING_QUERY"));		

	MAXIMUM_PERCENTAGE_PERTURBATION_FOR_GENERATING_ANSWER = Double.parseDouble(properties.getProperty("MAXIMUM_PERCENTAGE_PERTURBATION_FOR_GENERATING_ANSWER"));

	RATE_OF_LEARNING_INTERESTS = Double.parseDouble(properties.getProperty("RATE_OF_LEARNING_INTERESTS"));

	RATE_OF_LEARNING_FROM_GOOD_ANSWERS = Double.parseDouble(properties.getProperty("RATE_OF_LEARNING_FROM_GOOD_ANSWERS"));

	ALPHA = Double.parseDouble(properties.getProperty("ALPHA"));

	PERCENTAGE_DECAY_FOR_NO_ANSWER = Double.parseDouble(properties.getProperty("PERCENTAGE_DECAY_FOR_NO_ANSWER"));

	BETA = Double.parseDouble(properties.getProperty("BETA"));

	W = Double.parseDouble(properties.getProperty("W"));

	SIMILARITY_THRESHOLD_FOR_ASKING_NEIGHBOR = Double.parseDouble(properties.getProperty("SIMILARITY_THRESHOLD_FOR_ASKING_NEIGHBOR"));

	SIMILARITY_THRESHOLD_FOR_ANSWERING_QUESTIONER = Double.parseDouble(properties.getProperty("SIMILARITY_THRESHOLD_FOR_ANSWERING_QUESTIONER"));

	SIMILARITY_THRESHOLD_FOR_REFERING_QUESTIONER_TO_NEIGHBOR = Double.parseDouble(properties.getProperty("SIMILARITY_THRESHOLD_FOR_REFERING_QUESTIONER_TO_NEIGHBOR"));

	SIMILARITY_THRESHOLD_FOR_EVALUATING_ANSWER = Double.parseDouble(properties.getProperty("SIMILARITY_THRESHOLD_FOR_EVALUATING_ANSWER"));



	//----------------------------------------------------------------------			

	//	INPUT / OUTPUT

	//----------------------------------------------------------------------	

		

	DECIMAL_STRING_LENGTH = Integer.parseInt(properties.getProperty("DECIMAL_STRING_LENGTH"));

	OPTION_WRITE_READABLE_STATE_FILES = getBoolean( "OPTION_WRITE_READABLE_STATE_FILES" );

	OPTION_WRITE_AGENT_LOGS = getBoolean( "OPTION_WRITE_AGENT_LOGS" );		

	OPTION_WRITE_AGENT_STATISTICS = getBoolean( "OPTION_WRITE_AGENT_STATISTICS" );		

    }

	

    //=======================================================================	

	

    private static boolean getBoolean( String propertyName )

    {

	boolean booleanValue = false;

		

	String stringBoolean = properties.getProperty( propertyName );

		

	if ( stringBoolean.equalsIgnoreCase("TRUE") )

	    {

		booleanValue = true;

	    }

		

	return booleanValue;

    }

	

    //=======================================================================

}
