//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//============================================================================

package edu.ncsu.simulators;

import edu.ncsu.simulators.io.LogWriter;
import edu.ncsu.simulators.util.ErrorHandler;

import java.io.File;
import java.io.FileWriter;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.naming.Context;
import javax.naming.InitialContext;

/**
 * Maintains the global variables for the system.
 */
public class SimulationContext 
{
	// the JNDI initial context
	private static Context jndiContext = null;	
	// the root directory for the simulator
	private static File dirRoot = null;
	// the directory that contains all of the log directories
	private static File dirLogParent = null;
	// the directory that contains all of the input files
	private static File dirResources = null;
	// the log directory that contains all of the log files for the current
	// execution of the simulator
	private static File dirLog = null;
	// the FileWriter for the main log file for the current execution of the
	// simulator
	private static LogWriter logWriter = null;
	// the name of the file that contains all of the parameters for the
	// current execution of the simulator
	private static String parameterFileName = null;
	// the QueueConnection that all receivers and senders on this JVM will use
	private static QueueConnection queueConnection = null;	
	
	static
	{
		initialize();
	}
	
	//=======================================================================	
	
	private SimulationContext() 
	{	
	}
	
	//=======================================================================
	/**
	 * Initializes the global variables.
	 * Should only be called once.
	 */
	private static void initialize()
	{
		try
		{
			// set the Root Directory using the "root" environment variable
			dirRoot = new File ( System.getProperty("root") );		

			// set the Logging Directory
			dirLogParent = new File ( dirRoot.getAbsolutePath() + File.separator + "logs" );	
			
			// set the Resources Directory
			dirResources = new File ( dirRoot.getAbsolutePath() + File.separator + "resources" );
						
			// Get the JNDI Initial Context
			jndiContext = new InitialContext();
			
			// Get the Queue Connection Factory
			QueueConnectionFactory queueConnectionFactory = (QueueConnectionFactory)jndiContext.lookup( Constants.JNDI_QUEUE_CONNECTION_FACTORY );
			
			// Create the Queue Connection
			queueConnection = queueConnectionFactory.createQueueConnection();
		}
		catch ( Exception e )
		{
			ErrorHandler.exit( e, "Error occured during SimulationContext.initialize" );
		}				
	}
	
	//=======================================================================	

	public static void setLogDir( File directory ) 
	{	
		dirLog = directory;
	}	
	
	//=======================================================================	

	public static void setParameterFileName( String fileName ) 
	{	
		parameterFileName = fileName;
	}
	
	//=======================================================================	

	public static void setSimulatorLog( File fileSimulatorLog ) 
	{	
		try
		{
			// set the FileWriter for the Simulator Log
			logWriter = new LogWriter( fileSimulatorLog );
		}
		catch ( Exception e )
		{
			ErrorHandler.exit( e, "Could not create FileWriter for " + fileSimulatorLog );
		}
	}
	
	//=======================================================================	
	/**
	 * Here are all the public get methods for the global variables
	 */

	public static Context getJNDIContext() { return jndiContext; }			
	public static File getLogDir() { return dirLog; }		
	public static File getLogParentDir() { return dirLogParent; }		
	public static String getParameterFileName() { return parameterFileName; }
	public static QueueConnection getQueueConnection() { return queueConnection; }		
	public static File getResourceDir() { return dirResources; }
	public static LogWriter getSimulatorLog() { return logWriter; }	
	
	//=======================================================================		
}
