//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 

//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//      the guidance of Dr. Munindar Singh. 
//==========================================================================



package edu.ncsu.simulators;



import edu.ncsu.simulators.agent.*;

import edu.ncsu.simulators.io.*;

import edu.ncsu.simulators.util.ErrorHandler;

import edu.ncsu.simulators.util.FileManager;

import edu.ncsu.simulators.util.Statistics;



import java.io.File;

import java.io.FileReader;

import java.io.FileWriter;

import java.io.StreamTokenizer;

import java.util.Calendar;

import java.util.Date;

import java.util.Hashtable;



import javax.naming.Context;

import javax.naming.InitialContext;

import javax.naming.NamingException;



import javax.jms.QueueConnectionFactory;

import javax.jms.QueueConnection;

import javax.jms.QueueSession;

import javax.jms.QueueSender;

import javax.jms.Queue;

import javax.jms.TextMessage;

import javax.jms.Session;

import javax.jms.JMSException;



/**
 * The Simulator class contains the Main method for the system which loads 
 * or generates a set of agents, runs the simulation, and outputs the results.
 * <p>
 * The Simulator begins by reading in a user supplied generation specification
 * file (which broadly specify what kinds of agents should be generated) OR 
 * a user supplied state file (which precisely specify the state of every agent).
 * It also reads in a user supplied parameter file (which specifies the values
 * for a set of parameters that are used by the Simulator and its agents).
 * In addition, it creates, or empties, a user specified directory for logging
 * all of the output.
 * <p>
 * Next, the Simulator saves in the log directory copies of the various input 
 * files along with the initial state of the agents before the simulation begins.
 * <p>
 * Now the simulation is ready to begin looping.  Each loop starts with the 
 * Simulator triggering each of the agents to generate their next question. 
 * The agents will then go off and do their thing, which is to ask those 
 * questions of some of their neighbors - which will lead to referrals and/or 
 * answers, and possibly questions of the referred agent. The details of this
 * are handled by the Agent class and the implementations of the various 
 * policies that each agent uses. In any case, the Simulator will wait until 
 * the system has settled down to the point where agents have not sent or 
 * received messages from each other for a period of time (as specified by the 
 * parameter file).
 * <p>
 * Once the system has settled down, the Simulator will trigger each agent
 * to learn based on the answers that it did or did not get to the question
 * that it asked.  After a certain number of loops (as specified by the 
 * parameter file), the Simulator will trigger the agents to revaluate their
 * neighbors and possibly select new neighors. The details of this are handled
 * by the Agents and their implementations of the neighbor selection policy.
 * <p>
 * @author Mike Suralik
 */

public class Simulator 
{
    public static final String FILE_SUFFIX_GENERATION_SPECIFICATION = ".GEN";
    public static final String FILE_SUFFIX_AGENT_STATES = ".STATE";	
    public static final String FILE_SUFFIX_PARAMETERS = ".CONF";
    public static boolean FLG_IGNORE_MESSAGES = true;

    private static Agent[] agents = null;
    private static int numAgents = 0;
    private static int numDimensions = 0;
    private static AgentModel defaultAgentModel = null;
    private static int stateNumber = 0;


    //=======================================================================
    /**
     * Loads or generates a set of agents, runs the simulation, and outputs the 
     * results.  See the class documentation for more details.
     *
     * @param arguments an array of string arguments accepted from the command line, where
     * <ul>
     * <li> arg1 = the root directory of all the simulator related subdirectories
     * <li> arg2 = a *.GEN file which indicates how to generate agents OR </br>
     * a *.STATE file which contains the state of existing agents
     * <li> arg3 = a *.CONF file which contains simulation parameters
     * <li> arg4 = the subdirectory of (root-dir)\\logs to be used for logging data
     * <li> arg5 (OPTIONAL) = the number used to seed the random number generator
     */
    public static void main( String[] args ) {

	try {	

	//--------------------------------------------------------------
	//	VALIDATE COMMAND LINE PARAMETERS
	//--------------------------------------------------------------	

	    String agentsFileName = null;
	    String parametersFileName = null;				
	    String logSubdirectoryName = null;		
	    Long randomNumberGeneratorSeed = null;

	    // if the number of parameters is unexpected, then abort

	    if ( (args.length < 3) || (args.length > 4) ) {

		printUsage();
		ErrorHandler.exit( "Incorrect number of arguments" );
	    }

	    // the 3 required arguments
	    
	    agentsFileName = args[0];
	    parametersFileName = args[1];				
	    logSubdirectoryName = args[2];		



	    // the 4th argument is optional

	    if ( args.length == 4 ) {

		String strSeed = args[3]; 
		try {
		    randomNumberGeneratorSeed = new Long( Long.parseLong(strSeed) );		
	
		}

		catch ( Exception e) {

		    printUsage();
		    ErrorHandler.exit( e, "Invalid seed for random number generator" );

		}
	    }


	    // if any of the parameters have invalid values, then abort

	    if ( ((!agentsFileName.toUpperCase().endsWith(FILE_SUFFIX_GENERATION_SPECIFICATION)) &&

		  (!agentsFileName.toUpperCase().endsWith(FILE_SUFFIX_AGENT_STATES))) || 

		 (!parametersFileName.toUpperCase().endsWith(FILE_SUFFIX_PARAMETERS)) )

		{

		    printUsage();

		    ErrorHandler.exit( "Incorrect file extension" );

		}



	    if ( randomNumberGeneratorSeed != null ) {

		// if the user supplied a valid seed, then set it
		Statistics.initializeSeed( randomNumberGeneratorSeed );

	    }

	    else {

		// if the user did not supply a seed at all,
		// find out what the seed was so that we can record
		// it to the log later.
		randomNumberGeneratorSeed = Statistics.getSeed();
	    }



	    // Convert the file and directory names, into absolute paths

	    String agentsFullFileName = SimulationContext.getResourceDir() + File.separator + agentsFileName;				

	    String parametersFullFileName = SimulationContext.getResourceDir() + File.separator + parametersFileName;

	    String logFullSubdirectoryName = SimulationContext.getLogParentDir() + File.separator + logSubdirectoryName;		

			

	    if ( ! (new File(agentsFullFileName).exists())) { 

		printUsage();
		ErrorHandler.exit( agentsFullFileName + " does not exist" );
	    
	    }

	    if ( ! (new File(parametersFullFileName).exists())) {

		printUsage();				
		ErrorHandler.exit( parametersFullFileName + " does not exist" );				
	    }

			

	    // Initialize the parameter file name.

	    // This will be needed in order to load the constants.

	    SimulationContext.setParameterFileName( parametersFullFileName );			

			
	    System.out.println();			
	    System.out.println();			
	    System.out.println();
	    System.out.println( "===================================================================" );
	    System.out.println();
	    System.out.println( "                           SIMULATOR");
	    System.out.println();
	    System.out.println( "===================================================================" );
	    System.out.println();
			

		//--------------------------------------------------------------

		//	PREPARE THE LOGGING SUBDIRECTORY

		//--------------------------------------------------------------	

			

	    File dirLogSubdirectory = new File( logFullSubdirectoryName );

			

	    // If the directory already exists, then either clean it out

		// or allow the user to abort

	    if ( (dirLogSubdirectory.exists()) && (dirLogSubdirectory.listFiles().length>0)) {

		System.out.println( dirLogSubdirectory.getAbsolutePath() + " is not empty!" );	  
		System.out.println();				

		int optionDelete = -1;

		while ( (optionDelete!=89) && (optionDelete!=121) && (optionDelete!=78) && (optionDelete!=110) ) {
		    if ( (optionDelete!=13) && (optionDelete!=10)) {

			    System.out.print( "Delete all existing files in that directory {Y/N}? " );				
		    }

		    optionDelete = System.in.read();					

		}

		// If option = yes, then delete the contents of the specified log directory

		if ( (optionDelete==89) || (optionDelete==121)) {

		    System.out.println( "Deleting files..." );					

		    File[] files = dirLogSubdirectory.listFiles();

		    for ( int i=0; i<files.length; i++) {

			files[i].delete();
		    }
		}

		// If option = no, then delete the contents of the specified log directory				
		else {
		    System.out.println( "Can not start simulator. An empty directory must be specified for logging." );					
		    System.exit( 1 );
		}		

	    }

	    // If the directory does not exist, then create the directory

	    else {

		System.out.println( "Creating log directory: " + dirLogSubdirectory.getAbsolutePath() );							
		dirLogSubdirectory.mkdirs();
	    }


	    File fileSimulatorLog = new File( dirLogSubdirectory.getAbsolutePath() + File.separator + "Simulator.LOG" );			

	    System.out.println( "Logging simulator events to: " + fileSimulatorLog.getAbsolutePath() );									

	
	    // set the Logging directory that will be used
	    // set the primary Log file that the simulator will use

	    SimulationContext.setLogDir( dirLogSubdirectory );

	    SimulationContext.setSimulatorLog( fileSimulatorLog );			

	    //***************************************************
	    //	THE LOG HAS BEEN INITIALIZED
	    //	IT IS SAFE TO USE THE LOG AFTER THIS POINT
	    //***************************************************			

	    log( "The random number generator has been seeded with: " + randomNumberGeneratorSeed );
		
	    //--------------------------------------------------------------
	    //	FROM THE RESOURCE DIRECTORY,
	    //	COPY THE PARAMETERS FILE AND THE AGENTS FILE 
	    //   TO THE LOG DIRECTORY IN ORDER TO HAVE A RECORD OF THE 
	    //	INITIAL CONDITIONS OF THE SIMULATOR
	    //--------------------------------------------------------------	

	    FileManager.copy( agentsFullFileName, SimulationContext.getLogDir() + File.separator + agentsFileName );

	    FileManager.copy( parametersFullFileName, SimulationContext.getLogDir() + File.separator + parametersFileName );


	    //--------------------------------------------------------------
	    //	POPULATE THE AGENTS ARRAY USING THE SUPPLIED (agents-file)
	    //	WHICH WILL BE EITHER A GENERATATION SPECIFICATION FILE
	    //   OR A STATE FILE.
	    //--------------------------------------------------------------	

	    SimulationGenerator generator = null;

	    if ( agentsFullFileName.toUpperCase().indexOf(".GEN") != -1 ) {

		log( "Generating the simulation from the generation specification file " + agentsFullFileName );				
		generator = new SpecificationReader( agentsFullFileName );		
	    }

	    else if ( agentsFullFileName.toUpperCase().indexOf(".STATE") != -1 ) {

		log( "Generating the simulation from the state file: " + agentsFullFileName );			
		generator = new StateReader( agentsFullFileName );			
	    }

	    else {

		System.out.println( "The name of the input file must end with \".GEN\" or \".STATE\" to indicate whether it is a generation specification file or a state file!" );
		System.exit( 1 );
	    }



	    agents = generator.getAgents();

	    numAgents = agents.length;

	    numDimensions = generator.getDimensions();

	    defaultAgentModel = generator.getDefaultAgentModel();

	    // give the error handler a pointer to the agents so that it
	    // can release their resources if something bad happens

	    ErrorHandler.setAgents( agents );
	    //--------------------------------------------------------------
	    //	SAVE THE INITIAL STATE TO A FILE
	    //--------------------------------------------------------------
	    System.out.println("Number of int. is " + Constants.NUMBER_OF_INTERACTIONS);
	    log( "Saving the initial state of the agents to a file..." );		
	    saveAgents( stateNumber++ );				

	    //--------------------------------------------------------------			
	    //   START THE QUEUE CONNECTION (i.e. message delivery)
	    //	AND ALLOW TIME FOR AGENTS TO RECEIVE AND IGNORE ANY
	    //	UNDELIVERED MESSAGES FROM PREVIOUS RUNS OF THE SIMULATOR
	    //
	    //   WARNING: IT APPEARS THAT IF THERE ARE UNDELIVERED MESSAGES
	    //            FROM PREVIOUS RUNS, THEY WILL SOMETIMES STAY IN
	    //            THE PERSISTENCE STORE AND REMAIN UNDELIVERED
	    //            EVEN AFTER WE ALLOW EXTRA TIME FOR THEIR DELIVERY.
	    //            THEREFORE, IT WOULD BE BEST TO DELETE ALL OF THE
	    //            MESSAGE QUEUES PRIOR TO STARTING THE PROGRAM -
	    //            TO BE ON THE SAFE SIDE.
	    //--------------------------------------------------------------			

	    SimulationContext.getQueueConnection().start();

	    boolean flgStillReceivingRogueMessages = true;

	    while ( flgStillReceivingRogueMessages) {

		flgStillReceivingRogueMessages = false;
		log( "Waiting so that JMS can deliver any undelivered messages from previous simulation runs..." );
		log( "Messages will be ignored by agents" );					      
		Thread.currentThread().sleep( 5000 );

		for( int indexAgents=0; indexAgents < numAgents; indexAgents++) {

		    if ( agents[indexAgents].getMessagesReceived() > 0) {

			flgStillReceivingRogueMessages = true;				

			for( int i=0; i < numAgents; i++) {
			    agents[i].resetMessageCounters();
			}	
			break;
		    }
		}			
	    }


	    //--------------------------------------------------------------			
	    //	RESET THE MESSAGE COUNTERS 
	    //--------------------------------------------------------------	

	    for( int indexAgents=0; indexAgents < numAgents; indexAgents++) {

		agents[indexAgents].resetMessageCounters();
	    }

	    //--------------------------------------------------------------			
	    //	NOTIFY AGENTS TO PROCESS ALL SUBSEQUENT MESSAGES
	    //--------------------------------------------------------------	

	    log( "...Finished waiting for undelivered messages." );			
	    log( "Notifying agents that they can process future messages..." );

	    for( int indexAgents=0; indexAgents < numAgents; indexAgents++){
		    
		agents[indexAgents].setMessageListeningState(Constants.LISTEN_TO_ALL_MESSAGES);
	    }
	    //--------------------------------------------------------------
	    //	RUN THE SIMULATION
	    //--------------------------------------------------------------		

	    log();
	    log( "BEGINNING THE SIMULATION RUN..." );		
	    log();

	    for ( int indexNeighborSetChanges=0; indexNeighborSetChanges < Constants.NUM_NEIGHBOR_SET_CHANGES; indexNeighborSetChanges++) {

		for ( int indexQuestionsForThisNeighborSet=0; indexQuestionsForThisNeighborSet < Constants.NUM_QUESTIONS_PER_NEIGHBOR_SET; indexQuestionsForThisNeighborSet++) {

		    // Run the garbage collector before each set of questions
		    //System.gc();
		    log( "-----------------------------------------------------------------------------" );
		    log( "Allowing agents to ask question #" + (indexQuestionsForThisNeighborSet+1) + " after neighbor set change #" + indexNeighborSetChanges );	
		    log( "-----------------------------------------------------------------------------" );
		    log( "Notify agents to process any future messages..." );							
		    for( int indexAgents=0; indexAgents < numAgents; indexAgents++) {
			    agents[indexAgents].startAcceptingResponses();									
		    }
		    log( "Make agents ask their next question..." );						
		    for( int indexAgents=0; indexAgents < numAgents; indexAgents++) {
			    agents[indexAgents].askQuestion();
		    }

		    boolean flgAgentsStillBusy = true;

		    while (flgAgentsStillBusy) {

			log( "Waiting for agents to stop receiving messages..." );							
			Thread.currentThread().sleep( Constants.WAITING_LIMIT_FOR_MESSAGES );
			flgAgentsStillBusy = false;

			long currentTime = System.currentTimeMillis();
			for( int indexAgents=0; indexAgents < numAgents; indexAgents++) {

			    Long timeOfEvent = agents[indexAgents].getTimeOfMostRecentMessageEvent();

			    // if timeSent = null, that means the agent
			    // has not yet sent any messages at all
			    if ( (timeOfEvent != null) && ((currentTime - timeOfEvent.longValue()) < Constants.WAITING_LIMIT_FOR_MESSAGES)) {
				    flgAgentsStillBusy = true;
				    break;
			    }
			}						
		    }											

		    log( "Agents have not sent a message in a while" );
		    log( "Stopping agents from processing any future messages..." );		      

		    for( int indexAgents=0; indexAgents < numAgents; indexAgents++) {

			    agents[indexAgents].stopAcceptingResponses();				
		    }	

		    log( "Allowing agents to learn from responses to their last question..." );		

		    for( int indexAgents=0; indexAgents < numAgents; indexAgents++) {

			agents[indexAgents].learnFromResponses();
			//after learning is over, the best answer is known so it can now be cached
			//agents[indexAgents].cacheBestResponse();
		    }
                                        
		}

		log( "Allowing agents to change their neighbors..." );				

		for( int indexAgents=0; indexAgents < numAgents; indexAgents++) {

		    agents[indexAgents].changeNeighbors();					      
		}	

		log( "Saving the state of the agents to a file..." );		
		saveAgents( stateNumber++ );					

	    }						

	    log();
	    log( "...ENDING THE SIMULATION RUN" );							
	    log();

	    //--------------------------------------------------------------			
	    //	RELEASE ALL RESOURCES
	    //--------------------------------------------------------------	
	    log( "Releasing all resources..." );

	    // for each agent
	    for ( int id=0; id < numAgents; id++) {

		agents[id].close();
	    }					
	    SimulationContext.getQueueConnection().close();

	    //--------------------------------------------------------------			
	    //	PRINT DIAGNOSTICS
	    //--------------------------------------------------------------	

	    long totalSent = 0;
	    long totalReceived = 0;

	    log( "-----------------------------------------------------------------------------" );
	    log( "Messaging Statistics:" );
	    log( "-----------------------------------------------------------------------------" );

	    // print the number of messages sent and received by the agents
	    for ( int id=0; id < numAgents; id++) {

		log( "Agent(" + id + ") sent " + agents[id].getMessagesSent() );
		totalSent = totalSent + agents[id].getMessagesSent();
		log( "Agent(" + id + ") received " + agents[id].getMessagesReceived() );
		totalReceived = totalReceived + agents[id].getMessagesReceived(); 
	    }	

	    log( "Total messages sent = " + totalSent );	
	    log( "Total messages received = " + totalReceived );

	    // print the number of messages received by the agents
	    if (agents[0].getXMLTag() == Constants.XML_TAG_SERVICE_AGENT) {
		
		FileWriter outputFileWriter = new FileWriter(dirLogSubdirectory.getAbsolutePath() + File.separator + "OwnServices.LOG", true);			
		for ( int id=0; id < numAgents; id++) {
		     outputFileWriter.write(String.valueOf(((ServiceAgent)agents[id]).getOwnGoodProviders()));
		     outputFileWriter.write("\t");		
		     outputFileWriter.write(String.valueOf(((ServiceAgent)agents[id]).getOwnServiceProviders()));
		     outputFileWriter.write("\t");		
		     outputFileWriter.write(String.valueOf(((ServiceAgent)agents[id]).getNumberOfUsefulPromotedProviders()));
		     outputFileWriter.write("\t");		
		     outputFileWriter.write(String.valueOf(((ServiceAgent)agents[id]).getNumberOfPromotedProviders()));

		     outputFileWriter.write("\n");		
		}	
		outputFileWriter.close();			
		
	    } // end of if ()
	    


	}

	catch( Exception e ) {
	    ErrorHandler.exit( e, "Untrapped error caught by Simulator.main" );
	}

	finally {

	    try {
		SimulationContext.getSimulatorLog().close();
	    }
	    catch ( Exception e ) {
		e.printStackTrace();
	    }

	}

    }



    //=======================================================================	

	

    public static int getAgentCount() { return numAgents; }

    public static AgentModel getDefaultAgentModel() { return defaultAgentModel; }

    public static int getDimensionCount() { return numDimensions; }	

	

    //=======================================================================

    public static void log() {
	log( "" );
    }	

    //=======================================================================

    public static void log( String text ) {

	SimulationContext.getSimulatorLog().log( text );
	System.out.println( text );
    }	
    //=======================================================================

    private static void printUsage() {

	System.out.println();					
	System.out.println( "------------------------------------------------------------" );				
	System.out.println();				
	System.out.println( "USAGE: java -Droot=(root-dir) simulator (agents-file) (config-file) (log-dir) [seed]" );
	System.out.println();		
	System.out.println( "   root-dir   :  the root directory of all the simulator related subdirectories" );				
	System.out.println( "   agents-file:  a *.GEN file which indicates how to generate agents OR" );
	System.out.println( "                 a *.STATE file which contains the state of existing agents" );
	System.out.println( "   config-file:  a *.CONF file which contains simulation parameters" );				
	System.out.println( "   log-dir    :  the subdirectory of (root-dir)\\logs to be used for logging data" );	
	System.out.println( "   seed		:  OPTIONAL, the number used to seed the random number generator" );			
	System.out.println( "" );
	System.out.println( "   Both the (agents-file) and the (config-file) must be located in" );
	System.out.println(	"   the (root-dir)\\resources directory" );			
	System.out.println();				
	System.out.println( "------------------------------------------------------------" );								
	System.out.println();

    }

    //=======================================================================

    private static void saveAgents( int fileNameBase ) {
	try {
	    StateWriter outputFileWriter = new StateWriter( SimulationContext.getLogDir() + File.separator + Constants.SIMULATION_STATE_FILE_NAME_PREFIX + "_" + fileNameBase + Constants.SIMULATION_STATE_FILE_NAME_SUFFIX );
	    outputFileWriter.save( agents, numDimensions, defaultAgentModel );
	}
	catch( Exception e ) {
	    e.printStackTrace();
	}

    }	
    //=======================================================================	

}
