//==========================================================================

//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 

//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//      the guidance of Dr. Munindar Singh. 
//==========================================================================



package edu.ncsu.simulators.agent;
 


import edu.ncsu.simulators.*;
import edu.ncsu.simulators.dialogs.*;
import edu.ncsu.simulators.io.LogWriter;
import edu.ncsu.simulators.policies.*;
import edu.ncsu.simulators.util.*;
import edu.ncsu.simulators.vectors.*;

import net.n3.nanoxml.*;

import java.io.File;
import java.io.FileWriter;
import java.io.StreamTokenizer;
import java.io.StringReader;

import java.util.Calendar;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;
import java.util.Properties;
import java.util.Vector;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import javax.jms.QueueConnectionFactory;
import javax.jms.QueueConnection;
import javax.jms.QueueReceiver;
import javax.jms.QueueSession;
import javax.jms.QueueSender;
import javax.jms.Queue;
import javax.jms.Message;
import javax.jms.MapMessage;
import javax.jms.MessageListener;
import javax.jms.Session;
import javax.jms.JMSException;
	

/**

 * Agents interact with each other throughout the course of a simulation run
 * in an effort to get their questions answered, to learn about other agents,
 * and to develop a set of agent neighbors that will help them find the answers
 * to their questions.
 * <p>
 * Here are the intrinsic properties that make up every agent instance:
 * <ul>
 * <li> Expertise
 * <li> Interests
 * <li> Set of neighboring agents
 * <li> Set of models of other agents (which may or may not be neighbors)
 * <li> Policy for asking questions
 * <li> Policy for answering questions
 * <li> Policy for referring agents to other agents
 * <li> Policy for learning about other agents
 * <li> Policy for changing neighbors
 * </ul>
 * <p>
 * Users of the system can test the effect of these properties by manipulating
 * the generation specification file or the agent state file which spell out
 * the values of these for each agent.
 * <p>
 * Although each agent controls the details of its own behavior, it is the 
 * Simulator class that triggers the start of certain behaviors.  Listed below 
 * are the 3 methods that the Simulator uses to trigger the activity of the agent:
 * <p>
 * <ul>
 * <li> askQuestion() -  causes the Agent to generate a new query and use its 
 * AskingPolicy to decide who to send the question to.
 * <li> learnFromResponses() - causes the Agent to use its LearningPolicy to
 * learn about the other agents in the system that it interacted with since
 * asking the last question.
 * <li> changeNeighbors() - causes the Agent to use its ChangeNeighborsPolicy 
 * to re-evaluate the quality of its neighbors and the other agents that it
 * has models of, possibly leading it to change its neighbors.
 * </ul>
 * <p>
 * Agents communicate with each other via the asynchronous messaging capability
 * provided by JBoss Message Queue.  To send a message to a particular agent, 
 * an agent sends a message to that agent's queue.  When an agent is created,
 * it registers with JBoss Message Queue to listen to its queue.  When a
 * message arrives in its queue, JBoss Message Queue automatically invokes the 
 * onMessage method of the Agent object and passes the waiting message as a parameter.
 * The Agent class implements the MessageListener interface which enables them
 * to listen to the JBoss Message Queue for any messages that have been sent to
 * them.
 */

public class Agent implements MessageListener {	
	

    /** ------------ THE AGENT's FUNDAMENTAL ATTRIBUTES ---------- **/
    public int id = -1;

    public Expertise expertise = null;
    public Interests interests = null;
    public Neighbors neighbors = null;

	

    /** ------------ THE AGENT's DECISION MAKING POLICIES -------- **/

    // policy for deciding which agents we should send a particular query to
    public AskingPolicy askingPolicy = null;

    // policy for deciding which queries to answer, and how to answer them
    public AnsweringPolicy answeringPolicy = null;

    // policy for deciding which agents we should refer a questioning agent to
    public ReferringPolicy referringPolicy = null;

    // policy for learning new expertise for those agents who answered questions
    // and for learning new sociability for those agents who referred us to the answering agents
    public LearningPolicy learningPolicy = null;

    // policy for adding/removing agents to the set of neighbors
    public ChangingNeighborsPolicy changingNeighborsPolicy = null;
        
    // policy for handling caches
    public CachingPolicy cachingPolicy = null;

    // policy for generating queries
    public QueryGenerationPolicy queryGenerationPolicy = null;

    /** ------------ THE AGENT's MEMORY --------------------------- **/

    // models of all the agents that this agent knows anything about
    public AgentModels agentModels = null;

    // the query that the agent is currently trying to find the answer to
    public Query currentQuery = null;

    // the number that identifies the current query
    public long currentQueryID = -1;

    // a collection of id's identifying the agents to whom the current query has been sent
    public HashSet setAgentsWhomHaveBeenSentCurrentQuery = null;

    public ReferralChain referralChain = null;
    public UnprocessedReferrals unprocessedReferrals = null;
        
    //The agent's cache
    public Cache cache = null;

    /** ------------ COUNTERS & STATISTICS ------------------------ **/

    // a counter of the number of queries (not question messages) that this agent has generated
    public long counterQueriesGenerated = 0;

    // a counter of the number of referral chains that this agent has started
    // used as an ID generator
    public long counterReferralChains = 0;

    // a counter of the number of messages sent by this agent
    // maintained for statistics
    public int counterMessagesReceived = 0;

    // a counter of the number of messages received by this agent
    // maintained for statistics
    public int counterMessagesSent = 0;

    /** ------------ COMMUNICATION INFRASTRUCTURE ----------------- **/

    /** Queue session, hold on to this so you may close it. Also used to create messages. */
    public QueueSession queueSession = null;

    /** Use this to send messages. */ 
    public QueueSender queueSender = null;

    /** Use this to receive messages. */
    public QueueReceiver queueReceiver = null;


    /** ------------ OTHER INFRASTRUCTURE ------------------------- **/


    // the random number generator used by this agent
    public Random random = null;

    public int stateMessageListening = Constants.IGNORE_ALL_MESSAGES;
    public Long timeMostRecentMessageEvent = null;
    public LogWriter agentLogWriter = null;
    public FileWriter fwLearningPolicyLog = null;

    //=======================================================================

    public Agent() {

    }
    public Agent (int i) {
	init (i);
    }

    /**
     * Initializes an agent with an id. Creates the log and statistics
     * files. Opens the connections to the message queues.
     *
     * @param id an <code>int</code> value 
     */

    public void init (int id) {

	setID(id);

	random = Statistics.getRandomNumberGenerator();

	File fileCreditPolicyLog = new File( SimulationContext.getLogDir() + File.separator + Constants.AGENT_LOG_FILE_NAME_PREFIX + id + ".STATISTICS" );		

	try {
	    if ( Constants.OPTION_WRITE_AGENT_LOGS ) {

		agentLogWriter = new LogWriter( SimulationContext.getLogDir() + File.separator + Constants.AGENT_LOG_FILE_NAME_PREFIX + getIDString( id ) + Constants.AGENT_LOG_FILE_NAME_SUFFIX  );
	    }
	    
	    if ( Constants.OPTION_WRITE_AGENT_STATISTICS ) {
		fwLearningPolicyLog = new FileWriter( fileCreditPolicyLog );
	    }

	}
	catch( Exception e ) {
	    e.printStackTrace();
	}				

	/** ------------ THE AGENT's MEMORY --------------------------- **/		

	agentModels = new AgentModels();		

	currentQuery = null;			
	setAgentsWhomHaveBeenSentCurrentQuery = new HashSet();

	referralChain = null;
	unprocessedReferrals = new UnprocessedReferrals();
	cache = new Cache();

	try {		
	    // Create the session
	    queueSession = SimulationContext.getQueueConnection().createQueueSession( false, Session.AUTO_ACKNOWLEDGE );          

	    // Create sender (not associated with any particular queue)
	    queueSender = queueSession.createSender( null );
	    // Look up the destination
	    Queue queueInboundMessages = (Queue)SimulationContext.getJNDIContext().lookup( getQueueName( id ) );
	    // Create a subscriber
	    queueReceiver = queueSession.createReceiver( queueInboundMessages );
	    // Set the message listener, which is this class since we implement
	    // the MessageListener interface
	    queueReceiver.setMessageListener( this);  						
	}

	catch ( Exception e ) {
	    e.printStackTrace();
	}

    }
	
    /**
     * Generates a new question and sends it to all of the neighbors
     * who this agent believes are knowledgeable about the query.
     * 
     */

    public void askQuestion() {			
	//Generates a new query and stores in it 'currentQuery' member
	generateQuery();

	// Create a new referral chain specifically for this question.
	referralChain = new ReferralChainDAG( currentQuery, this.id );

	// Initialize for a new query
	//setAgentsWhomHaveBeenSentCurrentQuery.clear();


	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Because the policy implementation could be written by anyone, 
	// we want to guard against the policy developer altering our objects.
	// So, we clone the objects that are passed as "read-only" parameters.
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	-				
	Iterator iteratorNeighborsToAsk = askingPolicy.getNeighborsToAsk( currentQuery.copy(), neighbors.copy(), agentModels.copy() );

	long referralChainID = counterReferralChains++;		

	while (iteratorNeighborsToAsk.hasNext()) {

	    Integer neighborID = (Integer)iteratorNeighborsToAsk.next();		

	    // add the neighbor that we just sent the question to as a referral
	    
	    referralChain.addReferral( this.id, neighborID.intValue() );		

	    // We synchronize on this block to make sure that the 
	    // setAgentsWhomHaveBeenSentCurrentQuery structure accurately reflects
	    // those agents whom have been sent the query.
	    // Otherwise, another thread could possibly send a message to
	    // the same neighbor if it was referred by some other agent.	

	    synchronized (setAgentsWhomHaveBeenSentCurrentQuery) {

		sendCurrentQuery( neighborID.intValue(), referralChainID  );
	    }

	}	

    }	

    //=======================================================================

    /**
     * <code>sendCurrentQuery</code> 
     * Creates a question message containing the current query, and then
     * sends the message to the specified agent.
     *
     * @param agentID an <code>int</code> To whom the message will go
     * @param referralChainID a <code>long</code> The referral chain id for the query
     */
    public void sendCurrentQuery(int agentID, long referralChainID) {		

	try {

	    MapMessage msgQuestion = queueSession.createMapMessage();

	    msgQuestion.setString( Constants.MESSAGE_FIELD_TYPE, Constants.MESSAGE_QUESTION );
	    msgQuestion.setInt( Constants.MESSAGE_FIELD_SENDER_ID, this.id );
	    msgQuestion.setString( Constants.MESSAGE_FIELD_QUERY, currentQuery.toPreciseString() );
	    msgQuestion.setLong( Constants.MESSAGE_FIELD_QUERY_ID, currentQueryID ); 
	    msgQuestion.setLong( Constants.MESSAGE_FIELD_REFERRAL_CHAIN_ID, referralChainID  ); 				
	    log( "sending query to agent(" + agentID + ")" );

	    send( msgQuestion, agentID );
	    rememberSentCurrentQuery( agentID );	
	}
	catch( Exception e ) {
	    e.printStackTrace();
	}

    }

	

    //=======================================================================

    /**
     * Close all messaging and file connections.
     */

    public void close() {

	try {

	    if ( queueSender != null ) 
		queueSender.close(); 

	    if ( queueReceiver != null ) {
		queueReceiver.setMessageListener(null);
		queueReceiver.close();
	    }
	    
	    if ( queueSession != null) 
		queueSession.close(); 

	    if ( agentLogWriter != null)  
		agentLogWriter.close(); 

	    if ( fwLearningPolicyLog != null )  
		fwLearningPolicyLog.close();

	}

	catch( Exception e ) {
	    e.printStackTrace();
	}

    }		

    //=======================================================================

    /**
     * Generates a new query for this agent based on its interests
     * @returns ID of the query that has been generated
     */

    public void generateQuery() {
	long currentQueryID = counterQueriesGenerated++;	

	currentQuery = queryGenerationPolicy.generateQuery(expertise, interests, random);
	//	generateInterestQuery();

	log( "-----------------------------------------------------------------" );
	log( " QUERY #" + currentQueryID );
	log( "-----------------------------------------------------------------" );
	log( "generated query #" + currentQueryID + " : (" + currentQuery.toTruncatedString() + ")" );
    }


    //=======================================================================

    /**
     * Returns this agent's Expertise
     */

    public Expertise getExpertise() {
	return expertise;

    }	

    /**
     * Returns the agent's neighbors
     *
     * @return a <code>Neighbors</code> value
     */
    public Neighbors getNeighbors() {
	return neighbors;

    }	

    /**
     * Returns the agent's models of others
     *
     * @return an <code>AgentModels</code> value
     */
    public AgentModels getAgentModels() {
	return agentModels;

    }	

    //=======================================================================

    /**
     * Returns this agent's Interests
     */

    public Interests getInterests() {
	return interests;
    }

    //========================================================================		
    /**
     * Returns this agent's id
     */

    public int getID() {
	return id;
    }


    /**
     * Sets the agent's id
     *
     * @param i an <code>int</code> is the agent's id
     */
    public void setID(int i) {
	id = i;
    }

    //=======================================================================

    /**
     * Converts the an int ID into a string ID - with the appropriate leading 0's
     */

    //changed from private to public by raghuram on Mar 25th '02
    public static final String getIDString( int id ) {

	String stringID = Integer.toString( id );
	StringBuffer bufID = new StringBuffer();

	for( int digits=stringID.length(); digits<4; digits++ )
	    bufID.append( "0" );

	bufID.append( stringID );

	return bufID.toString();

    }	

    

    //=======================================================================



    // Setting the policies for this agent

	

    public void setAskingPolicy(AskingPolicy policyImplementation) { 
	this.askingPolicy = policyImplementation; 
    }

    public void setAnsweringPolicy(AnsweringPolicy policyImplementation) { 
	this.answeringPolicy = policyImplementation; 
    }

    public void setReferringPolicy(ReferringPolicy policyImplementation) { 
	this.referringPolicy = policyImplementation; 
    }

    public void setLearningPolicy( LearningPolicy policyImplementation) {

	this.learningPolicy = policyImplementation; 
	this.learningPolicy.setFileWriter( this.fwLearningPolicyLog );
    }

    public void setChangingNeighborsPolicy( ChangingNeighborsPolicy policyImplementation) {

	this.changingNeighborsPolicy = policyImplementation; 
	this.changingNeighborsPolicy.setLogWriter( this.agentLogWriter );
    }	
        
    public void setCachingPolicy( CachingPolicy policyImplementation ){ 
	this.cachingPolicy = policyImplementation; 
    }

    public void setQueryGenerationPolicy (QueryGenerationPolicy policyImplementation ){ 
	this.queryGenerationPolicy = policyImplementation;
    }
    //=======================================================================

    /**
     * Sets this agent's Neighbors
     */

    public void setAgentModels(AgentModels agentModels) {
	this.agentModels = agentModels;
    }	

    //=======================================================================

    /**
     * Sets this agent's Neighbors
     */

    public void setNeighbors(Neighbors neighbors) {
	this.neighbors = neighbors;
    }

    //=======================================================================

    /**
     * Sets this agent's Expertise
     */

    public void setExpertise( Expertise expertise) {
	this.expertise = expertise;
    }

    //=======================================================================

    /**
     * Sets this agent's Interests
     */

    public void setInterests( Interests interests) {
	this.interests = interests;
    }

    //=======================================================================

    /**
     * Returns the number of messages received by this agent since its creation
     */

    public int getMessagesReceived() {
	return counterMessagesReceived;
    }

    //=======================================================================

    /**
     * Returns the number of messages sent by this agent since its creation
     */

    public int getMessagesSent() {
	return counterMessagesSent;

    }

    //=======================================================================

    /**
     * Returns the JNDI name of this agent's queue
     */

    public static final String getQueueName(int id) {
	return ( "queue/queueAgent" + getIDString(id) );
    }

    //=======================================================================

    /**
     * Handles all incoming Answer messages
     */

    public void handleAnswerMessage( MapMessage msgAnswer ) throws javax.jms.JMSException {

	// parse the Answer message

	int answererID = msgAnswer.getInt( Constants.MESSAGE_FIELD_SENDER_ID );				

	Answer answer = Answer.parse( msgAnswer.getString(Constants.MESSAGE_FIELD_ANSWER) );
	long  queryID = msgAnswer.getLong( Constants.MESSAGE_FIELD_QUERY_ID );		
	long referralChainID = msgAnswer.getLong( Constants.MESSAGE_FIELD_REFERRAL_CHAIN_ID );
	
	if ( queryID != currentQueryID )
	    ErrorHandler.exit( "Received an answer to an old query!" );
		

	log( "received answer(" + answer.toTruncatedString() + ") from agent(" + answererID + ")" );
	referralChain.recordAnswer( answererID, answer );		

    }

    //=======================================================================
    /**
     * Handles all incoming Question messages
     */

    private void handleQuestionMessage( MapMessage msgQuestion ) throws javax.jms.JMSException {

	// parse the Question message

	int questionerID = msgQuestion.getInt(Constants.MESSAGE_FIELD_SENDER_ID);
	long queryID = msgQuestion.getLong(Constants.MESSAGE_FIELD_QUERY_ID);	
	long referralChainID = msgQuestion.getLong(Constants.MESSAGE_FIELD_REFERRAL_CHAIN_ID);

	Query query = Query.parse( msgQuestion.getString(Constants.MESSAGE_FIELD_QUERY) );		


	log( "received query(" + query.toTruncatedString() + ") from agent(" + questionerID + ")" );

	AgentModel questionerModel = agentModels.getModel( new Integer(questionerID) );		

	// Update the model of the agent who sent the question				
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Because the policy implementation could be written by anyone, 
	// we want to guard against the policy developer altering our objects.
	// So, we clone the objects that are passed as "read-only" parameters.
	//
	// Since the purpose of the learning policy here is to alter the 
	// questionerModel we MUST send the actual questionerModel rather
	// than a copy of it.
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	-			

	learningPolicy.learnFromQuestion( questionerModel, query.copy() );	
	// Consider generating an answer to the query.
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Because the policy implementation could be written by anyone, 
	// we want to guard against the policy developer altering our objects.
	// So, we clone the objects that are passed as "read-only" parameters.
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	-			
                
                
	//Before we generate the answer, lets check to see if it in the cache
	Answer answer = getResponseFromCache(query);
                
	if( answer == null) 
	    {
		//Not found in cache, therefore try to generate an answer
		answer = answeringPolicy.generateAnswer( questionerModel.copy(), query.copy(), this.expertise, this.interests, this.id, this.neighbors);			
	    }
	else
	    log("Retrieved answer from cache");
                    
	// if this agent generated an answer to the query, then send it to the questioning agent

	if ( answer != null) {

	    // create the answer message

	    MapMessage msgAnswer = queueSession.createMapMessage();

	    msgAnswer.setString( Constants.MESSAGE_FIELD_TYPE, Constants.MESSAGE_ANSWER );
	    msgAnswer.setInt( Constants.MESSAGE_FIELD_SENDER_ID, this.id );
	    msgAnswer.setLong( Constants.MESSAGE_FIELD_QUERY_ID, queryID );
	    msgAnswer.setLong( Constants.MESSAGE_FIELD_REFERRAL_CHAIN_ID, referralChainID );
	    msgAnswer.setString( Constants.MESSAGE_FIELD_ANSWER, answer.toPreciseString() );

	    // send the answer message
	    send( msgAnswer, questionerID );

	    log( "sent answer(" + answer.toTruncatedString() + ") to agent(" + questionerID + ")" );			

	}

	// otherwise, consider sending the questioning agent some referrals

	else {

	    // synchronize with code blocks that modify this agent's neighbors
	    synchronized( neighbors) {

		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// Because the policy implementation could be written by anyone, 
		// we want to guard against the policy developer altering our objects.
		// So, we clone the objects that are passed as "read-only" parameters.
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		Iterator iteratorReferrableNeighbors = referringPolicy.getNeighborsToReferTo( questionerModel.copy(), query.copy(), neighbors.copy(), agentModels.copy(), this.id );

		if ( iteratorReferrableNeighbors != null) {

		    // iterate over all the neighbors who this agent feels comfortable
		    // referring the questioning agent to				
		    while( iteratorReferrableNeighbors.hasNext()) {
			int neighborID = ((Integer)iteratorReferrableNeighbors.next()).intValue();
			// Make sure that we do NOT refer a questioner to himself
			if ( neighborID != questionerID) {
			    // create the referral message
			    MapMessage msgReferral = queueSession.createMapMessage();

			    msgReferral.setString( Constants.MESSAGE_FIELD_TYPE, Constants.MESSAGE_REFERRAL );
			    msgReferral.setInt( Constants.MESSAGE_FIELD_SENDER_ID, this.id );
			    msgReferral.setLong( Constants.MESSAGE_FIELD_QUERY_ID, queryID );
			    msgReferral.setLong( Constants.MESSAGE_FIELD_REFERRAL_CHAIN_ID, referralChainID );				
			    msgReferral.setInt( Constants.MESSAGE_FIELD_REFERRED_AGENT_ID, neighborID );

			    // send the referral message
			    send( msgReferral, questionerID );
			    log( "referring agent(" + neighborID + ") to agent(" + questionerID + ")" );		
			} //end if
		    } //end while
		} //end if
	    } //end synchrinized
	} //end else

    }

    //=======================================================================

    /**
     * Handles all incoming Referral messages
     */

    private void handleReferralMessage( MapMessage msgReferral ) throws javax.jms.JMSException {

	// parse the referral message

	int referringAgentID = msgReferral.getInt( Constants.MESSAGE_FIELD_SENDER_ID );
	int referredAgentID = msgReferral.getInt( Constants.MESSAGE_FIELD_REFERRED_AGENT_ID );		
	long referralChainID = msgReferral.getLong(Constants.MESSAGE_FIELD_REFERRAL_CHAIN_ID);			
	long queryID = msgReferral.getLong( Constants.MESSAGE_FIELD_QUERY_ID );				

	log( "received referral to agent(" + referredAgentID + ") from agent(" + referringAgentID + ")" );

	if ( queryID != currentQueryID )
	    ErrorHandler.exit( "Received a referral to an OLD query" );

	int lengthOfReferralChain = referralChain.getChainLength( referringAgentID );

	//if ( (referredAgentID != this.id) && (lengthOfReferralChain < Constants.MAXIMUM_REFERRAL_CHAIN_LENGTH) )

	if ( lengthOfReferralChain < Constants.MAXIMUM_REFERRAL_CHAIN_LENGTH) {

	    referralChain.addReferral( referringAgentID, referredAgentID );

	    // We synchronize on this block to make sure that the 
	    // setAgentsWhomHaveBeenSentCurrentQuery structure accurately reflects
	    // those agents whom have been sent the query.
	    // Otherwise, another thread could possibly send a message to
	    // the same neighbor if it was referred by some other agent.		
	    
	    synchronized (setAgentsWhomHaveBeenSentCurrentQuery) {
		// if this agent has not already sent the specified query to the referred agent,
		if ( !haveSentCurrentQuery(referredAgentID) && (referredAgentID != this.id)) {

		    // then ask the referred agent the answer to the query
		    sendCurrentQuery( referredAgentID, referralChainID );			

		}				
		else {
		     
		    Enumeration enumDescendantsOfReferredAgent = referralChain.getDescendants( referredAgentID );
		    while ( enumDescendantsOfReferredAgent.hasMoreElements() ) {

			Integer agentID = (Integer)enumDescendantsOfReferredAgent.nextElement();
			Enumeration enumReferralMessages = unprocessedReferrals.removeReferrals( referralChainID, agentID.intValue() ); 

			if ( enumReferralMessages != null ) {
			    while ( enumReferralMessages.hasMoreElements()) {
				log( "handling a stored referral message because the chain length is now less than the limit." );					
				MapMessage referralMessage = (MapMessage)enumReferralMessages.nextElement();
				handleReferralMessage( referralMessage );
			    } //while	
			} //if
		    } //while

		}//else			
	    }//synchronzed				

	} //if

	else {
	    log( "adding the referral to storage because using it would exceed the maximum chain length" );
	    unprocessedReferrals.addReferral( referralChainID, referringAgentID, msgReferral );	
	}

    }		



    //=======================================================================	

    /**
     * @return true if this agent has sent the specified query to the
     * specified agent; false otherwise
     */

    private boolean haveSentCurrentQuery( int _agentID) {

	return setAgentsWhomHaveBeenSentCurrentQuery.contains(new Integer(_agentID));
    }

    //=======================================================================

    /**
     * Implementation of the MessageListener interface for JMS.
     * If this agent has one or more messages waiting in it's JMS queue, 
     * JMS will eventually invoke this method and deliver one of the messages.
     *
     * This method simply sorts the messages by type and then invokes the
     * appropriate message handler.
     */

    public void onMessage(Message msg) {

	// record the time of the most recent message received

	timeMostRecentMessageEvent = new Long( System.currentTimeMillis() );

	counterMessagesReceived++;

	if ( stateMessageListening == Constants.IGNORE_ALL_MESSAGES ) {

	    log( "Agent(" + this.id + " ): Received a message, but is ignoring ALL messages" );
	    return;
	}

	MapMessage msgMap = null;

	// Unpack the message, be carefull when casting to correct message type
	// onMessage should not throw any application exceptions.

	try{			

	    msgMap = (MapMessage)msg;

	    String type = msgMap.getString( Constants.MESSAGE_FIELD_TYPE );

	    if ( type.equals(Constants.MESSAGE_QUESTION) )
		handleQuestionMessage( msgMap );

	    else if ( type.equals(Constants.MESSAGE_REFERRAL) )
		handleReferralMessage( msgMap );

	    else if ( type.equals(Constants.MESSAGE_ANSWER) )
		handleAnswerMessage( msgMap );
	    else
		ErrorHandler.exit( "Agent.onMessage() could not handle the type of message that it received" );				

	}

	catch(Exception e) { 
	    ErrorHandler.exit( e );

	}

    }

	
    public String getXMLTag(){

	return Constants.XML_TAG_AGENT;
    }

    //=======================================================================

    /**

     * Parses the agent represented by the specified xml element.

     */

    public static Agent parse( IXMLElement xmlAgent, Expertise ex, Interests i) {

	Agent agent = null;

	int agentID = -1;				

	Expertise expertise = ex;
	Interests interests = i;
	Neighbors neighbors = new Neighbors();			
	AgentModels agentModels = new AgentModels();	

	AskingPolicy p_asking = null;
	AnsweringPolicy p_answering = null;
	ReferringPolicy p_referring = null;
	LearningPolicy p_learning = null;
	ChangingNeighborsPolicy p_changingNeighbors = null;
	CachingPolicy p_cachingPolicy = new CachingPolicy_Standard();
	QueryGenerationPolicy p_query = new QueryGenerationPolicy_Standard();
	Double number = null;		

	try {

	    // Parse all the attributes

	    Properties attributes = xmlAgent.getAttributes();
	    agentID = Integer.parseInt( attributes.getProperty(Constants.XML_ATTRIBUTE_ID) );

	    // Parse all the child elements
	    Vector vecChildElements = xmlAgent.getChildren();

	    Enumeration enumChildElements = vecChildElements.elements();

	    while (enumChildElements.hasMoreElements()) {

		IXMLElement xmlChildElement = (IXMLElement)enumChildElements.nextElement();
		// parse Expertise

		if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_EXPERTISE))
		    expertise = expertise.parse( xmlChildElement );

		// parse Interests
		else if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_INTERESTS))
		    interests = interests.parse( xmlChildElement );

		// parse Neighbors
		else if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_NEIGHBORS))
		    neighbors = Neighbors.parse( xmlChildElement );

		// parse Agent Model
		else if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_AGENT_MODEL)) {
		    AgentModel model = AgentModel.parse( xmlChildElement );		
		    agentModels.addModel( model );
		}

		// load the policies that an agent will use unless
		// they are overridden at the group level
		else if (xmlChildElement.getName().equals(Constants.XML_TAG_POLICIES)) {

		    Enumeration enumChildrenOfPolicies = xmlChildElement.getChildren().elements();					
		    while (enumChildrenOfPolicies.hasMoreElements()) {
			IXMLElement xmlChildOfPolicies = (IXMLElement)enumChildrenOfPolicies.nextElement();
			if (xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_ASKING))
			    p_asking = (AskingPolicy)Class.forName( xmlChildOfPolicies.getContent() ).newInstance();
			else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_ANSWERING) )
			    p_answering = (AnsweringPolicy)Class.forName( xmlChildOfPolicies.getContent() ).newInstance();
			else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_REFERRING) )
			    p_referring = (ReferringPolicy)Class.forName( xmlChildOfPolicies.getContent() ).newInstance();
			else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_LEARNING) )
			    p_learning = (LearningPolicy)Class.forName( xmlChildOfPolicies.getContent() ).newInstance();
			else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_CHANGING_NEIGHBORS) )
			    p_changingNeighbors = (ChangingNeighborsPolicy)Class.forName( xmlChildOfPolicies.getContent() ).newInstance();
			else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_CACHING) )
			    p_cachingPolicy = (CachingPolicy)Class.forName( xmlChildOfPolicies.getContent() ).newInstance();
			else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_QUERY_GENERATION) )
			    p_query = (QueryGenerationPolicy)Class.forName( xmlChildOfPolicies.getContent() ).newInstance();
		    } //end while
		} //end if	
	    } //end while

	    agent = new Agent( agentID );

	    agent.setInterests( interests );
	    agent.setExpertise( expertise );
	    agent.setNeighbors( neighbors );
	    agent.setAgentModels( agentModels );	

	    agent.setAskingPolicy( p_asking );
	    agent.setAnsweringPolicy( p_answering );
	    agent.setReferringPolicy( p_referring );
	    agent.setLearningPolicy( p_learning );
	    agent.setChangingNeighborsPolicy( p_changingNeighbors );			
	    agent.setCachingPolicy(p_cachingPolicy);
	    agent.setQueryGenerationPolicy(p_query);
	}

	catch(Exception e) {
	    e.printStackTrace();
	}			

	return agent;		

    }	

    //=======================================================================

    /**
     * 
     */

    public void toXML (IXMLElement xmlAgent) {

	try {
	    // Set all the attributes
	    xmlAgent.setAttribute( Constants.XML_ATTRIBUTE_ID, Integer.toString(this.id) );

	    // Set the Expertise
	    IXMLElement xmlExpertise = xmlAgent.createElement( Constants.XML_TAG_EXPERTISE );
	    this.expertise.toTruncatedXML( xmlExpertise );			
	    xmlAgent.addChild( xmlExpertise );

	    // Set the Interests			
	    IXMLElement xmlInterests = xmlAgent.createElement( Constants.XML_TAG_INTERESTS );
	    this.interests.toTruncatedXML( xmlInterests );			
	    xmlAgent.addChild( xmlInterests );

	    // Set the Neighbors			
	    IXMLElement xmlNeighbors = xmlAgent.createElement( Constants.XML_TAG_NEIGHBORS );
	    this.neighbors.toXML( xmlNeighbors );			
	    xmlAgent.addChild( xmlNeighbors );

	    // Only do the following if the parameters indicate that 
	    // the state file must contain enough information to be 
	    // read in as an agent input file.
	    if (Constants.OPTION_WRITE_READABLE_STATE_FILES) {

		// ------ Set all of the Policies -----------
		IXMLElement xmlPolicies = xmlAgent.createElement( Constants.XML_TAG_POLICIES );
		xmlAgent.addChild( xmlPolicies );
		IXMLElement xmlPolicy = null;

		xmlPolicy = xmlAgent.createElement(Constants.XML_TAG_POLICY_ASKING);
		xmlPolicy.setContent( this.askingPolicy.getClass().getName() );
		xmlPolicies.addChild( xmlPolicy );			

		xmlPolicy = xmlAgent.createElement( Constants.XML_TAG_POLICY_ANSWERING );
		xmlPolicy.setContent( this.answeringPolicy.getClass().getName() );
		xmlPolicies.addChild( xmlPolicy );			

		xmlPolicy = xmlAgent.createElement( Constants.XML_TAG_POLICY_REFERRING );
		xmlPolicy.setContent( this.referringPolicy.getClass().getName() );
		xmlPolicies.addChild( xmlPolicy );			

		xmlPolicy = xmlAgent.createElement( Constants.XML_TAG_POLICY_LEARNING );
		xmlPolicy.setContent( this.learningPolicy.getClass().getName() );
		xmlPolicies.addChild( xmlPolicy );			

		xmlPolicy = xmlAgent.createElement( Constants.XML_TAG_POLICY_CHANGING_NEIGHBORS );
		xmlPolicy.setContent( this.changingNeighborsPolicy.getClass().getName() );
		xmlPolicies.addChild( xmlPolicy );			
                              
		xmlPolicy = xmlAgent.createElement( Constants.XML_TAG_POLICY_CACHING );
		xmlPolicy.setContent( this.cachingPolicy.getClass().getName() );
		xmlPolicies.addChild( xmlPolicy );			

		xmlPolicy = xmlAgent.createElement( Constants.XML_TAG_POLICY_QUERY_GENERATION );
		xmlPolicy.setContent( this.queryGenerationPolicy.getClass().getName() );
		xmlPolicies.addChild( xmlPolicy );			

		// ------ Set all of the Agent Models -----------				

		Iterator iteratorAgentModels = agentModels.getModels();
		while (iteratorAgentModels.hasNext()) {

		    AgentModel model = (AgentModel)iteratorAgentModels.next();
		    IXMLElement xmlAgentModel = xmlAgent.createElement( Constants.XML_TAG_AGENT_MODEL );
		    model.toXML( xmlAgentModel );			
		    xmlAgent.addChild( xmlAgentModel );			
		} //end while
	    } //end if

	} //end try

	catch (Exception e) {
	    e.printStackTrace();
	}				

    }	

    //=======================================================================	

    /**
     * Remember the fact that this agent has sent the speficied query to the 
     * specified agent.
     */

    private void rememberSentCurrentQuery (int _agentID) {

	Integer agentID = new Integer( _agentID );

	synchronized( setAgentsWhomHaveBeenSentCurrentQuery) {
	    setAgentsWhomHaveBeenSentCurrentQuery.add( agentID );

	}

    }

    //=======================================================================	

    /**
     * Send the specified JMS MapMessage to the queue that belongs to the 
     * specified agent.
     */

    private void send( MapMessage message, int agentID ) throws JMSException {

	// Send the message
	try {

	    Context context = new InitialContext();
	    Queue queue = (Queue)context.lookup( getQueueName(agentID) );

	    // record the time of the most recent message sent
	    timeMostRecentMessageEvent = new Long( System.currentTimeMillis() );			
	    queueSender.send( queue, message);
	    counterMessagesSent++;
	}

	catch(Exception e) {
	    e.printStackTrace();
	}	

    }

    //=======================================================================	

    /**
     * Determines whether this agent will process the messages that JMS 
     * delivers to it
     */

    public void setMessageListeningState (int state) {
	this.stateMessageListening = state;
    }

    //=======================================================================	

    /**
     * Based on our agent models, remove bad agents from our set of neighbors
     * and add good agents to our set of neighbors.
     */

    public void changeNeighbors() {

	log( "-----------------------------------------------------------------" );
	log( " CHANGING NEIGHBOR SET" );
	log( "-----------------------------------------------------------------" );


	// synchronize with code blocks that reference this agent's neighbors
		synchronized(neighbors) {

	    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	    // Because the policy implementation could be written by anyone, 
	    // we want to guard against the policy developer altering our objects.
	    // So, we clone the objects that are passed as "read-only" parameters.
	    //
	    // Since the purpose of the policy is to update the neighbors,
	    // we MUST pass the actual object, and not a copy.			
	    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	-
		

	changingNeighborsPolicy.changeNeighbors( this.neighbors, this.agentModels, this.expertise.copy(), this.interests.copy() );	
	    }
    }

    //=======================================================================

    public void learnFromResponses() {
	learningPolicy.learnFromResponses( this.referralChain, this.agentModels, this.expertise);

    }	
        
    public void cacheBestResponse() {

	Answer answer = null;
	Query query = null;
	//Try to cache only if caching policy is define
	if(cachingPolicy.isImplemented())
            {
                query = ((ReferralChainDAG)referralChain).getQuery();
                answer = ((ReferralChainDAG)referralChain).getBestAnswer();
                cache.store(query, answer);
            }
    }
        
    private Answer getResponseFromCache(Query query) {

	Answer answer = null;
            
	if(cachingPolicy.isImplemented())
	    answer = cache.fetch(query);
            
	return answer;
    }

    //=======================================================================

    public void resetMessageCounters() {

	counterMessagesReceived = 0;
	counterMessagesSent = 0;	
    }

    //=======================================================================

    public void startAcceptingResponses() {
	stateMessageListening = Constants.LISTEN_TO_ALL_MESSAGES;

    }

    //=======================================================================

    public void stopAcceptingResponses() {
	stateMessageListening = Constants.IGNORE_ALL_MESSAGES;
    }

    //=======================================================================

    public Long getTimeOfMostRecentMessageEvent() {
	return timeMostRecentMessageEvent;
    }	

    //=======================================================================

    private void log() {
	log( "" );
    }	

    //=======================================================================
    //changed from private to public - raghuram - March
    public void log(String text)  {

	if (agentLogWriter != null)
	    agentLogWriter.log(text);
    }	

    public LearningPolicy getLearningPolicy() {
	return learningPolicy;
    }

}
