package edu.ncsu.simulators.analysis;

import Jama.*; 
import java.util.*;
import java.lang.reflect.*;
/**
 * AuthorityMetrics.java
 *
 *
 * Created: Fri Nov  8 15:56:45 2002
 *
 * @author <a href="mailto:pyolum@eos.ncsu.edu">Pinar Yolum</a>
 * @version
 */

public class AuthorityMetrics {
    public AuthorityMetrics (){
	
    }
    public static final double PAGERANK_CONSTANT = 0.85;
    /**
     *  Converts the adjacency class to a Matrix class that can be
     *  used by the Jama package.
     *
     * @param a an <code>Adjacency</code> value
     * @return a <code>Matrix</code> value 
     */

    public static double[][] convertToMatrix(Adjacency a){

	double[][] doubletable= new double[a.size()][a.size()]; 
	Iterator current = null;
	int neighbor = 0;

	//initialize the adjacency to zeros.
	for (int i=0; i<a.size(); i++) {
	    for ( int j=0; j<a.size(); j++) {
		doubletable[i][j] = 0.0;
	    } // end of for ()
	} // end of for ()
	
	
	for (int i=0; i<a.size(); i++ ) {
	    current = a.get(i).iterator();
	    //if i doesnt't have outlinks, make it point to everyone else
	    if (!current.hasNext()) {
		for (int k=0;k<a.size();k++)
		    doubletable[i][k] = 1.0;

	    }
	    else {
	    //i has some outlinks		 
		while ( current.hasNext()) {
		    neighbor = ((Integer)current.next()).intValue();
		    doubletable[i][neighbor] = 1.0;
		} // end of while ()

	    } // end of else
	    
	} // end of for ()
	return doubletable;

    }

    public static Matrix uniform(int dimension) {

	//	System.out.println("dimension" + dimension);
	String s = Integer.toString(dimension);
	Double d = new Double(s);

	double[][] doubles = new double[dimension][dimension];
	for (int i=0; i<dimension; i++) {
	    for (int j=0; j<dimension; j++) {
		doubles[i][j] = 1/d.doubleValue();
	    } // end of for ()
	    
	} // end of for ()
	return new Matrix(doubles);
    }

    public static Matrix normalize(double[][] doubles, int dimension) {

	double sum = 0.0;

	for (int i=0; i<dimension; i++) {
	    sum = 0.0;
	    for (int j=0; j<dimension; j++) {
		sum += doubles[i][j];
	    }	
	    //System.out.println("sum " + sum);

	    for (int k=0; k<dimension; k++) 
		doubles[i][k] = doubles[i][k]/sum;
	}
	return new Matrix(doubles);
		
    }
    /**
     * Find  the maximum.
     *
     * @param d a <code>double[]</code> value
     * @param dimension an <code>int</code> value
     * @return an <code>int</code> value
     */
    public static double max (double[] d){

	double maximum = 0.0;
	//int index = 0;
	for (int i=0; i<Array.getLength(d); i++ ) {
	    //Check if you need absolute values
	    if (d[i]>maximum) {
		maximum = d[i];
		//index = i;
	    } // end of if ()
	} // end of for ()

	return maximum;
	
    }

    /**
     * Returns the index of d
     *
     * @param d a <code>double</code> value
     * @param m a <code>Matrix</code> value
     * @return an <code>int</code> value
     */
    public static int getIndex(double d, Matrix m){

	int index  = 0;

	for ( int i=0; i<m.getRowDimension() ; i++) {
	    for (int j=0; j<m.getColumnDimension() ; j++) {
		if (m.get(i, j) == d) {
		    index = i;
		} // end of if ()
	    } // end of for ()
	} // end of for ()
	return index;
	
    }

    /**
     * Finds the dominant eigenVector
     *
     * @param m a <code>Matrix</code> value
     * @return a <code>Matrix</code> value
     */
    public static Matrix eigenVector(Matrix m) {

	int primary = 0;
	//get the decomposition of m
	EigenvalueDecomposition e = m.eig();
	// get all the eigenvalues
	double[] values = e.getRealEigenvalues();

	Matrix eigenvalues = e.getD();
	//	eigenvalues.print(2,2);
	//find the index of the matrix that has the greates eigenvalue
	primary = getIndex(max(values), eigenvalues);
	
	Matrix eigenvectors = e.getV();
	//	eigenvectors.print(2,2);

	//get the column with the greates value
	Matrix dominant = eigenvectors.getMatrix(0, eigenvectors.getColumnDimension()-1, primary, primary);

	return dominant;

    }

    /**
     * Converts a matrix into a vector
     *
     * @param m a <code>Matrix</code> value
     * @return a <code>Vector</code> value
     */
    public static Vector toVector(Matrix m) {
	
	Vector v = new Vector();
	
	for (int i=0; i<m.getRowDimension(); i++) {
	    for (int j=0; j<m.getColumnDimension(); j++ ) {
		v. add(new Double(m.get(i, j)));
	    } // end of for ()
	    
	} // end of for ()

	return v;
    }

    /**
     *  Returns the pageRanks as a Matrix
     *
     * @param adjacencyTable an <code>Adjacency</code> value
     * @return a <code>Matrix</code> value
     */
    public static Matrix findPageRanks(Adjacency adjacencyTable) {

	Matrix U = uniform(adjacencyTable.size());
	Matrix M = normalize(convertToMatrix(adjacencyTable), adjacencyTable.size());
	Matrix pageRanks = 
	    ((U.times(1-PAGERANK_CONSTANT)).plus(M.times(PAGERANK_CONSTANT))).transpose();

	return eigenVector(pageRanks);
    }

    /**
     * Finds the hubs.
     *
     * @param adjacencyTable an <code>Adjacency</code> value
     * @return a <code>Matrix</code> value
     */
    public static Matrix findHubs(Adjacency adjacencyTable) {

	Matrix m = new Matrix(convertToMatrix(adjacencyTable));
	return eigenVector(m.times(m.transpose()));
    }

    /**
     * Finds the authorities.
     *
     * @param adjacencyTable an <code>Adjacency</code> value
     * @return a <code>Matrix</code> value
     */
    public static Matrix findAuthorities(Adjacency adjacencyTable) {

	Matrix m = new Matrix(convertToMatrix(adjacencyTable));
	return eigenVector(m.transpose().times(m));
    }

    /**
     * Returns true even if the difference between one element in two
     * tables are greater than diff.
     *
     * @param table1 a <code>double[][]</code> value
     * @param table2 a <code>double[][]</code> value
     * @param size an <code>int</code> value
     * @param diff a <code>double</code> value
     * @return a <code>boolean</code> value 
     */
    public static boolean lessDifferent(double[][] table1, double[][] table2, int size, double diff) {
	
	boolean ismore = true;
	int i = 0;
	int j = 0;

	while ((i<size) && (!ismore)) {

	    while ((j<size) && (!ismore)) {
		ismore = (Math.abs(table1[i][j]-table2[i][j]) > diff) ;
		j++;
	    }
	    i++;
	} // end of while ()
	
	return !ismore;

    }
    
}// AuthorityMetrics

