/*
 * Metrics.java
 *
 * Created on October 10, 2001, 3:09 PM
 */

package edu.ncsu.simulators.analysis;

import java.util.*;
import edu.ncsu.simulators.agent.*;
/**
 *
 * @author  pirnal
 * @version 
 */

  /** All of the methods here are static. All are some kind of metrics
   *  some of them are never used. 
   *  and finding components doesn't work!
   */
public class Metrics {

    public static final double PAGERANK_CONSTANT = 0.85;
    public static final double MIN_SOCIABILITY = 0.3;
    public static final double UTILITY_CONSTANT = 0.85;

    /** Creates new Metrics */
    public Metrics() {
    }
    
    /**
     * Calculates the utility of an edge (u, v) (1-\epsilon)(I_u*E_v)
     * + \epsilon*utility(u,w) where w gives the best utility for v,
     * w.
     *
     * @param a an <code>Adjacency</code> value
     * @param u an <code>int</code> value
     * @param v an <code>int</code> value
     * @param oldUtilities a <code>double[][]</code> value
     * @return a <code>double</code> value 
     */
    public static double utility(Adjacency a, VectorTable interest, VectorTable expertise, int u, int v, double[][] oldUtilities)  {

	double max = 0;
	int next = -1;
	int bestW = -1;

	double value = interest.get(u).howSimilartoExpertise(expertise.get(v)) * (1-UTILITY_CONSTANT);

	Set s = findOutgoingNodes (a, v);

	Iterator i = s.iterator();
	
	//for each edge (v,w)
	while (i.hasNext()) {

	    next = ((Integer)i.next()).intValue();
	    //find the highest utility edge
	    if (oldUtilities[v][next] > max) {
		max = oldUtilities[v][next];
		bestW = next;
		
	    } // end of if ()
	    
	} // end of while ()
	
	value += UTILITY_CONSTANT * oldUtilities[u][bestW];

	return value;
	    

    }

    /**
     * Apply utility method to each existing edge.
     *
     * @param a an <code>Adjacency</code> value
     * @param oldUtilities a <code>double[][]</code> value
     * @return a <code>double[][]</code> value
     */
    public static double[][] findUtilities(Adjacency a, VectorTable interest, VectorTable expertise, double[][] oldUtilities) {

	double[][] current = oldUtilities;
	TreeSet t = null;
	int pointer = -1;
	for (int i=0; i<a.size(); i++) {
	    t = a.get(i);
	    Iterator out = t.iterator();

	    while (out.hasNext()) {
		pointer = ((Integer)out.next()).intValue();
		current[i][pointer] = utility(a, interest, expertise, i, pointer, oldUtilities);
		
	    } // end of while ()
	    
	} // end of for ()

	return current;
	
    }


    /**
     * Iterate finding utility untill difference between successive
     * iterations are less than diff
     *
     * @param a an <code>Adjacency</code> value
     * @return a <code>double[][]</code> value 
     */

    public static double[][] iterateUtilities(Adjacency a, VectorTable interest, VectorTable expertise, double diff){
	
	boolean converged = false;

	double[][] utilitytable = new double[a.size()][a.size()];
	double[][] newutilitytable = new double[a.size()][a.size()];
	//initialize the adjacency to zeros.
	for (int i=0; i<a.size(); i++) {
	    for ( int j=0; j<a.size(); j++) {
		if (a.isNeighbor(i, j)) 
		    utilitytable[i][j] = 1.0;		    
		else 
		    utilitytable[i][j] = 0.0;		     
	    } // end of for ()
	} // end of for ()

	while (!converged) {

	    System.out.println("iterating");
	    newutilitytable = Metrics.findUtilities(a, interest, expertise, utilitytable);	    
	    converged = AuthorityMetrics.lessDifferent(utilitytable, newutilitytable, a.size(), diff);
	    utilitytable = newutilitytable;

	} // end of while ()

	return utilitytable;


    }


    public static DoubleVector totalUtilities(Adjacency a, VectorTable interest, VectorTable expertise, double diff){

	double sum = 0.0;
	DoubleVector sums = new DoubleVector();
	double[][] d = iterateUtilities(a, interest, expertise, diff);
	for (int i=0; i<a.size(); i++) {
	    sum = 0.0;
	    for (int j=0; j<a.size(); j++) {
		sum += d[i][j];
	    } // end of for ()
	    sums.add(sum);
	    
	} // end of for ()
	
	return sums;
    }
    /**
     *  <code>findPageRanks</code> finds the pageranks of the nodes.
     *
     * @param a an <code>Adjacency</code> value
     * @param initialPageRanks a <code>Vector</code> value
     * @return a <code>DoubleVector</code> value
     */
    public static DoubleVector findPageRanks(Adjacency a, DoubleVector initialPageRanks) {

	DoubleVector newRanks = new DoubleVector();
	Enumeration e = initialPageRanks.elements();

	Set s; //for the incoming edges B_u
	Iterator pointers = null;
	double sum = 0.0; 
	int pointer = -1;

	for (int i= 0; i<initialPageRanks.size(); i++) {
	    sum = 0.0;
	    s = findIncomingNodes (a, i);
	    pointers = s.iterator();
	    while (pointers.hasNext()) {
		//get the next node that points at i
		pointer = ((Integer)pointers.next()).intValue();
		//add its pagerank/number of neighbors
		sum += (initialPageRanks.elementAt(pointer))/numberOfOutgoingNodes(a, pointer);
		//System.out.println("pointer "+ pointer + "sum " + sum);
		

		
	    } // end of while ()

	    //System.out.println("agent " + i + "sum " + sum);

	    sum = (sum * PAGERANK_CONSTANT) + (1-PAGERANK_CONSTANT);
	    newRanks.add(i, sum);
	    
	} // end of for ()
	return newRanks;
	
    }

    /**
     * Keeps the weights for all agents, still for one dimension.
     *
     * @param allAgents an <code>IntHashTable</code> value
     * @param dimension an <code>int</code> value
     * @return an <code>IntHashTable</code> value
     */
    public static IntHashTable allWeights(IntHashTable allAgents, int dimension) {

	IntHashTable weights = new IntHashTable();
	for (int i=0; i<allAgents.size(); i++) {
	    weights.put(i, totalWeights((AgentModels)allAgents.at(i), dimension));
	} // end of for ()

	return weights;
    }
    

    /**
     * Keeps the weights from a single agent to all agents for one dimension
     *
     * @param models an <code>AgentModels</code> value
     * @param dimension an <code>int</code> value
     * @return an <code>IntHashTable</code> value
     */
    public static IntHashTable totalWeights(AgentModels models, int dimension) {

	Iterator i = models.getModels();
	IntHashTable weights = new IntHashTable();
	double sum = 0.0;

	while (i.hasNext()) {
	    AgentModel a = (AgentModel)i.next();
	    weights.put(a.getID(), a.getSociability().at(dimension));
	} // end of while ()

	return weights;
	    
    }

    public static DoubleVector findSociabilityRanks(Adjacency a, IntHashTable weights, DoubleVector initialRanks) {

	DoubleVector newRanks = new DoubleVector();
	Enumeration e = initialRanks.elements();
	IntHashTable temp = new IntHashTable();

	Set s; //for the incoming edges B_u
	Iterator pointers = null;
	double sum = 0.0; 
	double total =0.0;
	double sociability = 0.0;
	int pointer = -1;

	//	System.out.println("size " + weights.size());
	for (int i= 0; i<initialRanks.size(); i++) {
	    sum = 0.0;
	    s = findIncomingNodes (a, i);
	    //	    System.out.println(" s " + s);
	    
	    pointers = s.iterator();
	    while (pointers.hasNext()) {
		//get the next node that points at i
		pointer = ((Integer)pointers.next()).intValue();
		temp = (IntHashTable)(weights.at(pointer));
		//System.out.println(pointer);
		
		//System.out.println("i " + i + "size " + temp.size());

		sociability = ((Double)temp.at(i)).doubleValue();
		//System.out.println("sociability " + sociability);
		
		total = temp.totalWeights();
		//add its pagerank/number of neighbors
		if (sociability > MIN_SOCIABILITY) {
		    sum += (initialRanks.elementAt(pointer)*sociability)/total;		    
		} // end of if ()
		//System.out.println("pointer "+ pointer + "sum " + sum);
		

		
	    } // end of while ()

	    //System.out.println("agent " + i + "sum " + sum);
	    //if (s.size() != 0) {
	    //	sum = sum/s.size();		
	    //} // end of if ()

	    sum = (sum * PAGERANK_CONSTANT) + (1-PAGERANK_CONSTANT);
	    newRanks.add(i, sum);
	    
	} // end of for ()
	return newRanks;
    }

    public static DoubleVector iterateSociabilityRanks(double diff, Adjacency a, IntHashTable agents, DoubleVector pageRanks, int dimension){
    
	boolean converged = false; //Gets true when the difference between iterations is small

	//Keeps the weights for the dimension 
	IntHashTable weights = allWeights(agents, dimension);
	weights.fill(MIN_SOCIABILITY+0.001);

	DoubleVector pageRanks2 = new DoubleVector();
	while (!converged) {

	    System.out.println("iterating");
	    pageRanks2 = Metrics.findSociabilityRanks(a, weights, pageRanks);	    
	    converged = pageRanks.lessDifferent(pageRanks2, diff);
	    pageRanks = pageRanks2;

	} // end of while ()

	return pageRanks2;
	
    }

    

    /**
     *  <code>iteratePageRanks</code> method iterates for
     *  numberOfIterations times.
     *
     * @param numberOfIterations an <code>int</code> value
     * @param a an <code>Adjacency</code> value
     * @param pageRanks a <code>DoubleVector</code> value
     * @return a <code>DoubleVector</code> value 
     */
    public static DoubleVector iteratePageRanks(int numberOfIterations, Adjacency a, DoubleVector pageRanks){

	for (int i=0; i<numberOfIterations; i++) {
	    pageRanks = findPageRanks(a, pageRanks);
	    
	} // end of for ()
	
	return pageRanks;

    }

    /**
     * repeats finding the pageranks until the difference between two consecutive
     * findings is less than diff
     *
     * @param diff a <code>double</code> value
     * @param a an <code>Adjacency</code> value
     * @param pageRanks a <code>DoubleVector</code> value
     * @return a <code>DoubleVector</code> value
     */
    public static DoubleVector iteratePageRanks(double diff, Adjacency a, DoubleVector pageRanks){
    
	boolean converged = false; //Gets true when the difference between iterations is small

	DoubleVector pageRanks2 = new DoubleVector();
	while (!converged) {

	    System.out.println("iterating");
	    pageRanks2 = Metrics.findPageRanks(a, pageRanks);	    
	    converged = pageRanks.lessDifferent(pageRanks2, diff);
	    pageRanks = pageRanks2;

	} // end of while ()

	return pageRanks2;
	
    }

    /**
     * <code>findOutgoingNodes</code> returns j's neighbors.
     *
     * @param a an <code>Adjacency</code> value
     * @param j an <code>int</code> value
     * @return a <code>TreeSet</code> value
     */
    public static TreeSet findOutgoingNodes(Adjacency a, int j){
        return a.get(j);
    }


    /**
     *  <code>findIncomingNodes</code> returns nodes that point to j.
     *
     * @param a an <code>Adjacency</code> value
     * @param j an <code>int</code> value
     * @return a <code>Set</code> value
     */
    public static Set findIncomingNodes(Adjacency a, int j){
        Set s = new TreeSet();
        for (int i=0; i<a.size(); i++)
            if (a.get(i).contains(new Integer(j)))
                s.add(new Integer(i));
        return s;
    }

    /**
     * Returns the number of nodes that point to j
     *
     * @param a an <code>Adjacency</code> value
     * @param j an <code>int</code> value
     * @return an <code>int</code> value
     */
    public static int numberOfIncomingNodes(Adjacency a, int j){
        return findIncomingNodes(a, j).size();//-1
    }

    /**
     * Rerturns the number of nodes j points to.
     *
     * @param a an <code>Adjacency</code> value
     * @param j an <code>int</code> value
     * @return an <code>int</code> value
     */
    public static int numberOfOutgoingNodes(Adjacency a, int j){
        return findOutgoingNodes(a, j).size();//-1
    }


    /**
     * Aggregates the edge counts to say stg. like two nodes have 15
     *  indegree, three have 12, etc.
     *
     * @param v a <code>Vector</code> value
     * @return a <code>Hashtable</code> value 
     */
    public static Hashtable count(Vector v){
	Hashtable h = new Hashtable();
	Integer k = new Integer(0);
	int sum = 0;
	for (int i=0; i<v.size(); i++){
	    k = (Integer)v.get(i);
	    if (!h.containsKey(k)){
		for (int j=i; j<v.size();j++)
		    if ((((Integer)v.get(j))).intValue() == k.intValue()) sum++;
		h.put(k, new Integer(sum));
	    }
	    sum = 0; 
	}
	return h;
    }		

    /**
     *  Makes a hash table based on how many incoming edges each agent has
     *
     * @param a an <code>Adjacency</code> value
     * @return a <code>Hashtable</code> value
     */
    public static Hashtable incomingNodes(Adjacency a){
        Vector v = new Vector();
	for (int i =0; i<a.size(); i++)
	    v.add(new Integer(numberOfIncomingNodes(a, i)));
	return count(v);
    }

    /**
     * Makes a hash table based on how many outgoing edges each agent has
     *
     * @param a an <code>Adjacency</code> value
     * @return a <code>Hashtable</code> value
     */
    public static Hashtable outgoingNodes(Adjacency a){
        Vector v = new Vector();
	for (int i =0; i<a.size(); i++)
	    v.add(new Integer(numberOfOutgoingNodes(a, i)));
	return count(v);
    }
    
    /**
     * Returns the generality of the whole graph 
     *
     * @param a a <code>VectorTable</code> value
     * @return a <code>double</code> value
     */
    public static double generality(VectorTable a){
        double sum=0;
        for (int i=0;i<a.size();i++)
            sum += a.get(i).generality();
        return sum/a.size();
    }

    /**
     * Finds the similarity clustering between a node j and all the others  
     *
     * @param v a <code>VectorTable</code> value
     * @param a an <code>Adjacency</code> value
     * @param j an <code>int</code> value
     * @return a <code>double</code> value
     */
    public static double interestSimilarity(VectorTable v, Adjacency a, int j){
        //v being the interest vector table
        
        double sum=0.0;
        TreeSet neighbors = findOutgoingNodes(a, j);

        neighbors.add(new Integer(j));

        Integer[] list = (Integer[]) neighbors.toArray(new Integer[0]);

        //for all the possible edges
        for (int i=0;i<neighbors.size();i++)
            for (int k=0; k<neighbors.size(); k++)
            //check to see if there is an edge
              
            if (a.isNeighbor(list[i],list[k])){

                sum += v.get(list[i].intValue()).howSimilartoVector(v.get(list[k].intValue()));

            
            }
        if (neighbors.size()>1) return sum/(neighbors.size()*(neighbors.size()-1));    
            else return 0;
    }

    /**
     *  <code>plainClustering</code> method finds the clustering of j.
     *
     * @param v a <code>VectorTable</code> value
     * @param a an <code>Adjacency</code> value
     * @param j an <code>int</code> value
     * @return a <code>double</code> value
     */
    
    public static double plainClustering(Adjacency a, int j) {

        double sum=0.0;
        TreeSet neighbors = findOutgoingNodes(a, j);

        neighbors.add(new Integer(j));

        Integer[] list = (Integer[]) neighbors.toArray(new Integer[0]);

        //for all the possible edges
        for (int i=0;i<neighbors.size();i++)
            for (int k=0; k<neighbors.size(); k++)
            //check to see if there is an edge
		if (a.isNeighbor(list[i],list[k]))
		    sum ++;

        if (neighbors.size()>1) return sum/(neighbors.size()*(neighbors.size()-1));    
            else return 0;

    }

    /**
     * Returns the average clustering of all nodes
     *
     * @param v a <code>VectorTable</code> value
     * @param a an <code>Adjacency</code> value
     * @param agents a <code>Vector</code> value
     * @return a <code>double</code> value
     */
    public static double plainClustering(Adjacency a, Vector agents){
        double sum=0;

        for (int i=0; i<agents.size(); i++){
            sum += plainClustering(a, ((Integer)agents.get(i)).intValue());

        }
        return sum/agents.size();
            
    }

    /**
     * Returns the plainClustering for all agents in a double vector.
     *
     * @param v a <code>VectorTable</code> value
     * @param a an <code>Adjacency</code> value
     * @param agents a <code>Vector</code> value
     * @return a <code>DoubleVector</code> value 
     */

    public static DoubleVector plainClusteringIndividuals(Adjacency a, Vector agents){

	DoubleVector d = new DoubleVector();
	for (int i=0; i<agents.size(); i++) {
            d.add(plainClustering(a, ((Integer)agents.get(i)).intValue()));
        }
        return d;
    }
	
    /**
     * Average over all interest similarities for all nodes 
     *
     * @param v a <code>VectorTable</code> value
     * @param a an <code>Adjacency</code> value
     * @param agents a <code>Vector</code> value
     * @return a <code>double</code> value
     */
    public static double interestSimilarity(VectorTable v, Adjacency a, Vector agents){
        double sum=0;

        for (int i=0; i<v.size(); i++){
            sum += interestSimilarity(v, a, ((Integer)agents.get(i)).intValue());

        }
        return sum/v.size();
            
    }

    /**
     * Returns <code>interestSimilarity</code> for all agents in a
     * double vector.
     *
     * @param v a <code>VectorTable</code> value
     * @param a an <code>Adjacency</code> value
     * @param agents a <code>Vector</code> value
     * @return a <code>DoubleVector</code> value 
     */

    public static DoubleVector interestSimilarityIndividuals(VectorTable v, Adjacency a, Vector agents){

	DoubleVector d = new DoubleVector();
	for (int i=0; i<agents.size(); i++) {
            d.add(interestSimilarity(v, a, ((Integer)agents.get(i)).intValue()));
        }
        return d;
    }

    /**
     * Finds the direct quality metric
     *
     * @param edges a <code>Vector</code> value
     * @param agents a <code>Vector</code> value
     * @param interest a <code>VectorTable</code> value
     * @param expertise a <code>VectorTable</code> value
     * @return a <code>double</code> value
     */
    public static double nodeQuality(Vector edges, Vector agents, VectorTable interest, VectorTable expertise){
      //Traverse over all edges and take the average
        int i = 0;
        int j = 0; 
        double sum = 0.0;
        Enumeration en=edges.elements();
        while (en.hasMoreElements()){
            i = (((Integer)en.nextElement()).intValue());
            j = (((Integer)en.nextElement()).intValue());

            sum += interest.get(i).howSimilartoExpertise(expertise.get(j));
        }
        return sum/(edges.size()/2);
    }

    /**
     * Normalizes the quality based on how well that population could have done
     *
     * @param shortestpaths an <code>int[][]</code> value
     * @param agents a <code>Vector</code> value
     * @param interest a <code>VectorTable</code> value
     * @param expertise a <code>VectorTable</code> value
     * @return a <code>double</code> value
     */
    public static double quality_normalized_society(int[][] shortestpaths, Vector agents, VectorTable interest, VectorTable expertise){
        double sum = 0.0;
        double temp= 0.0;
        int k = 0;
        int l = 0;
        int size = agents.size();
        double matchsum = 0.0;//to calculate how well I could have done 
        for (int i= 0; i<size; i++)
            for (int j= 0; j<size; j++)
                if (i!=j){ 
                    k = ((Integer)(agents.get(i))).intValue();
                    l = ((Integer)(agents.get(j))).intValue();
                    temp = (interest.get(k).howSimilartoExpertise(expertise.get(l)));
                    matchsum +=temp;
                    sum += temp /shortestpaths[i][j];
                }
        return sum/matchsum;
    }


    /**
     * Find quality normalized based on perfect population
     *
     * @param shortestpaths an <code>int[][]</code> value
     * @param agents a <code>Vector</code> value
     * @param interest a <code>VectorTable</code> value
     * @param expertise a <code>VectorTable</code> value
     * @return a <code>double</code> value
     */
    public static double quality_normalized(int[][] shortestpaths, Vector agents, VectorTable interest, VectorTable expertise) {
       return quality(shortestpaths, agents, interest, expertise)/(agents.size()*(agents.size()-1));
                    
    }

    /**
     * Describe <code>quality</code> method here.
     *
     * @param shortestpaths an <code>int[][]</code> value
     * @param agents a <code>Vector</code> value
     * @param interest a <code>VectorTable</code> value
     * @param expertise a <code>VectorTable</code> value
     * @return a <code>double</code> value
     */
    public static double quality(int[][] shortestpaths, Vector agents, VectorTable interest, VectorTable expertise){
        double sum = 0.0;
        double temp= 0.0;
        int k = 0;
        int l = 0;
        int size = agents.size();
        //double matchsum = 0.0;//to calculate how well I could have done 
        for (int i= 0; i<size; i++) {
            k = ((Integer)(agents.get(i))).intValue();
            for (int j= 0; j<size; j++)
                if (i!=j){ 
                    l = ((Integer)(agents.get(j))).intValue();
                    temp = (interest.get(k).howSimilartoExpertise(expertise.get(l)));
                    sum += temp /shortestpaths[i][j];
                }
	}
        return sum;
    }

    /**
     * Puts all the node quality values into the vector v, and then
     * returns the best (number_of_neighbors*2) value
     *
     * @param shortestpaths an <code>int[][]</code> value
     * @param agents a <code>Vector</code> value
     * @param interest a <code>VectorTable</code> value
     * @param expertise a <code>VectorTable</code> value
     * @param a an <code>Adjacency</code> value
     * @return a <code>double</code> value 
     */
    public static double qualityMedian(int[][] shortestpaths, Vector agents, VectorTable interest, VectorTable expertise, Adjacency a){
        double sum = 0.0;

	double current = 0.0;
        int size = agents.size();
  
        for (int i= 0; i<size; i++){

	    //Add how well each agent is - measured with the best nth neighbor
	    current = individualQuality(shortestpaths, agents, interest, expertise, a, i);
	    //Don't average over experts (who have 0 quality)
	    if (current == 0 ) {
		size --;
		
	    } // end of if ()
	    sum += current;
        }

        return sum/size;
                    
    }
    
    public static double individualQuality(int[][] shortestpaths, Vector agents, VectorTable interest, VectorTable expertise, Adjacency a, int i){

        int l = 0;
        int k = 0;
        DoubleVector v = new DoubleVector();
        double temp= 0.0;

        int size = agents.size();
	//        v.clear();
        for (int j= 0; j<size; j++)
	    if (i!=j){ 
		//k is the id of the first agent (source) 
		k = ((Integer)(agents.get(i))).intValue();
		l = ((Integer)(agents.get(j))).intValue();
		//System.out.println("Interest " + k + "Expertise " + l);

                temp = (interest.get(k).howSimilartoExpertise(expertise.get(l)));
                v.add(temp /shortestpaths[i][j]);
	    }
	return v.best(size-(a.get(k).size()*2)-2);

    }
    public static void getNodes(int i, Adjacency a, TreeSet returnSet){
        TreeSet neighbors = findOutgoingNodes(a, i);
        Integer[] list = (Integer[]) neighbors.toArray(new Integer[0]);
        for (int k=0; k<neighbors.size(); k++){
            if (!returnSet.contains(list[k])){
                   returnSet.add(list[k]);
                   getNodes(list[k].intValue(), a, returnSet);
            }
        }
            
    }

    /**
     * Internal method. Returns true if all three elements are zero.
     *
     * @param question a <code>Vector</code> value
     * @return a <code>boolean</code> value
     */
    private static boolean allZeros(Vector question) {
	return ((((Double) question.get(0)).doubleValue()== 0) && 
		(((Double) question.get(1)).doubleValue()== 0) && 
		(((Double) question.get(2)).doubleValue()== 0));
    }

    /**
     * Finds the ratio of answers you already had (already in the
     * lattice) to all queries
     *
     * @param numberofAgents an <code>int</code> value
     * @param s a <code>String</code> value
     * @return a <code>Vector</code> value 
     */
    public static Vector numberOfExistingProviders(int numberOfIterations, int numberofAgents, String s){

	Vector number = new Vector();

	ScoreReader r = new ScoreReader();
	Vector v = r.parse(numberofAgents, s);
	Vector temp = new Vector();
	Vector question = new Vector();

	double oldProvider = 0;

	int numberOfQueries = ((Vector)v.get(0)).size();
	int chunk = (new Double(numberOfQueries/numberOfIterations)).intValue();

	for (int k = 0; k<numberOfIterations; k++) {

	    for (int i=0; i<numberofAgents; i++){

		temp = (Vector)(v.get(i));
		oldProvider = 0.0;
		for (int j=(k*chunk); j<((k+1)*chunk); j++){
		    question = (Vector)(temp.get(j));
  
		//Meaning used an existing provider
		if (allZeros(question)) 
		    oldProvider++;
		}
	    oldProvider = oldProvider/chunk;
	    number.add(new Double(oldProvider));
	    }
	}
	return number;
		       
    }

    /**
     * Describe <code>numberOfContactedAgents</code> method here.
     *
     * @param numberOfIterations an <code>int</code> value
     * @param numberofAgents an <code>int</code> value
     * @param s a <code>String</code> value
     * @return a <code>Vector</code> value
     */
    public static Vector numberOfContactedAgents(int numberOfIterations, int numberofAgents, String s){

	Vector number = new Vector();

	ScoreReader r = new ScoreReader();
	Vector v = r.parse(numberofAgents, s);
	Vector temp = new Vector();
	Vector question = new Vector();

	double sum = 0;

	int numberOfQueries = ((Vector)v.get(0)).size();
	int chunk = (new Double(numberOfQueries/numberOfIterations)).intValue();

	for (int k = 0; k<numberOfIterations; k++) {

	    for (int i=0; i<numberofAgents; i++){

		temp = (Vector)(v.get(i));
		sum = 0.0;
		for (int j=(k*chunk); j<((k+1)*chunk); j++){
		    question = (Vector)(temp.get(j));

		    sum += ((Double)question.get(0)).doubleValue() + 
			   ((Double)question.get(1)).doubleValue() +
			   ((Double)question.get(2)).doubleValue();

		}
		sum = sum/chunk; //Avg. number of contacted peers per query
		number.add(new Double(sum));
	    }
	}
	return number;
		       
    }

    /**
     * for each agent, find the ratio of correct answers to number of queries
     *
     * @param numberOfIterations an <code>int</code> value
     * @param numberofAgents an <code>int</code> value
     * @param s a <code>String</code> value
     * @return a <code>Vector</code> value
     */
    public static Vector gotGoodAnswer(int numberOfIterations, int numberofAgents, String s){
      //s is the dir name
	double sum = 0.0;
	double current = 0.0;
	double correct = 0.0;

	ScoreReader r = new ScoreReader();
	Vector v = r.parse(numberofAgents, s);
	Vector temp = new Vector();
	Vector question = new Vector();
       
	int numberOfQueries = ((Vector)v.get(0)).size();
	int chunk = (new Double(numberOfQueries/numberOfIterations)).intValue();

	Vector returnVector = new Vector();

	for (int k = 0; k<numberOfIterations; k++) {
	
	    for (int i=0; i<numberofAgents; i++){
		temp = (Vector)(v.get(i));
		correct = 0;
		for (int j=(k*chunk); j<((k+1)*chunk); j++){
		    
		    question = (Vector)(temp.get(j));
		  
		    if ((((Double) question.get(0)).doubleValue() > 0) || allZeros(question))
			 correct++;

		}
		correct = correct/chunk;
		returnVector.add(new Double(correct));
	    }
	    
	} // end of for ()


	return returnVector;
    }
    

    /**
     * Ratio of good answers you got from others to the number of queries
     *
     * @param numberOfIterations an <code>int</code> value
     * @param numberofAgents an <code>int</code> value
     * @param s a <code>String</code> value
     * @return a <code>Vector</code> value
     */
    public static Vector gotFromOthers(int numberOfIterations, int numberofAgents, String s){
      //s is the dir name
	double sum = 0.0;
	double current = 0.0;
	double correct = 0.0;

	ScoreReader r = new ScoreReader();
	Vector v = r.parse(numberofAgents, s);
	Vector temp = new Vector();
	Vector question = new Vector();
       
	int numberOfQueries = ((Vector)v.get(0)).size();
	int chunk = (new Double(numberOfQueries/numberOfIterations)).intValue();

	Vector returnVector = new Vector();

	for (int k = 0; k<numberOfIterations; k++) {
	
	    for (int i=0; i<numberofAgents; i++){
		temp = (Vector)(v.get(i));
		correct = 0;
		for (int j=(k*chunk); j<((k+1)*chunk); j++){
		    
		    question = (Vector)(temp.get(j));
		  
		    if (((Double) question.get(0)).doubleValue() > 0)
			 correct++;

		}
		correct = correct/chunk;
		returnVector.add(new Double(correct));
	    }
	    
	} // end of for ()


	return returnVector;
    }
    
  /** How many of the answers were correct */
    public static Vector accuracy(int numberofAgents, String s){
      //s is the dir name
	double sum = 0.0;
	double current = 0.0;
	double correct = 0.0;

	ScoreReader r = new ScoreReader();
	Vector v = r.parse(numberofAgents, s);
	Vector temp = new Vector();
	Vector question = new Vector();
       
	
	for (int i=0; i<numberofAgents; i++){
	    temp = (Vector)(v.get(i));
	    for (int j=0; j<temp.size(); j++){

		question = (Vector)(temp.get(j));
		//Can be changed. no matter how many good answers there are to a question, we count that as one.  
		if (((Double) question.get(0)).doubleValue() > 0) correct++;
		sum += ((Double)question.get(0)).doubleValue() + 
		      ((Double)question.get(1)).doubleValue() +
		      ((Double)question.get(2)).doubleValue();

            }

	}
	temp.clear();
	temp.add(new Double(correct));
	temp.add(new Double(sum));
	return temp;
    }


    /**
     * Returns the <code>reciprocity</code> coefficient.
     * Ratio of reciprocal realation to non-reciprocal relations. 
     *
     * @param numberOfAgents an <code>int</code> value
     * @param a an <code>Adjacency</code> value
     * @return a <code>double</code> value
     */
    public static double reciprocity(int numberOfAgents, Adjacency a){

	double reciprocal = 0.0;
	double total = 0.0;
	boolean oneway = false;
	boolean reverse = false;

	for (int i=0; i<numberOfAgents; i++) {

	    for (int j=0; j<numberOfAgents; j++) {
		oneway = a.isNeighbor(i, j);
		reverse = a.isNeighbor(j, i);

		if (oneway && reverse) {
		    //System.out.println("reciprocal " + i +" " + j);
		    
		    reciprocal += 2;
		    total += 2;
		} // end of if (
		else if (oneway || reverse ) {
		    total++;
		} // end of else ()
		
	    } // end of for ()
	} // end of for ()
	//System.out.println("reciprocity " + reciprocal + " total " + total);
	
	//Avoid division by zero. Total can't be zero, if reciprocal is not.
	//So no need to worry about reciprocal.
	if (total == 0 ) {
	    total = 1;
	    
	} // end of if ()
	
	return (reciprocal/total);

    }

        
  /** Not working */
    public static Vector findComponents(Adjacency a){
        Vector components = new Vector();
        boolean found = false;
        int j = 0;
        TreeSet newSet = new TreeSet();

        for (int i=0; i<a.size(); i++)
        {
            j=0;
            found = false;
            while (j<components.size() && (!found)){
                if (((TreeSet)components.get(j)).contains(new Integer(i)))
                    found = true;
                j++;
            }
            
            if (!found){ 

                getNodes(i, a, newSet);
                components.add(newSet.clone());
                newSet.clear();
            }
        }
        return components;
    }
    public static boolean isDisjoint(TreeSet s, TreeSet t){
        int size = 0;
        size = s.size();
        s.addAll(t);
        return ((size + t.size()) == s.size());  
    }

    public static Vector components(Vector comp){
        Vector v = (Vector)comp.clone();
        Vector t = new Vector();
        //for (int i=0; i<t.size(); i++)
        for (int j=0; j<v.size(); j++)
          for (int k=j; k<v.size(); k++)
             if (!isDisjoint((TreeSet)v.get(j), (TreeSet)v.get(k))){
                        ((TreeSet)v.get(j)).addAll((TreeSet)(((TreeSet)v.get(k)).clone()));
                        ((TreeSet)v.get(k)).clear();
                    }
        return v;
    }
}
