//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.dialogs;

import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.vectors.Answer;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * The Node is a class that is used by the ReferralChainDAG to encapsulate
 * the notion of an agent which has been referred by another agent and may
 * have referred other agents itself.
 */
public final class Node
{
	private int id = -1;
	private Answer answer = Constants.NULL_ANSWER;
	private Vector parents = null;
	private Vector children = null;
	private int depth = 0;
	
	//========================================================================			

	public Node( int id )
	{	
		this.id = id;
		this.parents = new Vector();
		this.children = new Vector();
		this.depth = 0;
	}
	
	//========================================================================	
	
	public void addChild( Node child )
	{
		children.addElement( child );
		child.addParent( this );
	}
	
	//========================================================================	
	
	private void addParent( Node parent )
	{
		parents.addElement( parent );
		
		// adding the parent changed the depth of this node
		calculateDepth();		
	}
	
	//========================================================================			
	/**
	 * Recalculates the current depth of this node.
	 * The depth of a node equals the depth of its least deep parent + 1.
	 * If the depth of this node changes, then the depth of its descendants
	 * may have changed, so this method must recursively be invoked on its 
	 * children when this happens.
	 * This method assumes that the depth of its parents has already been
	 * correctly calculated.
	 */
	private void calculateDepth()
	{
		int originalDepth = this.depth;
		int newDepth = 0;
		int depthOfParent = 0;
		int minimumDepthAmongParents = Integer.MAX_VALUE;	
		
		// determine what the new depth of this node is
		if ( parents.size() == 0 )
		{
			newDepth = 0;
		}
		else
		{	
			Enumeration enumParents = parents.elements();

			while( enumParents.hasMoreElements() )
			{
				Node parent = (Node)enumParents.nextElement();

				depthOfParent = parent.getDepth();

				if ( depthOfParent < minimumDepthAmongParents )
				{
					minimumDepthAmongParents = depthOfParent;
				}
			}	
			
			newDepth = minimumDepthAmongParents + 1;
		}
		
		// if the new depth of this node is different from the oringal depth,
		// then we need save the new depth.
		// Furthermore, the depth of this nodes descendants may have changed.
		// So we recursively call this function on this node's children.
		if ( newDepth != originalDepth )
		{
			// save the new depth of this node
			this.depth = newDepth;
			
			// update the depth of this node's children
			Enumeration enumChildren = children.elements();

			while( enumChildren.hasMoreElements() )
			{
				Node child = (Node)enumChildren.nextElement();
				child.calculateDepth();
			}	
		}
	}		
	
	//========================================================================	
	
	public Vector getChildren()
	{
		return children;
	}
	
	//========================================================================	
	
	public int getDepth()
	{
		return depth;
	}	
	
	//========================================================================	
	
	public Vector getParents()
	{
		return parents;
	}	

	//========================================================================	
	
	public Node getDescendant( int id )
	{
		//System.out.println( "begin getDescendant(" + id + ")" );
		//System.out.println( "this id = (" + this.id + ")" );		
		
		if ( this.id == id )
		{
			//System.out.println( "end getDescendant(" + id + ") A" );					
			return this;
		}
		else
		{
			Node nodeTarget = null;

			Enumeration enumChildren = children.elements();

			while( enumChildren.hasMoreElements() )
			{
				Node child = (Node)enumChildren.nextElement();

				nodeTarget = child.getDescendant( id );

				if ( nodeTarget != null )
				{
					//System.out.println( "end getDescendant(" + id + ") B" );					
					return nodeTarget;
				}
			}		
			//System.out.println( "end getDescendant(" + id + ") C" );				
			return null;
		}	
	}	
	
	//========================================================================	
	
	public Vector getDescendants( )
	{
		//System.out.println( "Node.getDescendants(" + this.id + ")" );
		
		Vector descendants = null;
		
		Enumeration enumChildren = children.elements();

		while( enumChildren.hasMoreElements() )
		{
			Node child = (Node)enumChildren.nextElement();

			Vector d = child.getDescendants();

			if ( descendants == null )
			{
				descendants = d;
			}
			else
			{
				descendants.addAll( d );
			}
		}	
		
		if ( descendants == null )
		{
			descendants = new Vector();
		}
		
		descendants.add( new Integer(this.id) );
		
		//System.out.println( "returning: " + descendants );		
		//System.out.println( "Node.getDescendants(" + this.id + ")" );
		
		return descendants;
	}		
	
	//========================================================================	
	
	public boolean isDescendantOf( int id )
	{
		boolean flgDescendant = false;
		
		if ( this.id == id )
		{
			return true;
		}
		else
		{
			Enumeration enumParents = parents.elements();

			while( enumParents.hasMoreElements() )
			{
				Node parent = (Node)enumParents.nextElement();

				flgDescendant = parent.isDescendantOf( id );

				if ( flgDescendant )
				{
					return true;
				}
			}		
			
			return false;
		}
	}
	
	//========================================================================			
	
	public int getID()
	{
		return id;
	}	
	
	//========================================================================			
	
	public Answer getAnswer()
	{
		return answer;
	}
	
	//========================================================================			
	
	public void setAnswer( Answer answer )
	{
		this.answer = answer;
	}
	
	//========================================================================			
	
	public void recordEdges( int[][] edges )
	{
		Enumeration enumChildren = children.elements();
	
		while( enumChildren.hasMoreElements() )
		{
			Node child = (Node)enumChildren.nextElement();
			edges[this.id][child.getID()] = 1;
		}
		
		enumChildren = children.elements();

		while( enumChildren.hasMoreElements() )
		{
			Node child = (Node)enumChildren.nextElement();
			child.recordEdges( edges );
		}
	}
	
	//========================================================================			
	
	public String toString()
	{
		return "Node:" + String.valueOf( id );
	}
	
	//========================================================================			
}