//==========================================================================

//	This source code is part of the Agent Simulation Test Bed system which

//	was developed for simulating interaction among agents who are attempting 

//	to acquire information from each other and learn about each other. 

//

//	The system was created at North Carolina State Univeristy, 

//	Department of Computer Science between September and December 2001. 

//	It was implemented by Michael J. Suralik based on the specifications 

//	developed by Pinar Yolum and Bin Yu, who were conducting research under 

//   the guidance of Dr. Munindar Singh. 

//==========================================================================

package edu.ncsu.simulators.dialogs;



import edu.ncsu.simulators.Simulator;

import edu.ncsu.simulators.Constants;

import edu.ncsu.simulators.util.ErrorHandler;

import edu.ncsu.simulators.vectors.Answer;

import edu.ncsu.simulators.vectors.Query;



import java.util.Enumeration;

import java.util.Hashtable;

import java.util.Vector;



/**

 * The ReferralChainDAG is an implementation of the ReferralChain interface

 * which utilizes a directed acyclic graph.  The most interesting feature of

 * this implementation is that it will ignore the addition of referrals from 

 * a referring agent to another agent which was an ancestor of the referring 

 * agent (i.e. it will not allow a cycle).

 */

public class ReferralChainDAG implements ReferralChain

{	

	private Node root = null;	

	private Query query = null;

	private Hashtable hashAnswers = null;
        
        private Answer bestAnswer = null;
                
	

	//========================================================================			

	

	public ReferralChainDAG( Query query, int id )

	{

		//System.out.println( "Adding RCDAG with root = " + id );

		this.root = new Node(id);		

		this.query = query;

		this.hashAnswers = new Hashtable();

	}

	

	//========================================================================			

	/**

	 * Since this method may change the length of a referral, we must 

	 * synchronize it with the getChainLength method.

	 */

	synchronized public void addReferral( int referringAgentID, int referredAgentID )

	{

		//System.out.println( "ReferralChainDAG.addReferral(" + referringAgentID + "," + referredAgentID + ")" );

		

		Node nodeReferring = root.getDescendant( referringAgentID );

		

		if ( nodeReferring != null )

		{

			// do NOT add the referral if the referring agent has referred

			// someone to one of its ancestors, because that would create a cycle

			if ( !nodeReferring.isDescendantOf(referredAgentID) )

			{		

				Node nodeReferred = root.getDescendant( referredAgentID );



				if ( nodeReferred == null )

				{

					nodeReferred = new Node( referredAgentID );

				}



				nodeReferring.addChild( nodeReferred );

			}

		}

		else

		{

			// there is a programming error if the referring agent is not in the chain

			ErrorHandler.exit( "ERROR: Attempted to add a referral to " + referredAgentID + " from a referring agent " + referringAgentID + " that is not in the chain!" );

		}		

		

		// print the updated graph

		//System.out.println( this );

	}

	

	//========================================================================			

	

	public Enumeration getAncestorsByGeneration( int agentID )

	{

		//System.out.println( "getAncestorsByGeneration(" + agentID + ")" );

		

		Vector generations = new Vector();

		int dummyCounter = 0;

		

		Node progenyNode = root.getDescendant( agentID );

		Enumeration enumParentNodes = progenyNode.getParents().elements();

		

		while ( enumParentNodes.hasMoreElements() )

		{

			Vector parentIDs = new Vector();			

			Vector grandparentNodes = new Vector();



			while ( enumParentNodes.hasMoreElements() )

			{

				Node parent = (Node)enumParentNodes.nextElement();

				

				// never include the root id as an ancestor!

				// we don't want to manipulate the sociability of questioning agent

				if ( parent.getID() != root.getID() )

				{

					parentIDs.addElement( new Integer(parent.getID()) );

					grandparentNodes.addAll( parent.getParents() );

				}

			}

			

			generations.addElement( parentIDs.elements() );

		

			enumParentNodes = grandparentNodes.elements();		

		}

		

		return generations.elements();

	}

	

	//========================================================================			

	/**

	 * Since this method depends on the tree structure not changing, we must 

	 * synchronize it with the addReferral method

	 */	

	synchronized public Hashtable getAnswers()

	{

		Hashtable hashAnswers = new Hashtable();

		

		Vector queue = (Vector)root.getChildren().clone();

		

		while ( queue.size() > 0 )

		{

			Node node = (Node)queue.elementAt( 0 );		

			queue.removeElementAt( 0 );		



			hashAnswers.put( new Integer(node.getID()), node.getAnswer() );

			

			if ( node.getChildren().size() != 0 )

			{

				queue.addAll( node.getChildren() );

			}

		}

		

		return hashAnswers;

	}

	

	//========================================================================			

	/**

	 * Since the addReferral method may change the length of a referral, we must 

	 * synchronize this method with the addReferral method.

	 */	

	synchronized public int getChainLength( int agentID )

	{

		//System.out.println( "START getChainLength(" + agentID + ")" );

		//System.out.println( "root id = (" + root.getID() + ")" );

		

		Node nodeAgent = root.getDescendant( agentID );

		

		if ( nodeAgent == null )

		{

			ErrorHandler.exit( "ERROR: Attempted to get the chain length for an agent(" + agentID + ") that is not in the chain!" );

		}

		

		//System.out.println( "END getChainLength(" + agentID + ")" );		

		return nodeAgent.getDepth();

	}		

	

	//========================================================================		

	

	public Enumeration getDescendants( int agentID )

	{

		//System.out.println( "BEGIN ReferralChainDAG.getDescendants(" + agentID + ")" );

		

		Vector vecDescendants = new Vector();

		

		Node nodeAncestor = root.getDescendant( agentID );

		

		if ( nodeAncestor != null )

		{

			vecDescendants = nodeAncestor.getDescendants();

		}

		

		//System.out.println( "returning: " + vecDescendants );

		//System.out.println( "END ReferralChainDAG.getDescendants(" + agentID + ")" );				

		

		return vecDescendants.elements();		

	}

	

	//========================================================================			

	

	public Query getQuery()

	{		

		return this.query;

	}			

	

	//========================================================================	

	

	public void recordAnswer( int answeringAgentID, Answer answer )

	{

		//System.out.println( "recordAnswer(" + answeringAgentID + ")" );

		

		Node nodeAnsweringAgent = root.getDescendant( answeringAgentID );

		

		if ( nodeAnsweringAgent == null )

		{

			ErrorHandler.exit( "ERROR: Attempted to record an answer for an agent that is not in the chain!" );

		}

		

		nodeAnsweringAgent.setAnswer( answer );

		

		//hashAnswers.put( new Integer(answeringAgentID), answer );

	}	

	

	//========================================================================	

	

	public String toString()

	{

		StringBuffer sbuf = new StringBuffer();

		int[][] edges = new int[Simulator.getAgentCount()][Simulator.getAgentCount()];

		

		// populate the edges graph

		root.recordEdges( edges );

				

		// print the edges graph to a string buffer



		sbuf.append( "\n" );		

		sbuf.append( "   REFERRAL GRAPH \n" );		

		sbuf.append( "\n" );		

		sbuf.append( "   " );

			

		for ( int v=0; v<Simulator.getAgentCount(); v++ )

		{			

			sbuf.append( v + " " );

		}

		

		sbuf.append( "\n" );

		sbuf.append( "   " );		

		

		for ( int v=0; v<Simulator.getAgentCount(); v++ )

		{			

			sbuf.append( "--" );

		}		

		

		sbuf.append( "\n" );		

		

		for ( int u=0; u<Simulator.getAgentCount(); u++ )

		{

			sbuf.append( u + "| " );

			

			for ( int v=0; v<Simulator.getAgentCount(); v++ )

			{			

				sbuf.append( edges[u][v] + " " );

			}

			

			sbuf.append( "\n" );

		}

		

		return sbuf.toString();

	}	
        
        public void setBestAnswer(Answer bestAnswer) 
        {
            this.bestAnswer = bestAnswer;
        }
        
        public Answer getBestAnswer()
        {
            return bestAnswer;
        }
            

	

	//========================================================================		

}