//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.dialogs;

import edu.ncsu.simulators.Constants;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import javax.jms.MapMessage;

/**
 * The UnprocessedReferrals object is used to store referrals which have not
 * been processed yet, but need to be maintained for some reason.  
 * <p>
 * Currently, it is used to maintain referrals which would have extended a 
 * referral chain past a maximum allowed length.  There is a reason to maintain 
 * the referral, rather than throw it away, because we may discover later that 
 * a different referral had the effect of shortening the length of the chain, 
 * and that this referral is now elegible to be processed.
 */
public final class UnprocessedReferrals
{
	private Hashtable hashReferralChains = null;
	
	//========================================================================			

	public UnprocessedReferrals()
	{	
		this.hashReferralChains = new Hashtable();
	}
	
	//========================================================================	
	
	public void addReferral( long _chainID, int _referringAgentID, Object referral )
	{
		// print debugging info
		try
		{
			MapMessage referralMessage = (MapMessage)referral;
			int referredAgentID = referralMessage.getInt( Constants.MESSAGE_FIELD_REFERRED_AGENT_ID );	
			//System.out.println( "UnprocessedReferrals.addReferral( chain=" + _chainID + ", referringAgentID=" + _referringAgentID + ", referredAgentID=" + referredAgentID ); 
		}
		catch ( Exception e )
		{
			e.printStackTrace();
		}
		
		Long chainID = new Long( _chainID );
		Integer referringAgentID = new Integer( _referringAgentID );
		
		Hashtable hashReferringAgents = (Hashtable)hashReferralChains.get( chainID ); 
		
		if ( hashReferringAgents == null )
		{
			hashReferringAgents = new Hashtable();
			hashReferralChains.put( chainID, hashReferringAgents );
		}
		
		Vector vecReferrals = (Vector)hashReferringAgents.get( referringAgentID );
		
		if ( vecReferrals == null )
		{
			vecReferrals = new Vector();
			hashReferringAgents.put( referringAgentID, vecReferrals );						
		}
		
		vecReferrals.addElement( referral );	
	}
	
	//========================================================================	
	
	public Enumeration removeReferrals( long _chainID, int _referringAgentID )
	{
		Long chainID = new Long( _chainID );
		Integer referringAgentID = new Integer( _referringAgentID );
		
		Enumeration enum = null;
		
		Hashtable hashReferringAgents = (Hashtable)hashReferralChains.get( chainID ); 
		
		if ( hashReferringAgents != null )
		{
			Vector vecReferrals = (Vector)hashReferringAgents.remove( referringAgentID );
			
			if ( vecReferrals != null )
			{
				enum = vecReferrals.elements();
			}
		}
		
		return enum;
	}
	
	//========================================================================	
	
	public void removeReferralChain( long _chainID )
	{
		Long chainID = new Long( _chainID );
		
		// remove the chain
		Hashtable hashReferringAgents = (Hashtable)hashReferralChains.remove( chainID );

		Enumeration enumVectorsReferredAgents = hashReferringAgents.elements();

		while ( enumVectorsReferredAgents.hasMoreElements() )
		{
			Vector vecReferrals = (Vector)enumVectorsReferredAgents.nextElement();
			vecReferrals.clear();
		}
		
		hashReferringAgents.clear();
	}
	
	//========================================================================	
	
	public String toString( long _chainID )
	{
		Long chainID = new Long( _chainID );
		StringBuffer sbuf = new StringBuffer();
		
		Hashtable hashReferringAgents = (Hashtable)hashReferralChains.get( chainID ); 
		
		sbuf.append( "\n" );		
		sbuf.append( "UNPROCESSED REFERRALS \n" );		
		sbuf.append( "\n" );		
		sbuf.append( "chain: " + chainID + "\n" );
		
		if ( hashReferringAgents != null )
		{
			Enumeration enumReferringAgents = hashReferringAgents.keys();			
			Enumeration enumVectorsOfReferrals = hashReferringAgents.elements();
			
			while ( enumVectorsOfReferrals.hasMoreElements() )
			{
				Vector vecReferrals = (Vector)enumVectorsOfReferrals.nextElement();
				Integer referringAgentID = (Integer)enumReferringAgents.nextElement();
				
				sbuf.append( "agent(" + referringAgentID + ")" );				
			
				if ( vecReferrals != null )
				{
					Enumeration enumReferralMessages = vecReferrals.elements();
				
					sbuf.append( " referred to agents: " ); 											
					
					while ( enumReferralMessages.hasMoreElements() )
					{
						try
						{
							MapMessage referralMessage = (MapMessage)enumReferralMessages.nextElement();
							int referredAgentID = referralMessage.getInt( Constants.MESSAGE_FIELD_REFERRED_AGENT_ID );
							sbuf.append( referredAgentID + " " ); 		
						}
						catch( Exception e )
						{
							e.printStackTrace();
						}
					}
					
					sbuf.append( "\n" ); 						
				}
			}
		}
			
		return sbuf.toString();
	}
}