//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.io;

import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.util.ErrorHandler;

import net.n3.nanoxml.*;

import java.util.Enumeration;
import java.util.Properties;

/**
 * The Dimension class is a utility class used by the SpecificationReader to temporarily store the data from a specification generation file in a format that is easy to work with.
 * <p>
 * Specification Readers load the data from the generation specification file
 * and parse it into Group objects and Dimension objects.  The 
 * SpecificationReader then uses the Dimension objects to create the agents and
 * other objects necessary for the Simulator.  Once the agents have been 
 * created, Dimension objects are no longer needed.
 * <p>
 * A Dimension object represents a value of or a range of values for a 
 * particular dimension of both an Expertise vector and an Interests vector.
 */
public class Dimension
{
	// The id number for this dimension
	private int id = -1;
	
	// Denotes whether a precise value or a range of values has been 
	// specified for this dimension of the Expertise vector
	private boolean flgRangeOfExpertise = true;

	// Denotes the precise expertise value for a particular dimension
	// This is used if flgRangeOfExpertise = false.
	public double expertise_value = -1;
	
	// Denotes a range of expertise for a particular dimension
	// These are used if flgRangeOfExpertise = true.
	public double expertise_range_low = -1;
	public double expertise_range_high = -1;

	// Denotes whether a precise value or a range of values has been 
	// specified for this dimension of the Interests vector
	private boolean flgRangeOfInterests = true;
	
	// Denotes the precise interests value for a particular dimension
	// This is used if flgRangeOfInterests = false.
	public double interests_value = -1;	

	// Denotes a range of interests for a particular dimension
	// These are used if flgRangeOfInterests = true.	
	public double interests_range_low = -1;
	public double interests_range_high = -1;		
	
	//========================================================================		
	/**
	 * New dimension objects should only be created from the parse method.
	 */
	private Dimension()
	{
	}
	
	//========================================================================	
	
	public int getID() { return id; }	
	
	/**
	 * Returns the lowest allowed value for the Expertise vector on this dimension.
	 */
	public double getExpertiseRangeLow() { return expertise_range_low; }	

	/**
	 * Returns the highest allowed value for the Expertise vector on this dimension.
	 */	
	public double getExpertiseRangeHigh() { return expertise_range_high; }	

	/**
	 * Returns the precise value of the Expertise vector on this dimension.
	 */
	public double getExpertiseValue() { return expertise_value; }		
	
	/**
	 * Returns the lowest allowed value for the Interests vector on this dimension.
	 */
	public double getInterestsRangeLow() { return interests_range_low; }	
	
	/**
	 * Returns the highest allowed value for the Interests vector on this dimension.
	 */
	public double getInterestsRangeHigh() { return interests_range_high; }	
	
	/**
	 * Returns the precise value of the Interests vector on this dimension.
	 */	
	public double getInterestsValue() { return interests_value; }	
	
	//========================================================================		
	/**
	 * Set the precise value of the Expertise vector for this dimension.
	 */ 
	public void setExpertise( double value )
	{
		flgRangeOfExpertise = false;
		expertise_value = value;
	}		

	//========================================================================	
	/**
	 * Set the range of values of the Expertise vector for this dimension.
	 */ 	
	public void setExpertiseRange( double low, double high )
	{
		flgRangeOfExpertise = true;
		expertise_range_low = low;
		expertise_range_high = high;
	}
	
	//========================================================================		
	/**
	 * Set the precise value of the Interests vector for this dimension.
	 */ 	
	public void setInterests( double value )
	{
		flgRangeOfInterests = false;
		interests_value = value;
	}
	
	//========================================================================		
	/**
	 * Set the range of values of the Interests vector for this dimension.
	 */ 	
	public void setInterestsRange( double low, double high )
	{
		flgRangeOfInterests = true;
		interests_range_low = low;
		interests_range_high = high;
	}
	
	//========================================================================		
	/**
	 * Returns true if the expertise value is specified as a range for this
	 * dimension of the Expertise vector; false if it is specified as a precise
	 * value.
	 */
	public boolean hasRangeOfExpertise()
	{
		return flgRangeOfExpertise;
	}
	
	//========================================================================	
	/**
	 * Returns true if the interests value is specified as a range for this
	 * dimension of the Interests vector; false if it is specified as a precise
	 * value.
	 */	
	public boolean hasRangeOfInterests()
	{
		return flgRangeOfInterests;
	}
	
	//========================================================================	
	/**
	 * Parse an XML element into the Dimension object which it represents.
	 * This assumes that it has already been verified that the XML element 
	 * is a valid Dimension object.
	 *
	 * @param xmlDimension an XML element which represents a dimension from a 
	 * specification generation file.
	 *
	 * @return the Dimension object represented by the xmlDimension element.
	 */	
	public static Dimension parse( IXMLElement xmlDimension )
	{
		Dimension newDimension = new Dimension();
	
		try
		{
			// Parse all the attributes of the dimension
			Properties attributesOfDimension = xmlDimension.getAttributes();					
			
			// set the ID of the dimension
			newDimension.id = Integer.parseInt( attributesOfDimension.getProperty(Constants.XML_ATTRIBUTE_ID) );

			// Parse all the child elements of the dimension
			Enumeration enumChildrenOfDimension = xmlDimension.getChildren().elements();

			while( enumChildrenOfDimension.hasMoreElements() )
			{
				IXMLElement xmlChildOfDimension = (IXMLElement)enumChildrenOfDimension.nextElement();

				// parse Expertise
				if ( xmlChildOfDimension.getName().equalsIgnoreCase(Constants.XML_TAG_EXPERTISE) )
				{		
					// Parse all the attributes
					Properties attributesOfExpertise = xmlChildOfDimension.getAttributes();

					String strExpertiseValue = attributesOfExpertise.getProperty( Constants.XML_ATTRIBUTE_VALUE );
					String strExpertiseRangeLow = attributesOfExpertise.getProperty( Constants.XML_ATTRIBUTE_LOW );
					String strExpertiseRangeHigh = attributesOfExpertise.getProperty( Constants.XML_ATTRIBUTE_HIGH );					

					if ( (strExpertiseValue == null) && (strExpertiseRangeLow != null) && (strExpertiseRangeHigh != null) ) 
					{
						// set the range of the expertise for this dimension
						newDimension.setExpertiseRange( Double.parseDouble(strExpertiseRangeLow), Double.parseDouble(strExpertiseRangeHigh) );
					}
					else if ( (strExpertiseValue != null) && (strExpertiseRangeLow == null) && (strExpertiseRangeHigh == null) ) 
					{
						// set the value of the expertise for this dimension
						newDimension.setExpertise( Double.parseDouble(strExpertiseValue) );				
					}
					else
					{
						ErrorHandler.exit( "ERROR parsing " + xmlChildOfDimension );		
					}					
				}

				// parse Interests
				else if ( xmlChildOfDimension.getName().equalsIgnoreCase(Constants.XML_TAG_INTERESTS) )
				{		
					// Parse all the attributes
					Properties attributesOfInterests = xmlChildOfDimension.getAttributes();

					String strInterestsValue = attributesOfInterests.getProperty( Constants.XML_ATTRIBUTE_VALUE );
					String strInterestsRangeLow = attributesOfInterests.getProperty( Constants.XML_ATTRIBUTE_LOW );
					String strInterestsRangeHigh = attributesOfInterests.getProperty( Constants.XML_ATTRIBUTE_HIGH );					

					if ( (strInterestsValue == null) && (strInterestsRangeLow != null) && (strInterestsRangeHigh != null) ) 
					{
						// set the range of the interest for this dimension
						newDimension.setInterestsRange( Double.parseDouble(strInterestsRangeLow), Double.parseDouble(strInterestsRangeHigh) );						
					}
					else if ( (strInterestsValue != null) && (strInterestsRangeLow == null) && (strInterestsRangeHigh == null) ) 
					{
						// set the value of the interest for this dimension
						newDimension.setInterests( Double.parseDouble(strInterestsValue) );	
					}
					else
					{
						ErrorHandler.exit( "ERROR parsing " + xmlChildOfDimension );
					}					
				}			
			}
		}
		catch( Exception e )
		{
			e.printStackTrace();
		}
		
		return newDimension;
	}			
}
