//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.io;

import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.util.ErrorHandler;

import net.n3.nanoxml.*;

import java.lang.Comparable;
import java.util.Enumeration;
import java.util.Properties;

/**
 * The Group class is a utility class used by the SpecificationReader to temporarily store the data from a specification generation file in a format that is easy to work with.
 * <p>
 * Specification Readers load the data from the generation specification file
 * and parse it into Group objects.  The SpecificationReader then uses the
 * Group objects to create the agents and other objects necessary for the 
 * Simulator.  Once the agents have been created, group objects are no longer
 * needed.
 * <p>
 * A Group object represents a common specification for creating a group of 
 * agents, with similar characteristics (e.g. the same policy implementations).
 * Although all the agents from the same group may not be identical because 
 * their interests may be randomly generated within a certain, they will still 
 * have the same characteristics in the sense that their interests will fall
 * within a common range.
 */
public class Group
{
	// These default policy implementations will apply to all groups which
	// do not override them by specifying a particular policy implementation
	// of their own.
	private static Class classDefaultAskingPolicy = null;
	private static Class classDefaultAnsweringPolicy = null;
	private static Class classDefaultReferringPolicy = null;
	private static Class classDefaultLearningPolicy = null;
	private static Class classDefaultChangingNeighborsPolicy = null;
        private static Class classDefaultCachingPolicy = null;
        private static Class classDefaultQueryGenerationPolicy = null;

	// The policy implmentations for this Group instance.
	// For those policies with null implementations, the defaults will be used.
	private Class classAskingPolicy = null;
	private Class classAnsweringPolicy = null;
	private Class classReferringPolicy = null;
	private Class classLearningPolicy = null;
	private Class classChangingNeighborsPolicy= null;
        private Class classCachingPolicy = null;
        private Class classQueryGenerationPolicy = null;
	
	// This default neighbor configuration will apply to all groups which
	// do not override them by specifying a particular neighbor configuration
	// of their own.
	private static double defaultNeighborPercentage = -1;
	private static double defaultClustering = -1;

	// The neighbor configuration for this Group instance.
	// If none is specified (and the values remain -1) the default 
	// neighbor configuration will be used.
	private double neighborPercentage = -1;
	private double clustering = -1;	

	// This boolean indicates whether the size of the group has been specified
	// in absolute terms (number of individual agents) or in relative terms
	// (percentage of the agent population).
	private boolean flgProportionalToPopulation = true;
	private int individuals = -1;
	private double percentageOfPopulation = -1;
	
	// An array of specificications for each dimension
	private Dimension[] dimensions = null;
	
	//========================================================================		
	/**
	 * New groups should only be created by the Group.parse() method.
	 */
	private Group() 
	{
	}
	
	//========================================================================		
	/**
	 * Returns the object which contains the specification for generating
	 * the expertise and interests values for the dimension indicated by the id
	 */
	public Dimension getDimension( int id ){ return dimensions[ id ]; }	

	//========================================================================		
	/**
	 * Returns the absolute number of agents that belong to this group
	 */
	public int getIndividuals() { return individuals; }

	//========================================================================			
	/**
	 * Returns the percentage of agents in the population who belong to this
	 * group (the range of this value is a decimal between 0 and 100).
	 */
	public double getPercentageOfPopulation() { return percentageOfPopulation; }

	//========================================================================	
	//	METHODS FOR GETTING POLICY IMPLEMENTATIONS
	//========================================================================		
	
	/**
	 * Returns the Class that implements the AskingPolicy which agents in this
	 * group should use
	 */
	public Class getAskingPolicyClass()
	{ 
		Class implementation = this.classAskingPolicy;
		if ( implementation == null )
		{
			implementation = classDefaultAskingPolicy;
		}
		return implementation;
	}

	/**
	 * Returns the Class that implements the AskingPolicy which agents in this
	 * group should use
	 */	
	public Class getAnsweringPolicyClass()
	{ 
		Class implementation = this.classAnsweringPolicy;
		if ( implementation == null )
		{
			implementation = classDefaultAnsweringPolicy;
		}
		return implementation;
	}
	
	/**
	 * Returns the Class that implements the ReferringPolicy which agents in this
	 * group should use
	 */	
	public Class getReferringPolicyClass()
	{ 
		Class implementation = this.classReferringPolicy;
		if ( implementation == null )
		{
			implementation = classDefaultReferringPolicy;
		}
		return implementation;
	}
	
	/**
	 * Returns the Class that implements the LearningPolicy which agents in this
	 * group should use
	 */	
	public Class getLearningPolicyClass()
	{ 
		Class implementation = this.classLearningPolicy;
		if ( implementation == null )
		{
			implementation = classDefaultLearningPolicy;
		}
		return implementation;
	}
	
	/**
	 * Returns the Class that implements the ChangingNeighborsPolicy which 
	 * agents in this group should use
	 */	
	public Class getChangingNeighborsPolicyClass()
	{ 
		Class implementation = this.classChangingNeighborsPolicy;
		if ( implementation == null )
		{
			implementation = classDefaultChangingNeighborsPolicy;
		}
		return implementation;
	}
        
        /**
	 * Returns the Class that implements the CachingPolicy which 
	 * agents in this group should use
	 */	
	public Class getCachingPolicyClass()
	{ 
		Class implementation = this.classCachingPolicy;
		if ( implementation == null )
		{
			implementation = classDefaultCachingPolicy;
		}
		return implementation;
	}

        
        /**
	 * Returns the Class that implements the QueryGenerationPolicy which 
	 * agents in this group should use
	 */	
	public Class getQueryGenerationPolicyClass()
	{ 
		Class implementation = this.classQueryGenerationPolicy;
		if ( implementation == null )
		{
			implementation = classDefaultQueryGenerationPolicy;
		}
		return implementation;
	}

	//========================================================================			
	/**
	 * Returns the percentage of agents in the population that agents in this
	 * group should be neighbors with (the range of this value is a decimal 
	 * between 0 and 1).
	 */	
	public double getNeighborPercentage()
	{ 
		double np = this.neighborPercentage;
		if ( np == -1 )
		{
			np = defaultNeighborPercentage;
		}
		return np;
	}	

	//========================================================================				
	/**
	 * Returns the percentage of neighbors which are selected from the
	 * agent population because of their proximity to the agent in queustion
	 * via the ring, as opposed to being chosen randomly (the range of this 
	 * value is a decimal between 0 and 1).
	 */
	public double getClustering()
	{ 
		double c = this.clustering;
		if ( c == -1 )
		{
			c = defaultClustering;
		}
		return c;
	}	
	
	//========================================================================	
	//	METHODS FOR SETTING *DEFAULT* POLICY IMPLEMENTATIONS
	//========================================================================		

	/**
	 * Sets the Class which will serve as the default implementation of an 
	 * AskingPolicy.  This implementation will apply to all groups which 
	 * do not override it by specifying a particular implementation of their own.	
	 */
	public static void setDefaultAskingPolicyClass( Class classPolicyImplementation ){ classDefaultAskingPolicy = classPolicyImplementation; }

	/**
	 * Sets the Class which will serve as the default implementation of an 
	 * AnsweringPolicy.  This implementation will apply to all groups which 
	 * do not override it by specifying a particular implementation of their own.	
	 */	
	public static void setDefaultAnsweringPolicyClass( Class classPolicyImplementation ){ classDefaultAnsweringPolicy = classPolicyImplementation; }

	/**
	 * Sets the Class which will serve as the default implementation of a 
	 * ReferringPolicy.  This implementation will apply to all groups which 
	 * do not override it by specifying a particular implementation of their own.	
	 */	
	public static void setDefaultReferringPolicyClass( Class classPolicyImplementation ){ classDefaultReferringPolicy = classPolicyImplementation; }

	/**
	 * Sets the Class which will serve as the default implementation of a
	 * LearningPolicy.  This implementation will apply to all groups which 
	 * do not override it by specifying a particular implementation of their own.	
	 */	
	public static void setDefaultLearningPolicyClass( Class classPolicyImplementation ){ classDefaultLearningPolicy = classPolicyImplementation; }
	
	/**
	 * Sets the Class which will serve as the default implementation of a 
	 * ChangingNeighborsPolicy.  This implementation will apply to all groups which 
	 * do not override it by specifying a particular implementation of their own.	
	 */
	public static void setDefaultChangingNeighborsPolicyClass( Class classPolicyImplementation ){ classDefaultChangingNeighborsPolicy = classPolicyImplementation; }
        
        /**
	 * Sets the Class which will serve as the default implementation of a 
	 * CachingPolicy.  This implementation will apply to all groups which 
	 * do not override it by specifying a particular implementation of their own.	
	 */
        public static void setDefaultCachingPolicyClass (Class classPolicyImplementation){ classDefaultCachingPolicy = classPolicyImplementation; }

        public static void setDefaultQueryGenerationPolicyClass (Class classPolicyImplementation){ classDefaultQueryGenerationPolicy = classPolicyImplementation; }
        
        
        

	//========================================================================		
	// Methods for setting the default neighbor configuration for all groups
		
	public static void setDefaultNeighborPercentage( double percentage ){ defaultNeighborPercentage = percentage; }
	public static void setDefaultClustering( double clustering ){ defaultClustering = clustering; }
	
	//========================================================================			
	/**
	 * Returns true if the group size is specified in relative terms (i.e. a 
	 * percentage of the agent population); false if the group size is 
	 * specified in absolute terms (i.e. a precise number of individual agents)
	 */
	public boolean isProportionalToPopulation()
	{
		return flgProportionalToPopulation;
	}	
	
	//========================================================================		
	/**
	 * Parse an XML element into the Group object which it represents.
	 * This assumes that it has already been verified that the XML element 
	 * is a valid Group object.
	 *
	 * @param xmlGroup an XML element which represents a group from a 
	 * specification generation file.
	 * @param numDimensions the number of dimensions in the universe of 
	 * discourse which applies ot this group of agents (currently, this should 
	 * be the same for all groups and agents in the simulation)
	 *
	 * @return the Group object represented by the xmlGroup element.
	 */
	public static Group parse( IXMLElement xmlGroup, int numDimensions )
	{
		Group newGroup = new Group();
		
		try
		{	
			// set the number of dimensions that this group should specify
			newGroup.dimensions = new Dimension[ numDimensions ];
			
			// Parse all the attributes
			Properties attributesOfGroup = xmlGroup.getAttributes();
			String strIndividuals = attributesOfGroup.getProperty( Constants.XML_ATTRIBUTE_INDIVIDUALS );
			String strPercentageOfPopulation = attributesOfGroup.getProperty( Constants.XML_ATTRIBUTE_PERCENT_OF_POPULATION );
			
			if ( (strIndividuals == null) && (strPercentageOfPopulation != null) )
			{
				newGroup.flgProportionalToPopulation = true;				
				newGroup.percentageOfPopulation = Double.parseDouble( strPercentageOfPopulation );
			}
			else if ( (strIndividuals != null) && (strPercentageOfPopulation == null) )
			{
				newGroup.flgProportionalToPopulation = false;
				newGroup.individuals = Integer.parseInt( strIndividuals );
			}
			else
			{
				ErrorHandler.exit( "ERROR parsing " + xmlGroup );
			}
			
			Enumeration enumChildren = xmlGroup.getChildren().elements();
			
			while( enumChildren.hasMoreElements() )
			{
				IXMLElement xmlChild = (IXMLElement)enumChildren.nextElement();
			
				// Parse the ranges or value of expertise and interests that 
				// the agents in this group will have.
				if ( xmlChild.getName().equalsIgnoreCase(Constants.XML_TAG_DIMENSION) )
				{
					Dimension dimension = Dimension.parse( xmlChild );
			
					// save the dimension in this group object
					newGroup.dimensions[ dimension.getID() ] = dimension;
				}
				// Parse the types of neighbors the agents in this group will have
				else if ( xmlChild.getName().equalsIgnoreCase(Constants.XML_TAG_NEIGHBORS) )
				{
					Properties attributesOfNeighbors = xmlChild.getAttributes();					
					newGroup.neighborPercentage = Double.parseDouble( attributesOfNeighbors.getProperty(Constants.XML_ATTRIBUTE_PERCENTAGE_OF_AGENTS) );
					newGroup.clustering = Double.parseDouble( attributesOfNeighbors.getProperty(Constants.XML_ATTRIBUTE_CLUSTERING) );												
				}
				// Parse the policies that the agents in this group will use
				else if ( xmlChild.getName().equals(Constants.XML_TAG_POLICIES) )
				{
					Enumeration enumChildrenOfPolicies = xmlChild.getChildren().elements();					

					while( enumChildrenOfPolicies.hasMoreElements() )
					{
						IXMLElement xmlChildOfPolicies = (IXMLElement)enumChildrenOfPolicies.nextElement();

						if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_ASKING) )
						{							
							newGroup.classAskingPolicy = Class.forName( xmlChildOfPolicies.getContent() );
						}
						else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_ANSWERING) )
						{	

							newGroup.classAnsweringPolicy = Class.forName( xmlChildOfPolicies.getContent() );
						}								
						else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_REFERRING) )
						{							
							newGroup.classReferringPolicy = Class.forName( xmlChildOfPolicies.getContent() );
						}																
						else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_LEARNING) )
						{							
							newGroup.classLearningPolicy = Class.forName( xmlChildOfPolicies.getContent() );
						}										
						else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_CHANGING_NEIGHBORS) )
						{							
							newGroup.classChangingNeighborsPolicy = Class.forName( xmlChildOfPolicies.getContent() );
						}
                                                else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_CACHING) )
						{							
							newGroup.classCachingPolicy = Class.forName( xmlChildOfPolicies.getContent() );
						}
					}
				}					
			}
		}
		catch( Exception e )
		{
			e.printStackTrace();
		}
		
		return newGroup;
	}
}
