//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//      the guidance of Dr. Munindar Singh. 
//============================================================================


package edu.ncsu.simulators.io;


import edu.ncsu.simulators.Simulator;
import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.agent.*;
import edu.ncsu.simulators.policies.*;

import edu.ncsu.simulators.util.ErrorHandler;
import edu.ncsu.simulators.util.Statistics;
import edu.ncsu.simulators.util.RandomIterator;
import edu.ncsu.simulators.vectors.*;

import net.n3.nanoxml.*;

import java.io.File;
import java.io.FileInputStream;

import java.math.BigDecimal;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.Properties;
import java.util.Random;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Vector;

/**
 * The SpecificationReader is a SimulationGenerator which parses a
 * generation specification XML element from a file, and then creates
 * a collection of new agents which conform to the specified
 * characteristics.
 */

public final class SpecificationReader implements SimulationGenerator {		

    // ---------------------------------------------------------
    // The fields below are used to implement the SimulationGenerator interaface

    private Agent[] agents = null;	
    private int numDimensions = 0;
    private AgentModel defaultAgentModel = null;

    // ---------------------------------------------------------
    // During the loadGenerationSpecificationMethod(), the generation
    // specification file is loaded and parsed into the fields listed below.
    // Then, the createAgents() uses the data in those fields to create the 
    // agents.

	

    // a set of group objects parsed from the file where group size is 
    // relative to the size of the population

    private HashSet setGroupsWithProportionalSize = null;

    // a set of group objects parsed from the file where group size is 
    // absolute

    private HashSet setGroupsWithFixedSize = null;



    // the classes that provide the default implementations for various policies

    private Class classDefaultAskingPolicy = null;
    private Class classDefaultAnsweringPolicy = null;
    private Class classDefaultReferringPolicy = null;
    private Class classDefaultLearningPolicy = null;
    private Class classDefaultChangingNeighborsPolicy = null;
    private Class classDefaultCachingPolicy = null;
    private Class classDefaultQueryGenerationPolicy = null;

    // the default neighbor configuration 

    private double defaultNeighborPercentage = -1;
    private double defaultClustering = -1;

    // ---------------------------------------------------------

    private int numAgents = 0;	


    //========================================================================			
    /**
     * A SpecificationReader is passed the name of a specification generation XML 
     * file in the constructor.  The SpecificationReader first parses the contents
     * of the file into various java primitives and a collection of Group objects.  
     * Group objects are used to temporarily store the data that is required to 
     * create groups of similar agents.
     * <p>
     * Then, the SpecificationReader processes the Group objects and other parsed
     * data, using it to create a set of agents which adhere to the specifications
     * that were in the file. 
     *
     * @param fileName the name of a specification generation XML file which 
     * describes what kind of agent population to create	
     */

    public SpecificationReader( String fileName ) {

	try {	

	    setGroupsWithFixedSize = new HashSet();
	    setGroupsWithProportionalSize = new HashSet();


	    File fileSpecification = new File( fileName );

	    // Parse the Simulation Generation XML element from the specified file

	    IXMLParser parser = XMLParserFactory.createDefaultXMLParser();
	    IXMLReader reader = new StdXMLReader( new FileInputStream(fileSpecification) );

	    parser.setReader( reader );

	    IXMLElement xmlGenerationSpecification = (IXMLElement)parser.parse();

	    Properties attributes = xmlGenerationSpecification.getAttributes();
			

	    // Set the number of dimensions
	    this.numDimensions = Integer.parseInt( attributes.getProperty(Constants.XML_ATTRIBUTE_DIMENSIONS) );

	    // Set the number of agents
	    this.numAgents = Integer.parseInt( attributes.getProperty(Constants.XML_ATTRIBUTE_POPULATION_SIZE) );			

	    // load the data from the file into this object's instance variables
	    loadGenerationSpecification( xmlGenerationSpecification );


	    // set the default policy implementations that groups
	    // should use where none were explicitly specified.
	    Group.setDefaultAskingPolicyClass( classDefaultAskingPolicy );
	    Group.setDefaultAnsweringPolicyClass( classDefaultAnsweringPolicy );
	    Group.setDefaultReferringPolicyClass( classDefaultReferringPolicy );
	    Group.setDefaultLearningPolicyClass( classDefaultLearningPolicy );
	    Group.setDefaultChangingNeighborsPolicyClass( classDefaultChangingNeighborsPolicy );
	    Group.setDefaultCachingPolicyClass( classDefaultCachingPolicy );
	    Group.setDefaultQueryGenerationPolicyClass( classDefaultQueryGenerationPolicy);
                        
                        
	    // set the default neighbor configuration that groups
	    // should use where none were explicitly specified.
	    Group.setDefaultNeighborPercentage( defaultNeighborPercentage );
	    Group.setDefaultClustering( defaultClustering );

	    // Set the agents array

	    this.agents = createAgents();

	}

	catch ( Exception e ) {

	    ErrorHandler.exit( "Error occured while loading or processing the generation specification file " + fileName );

	}

    }



    //========================================================================				

    /**
     * Assigns the specified number of ring neighbors and random neighbors to 
     * the specified agent.
     *
     * @param agent the agent which needs to have his neighbors initialized
     * @param numRingNeighbors the number of neighbors which should be "ring" neighbors.  A ring
     * neighbor is a agent who is selected as a neighbor based on the fact that
     * the agent's ID is the closest ID (as measured on a circular number line)
     * to the ID of the specified agent.  For example if the agents in a 
     * universe had IDs 0,1,2...,10 and we assigned 5 ring neighbors to 
     * agent 2, then we would assign agents 10,0,1,3,4 as the ring neighbors.
     * @param numRandomNeighbors the number of neighbors which should be 
     * randomly selected from all the remaining agents after the ring neighbors
     * have been assigned.
     */

    public void assignNeighbors( Agent agent, int numRingNeighbors, int numRandomNeighbors ) {

	Neighbors neighbors = new Neighbors();

	int counterLowerAgents = agent.getID();
	int counterHigherAgents = agent.getID();

	// assign the ring neighbors

	for( int indexNeighbors=0; indexNeighbors < numRingNeighbors; indexNeighbors++ ) {

	    if ( (indexNeighbors % 2) == 0 ) {
		if ( counterLowerAgents == 0 )
			counterLowerAgents = this.numAgents - 1;
		    else
			counterLowerAgents--;

		    neighbors.addNeighbor( counterLowerAgents );
		}

	    else {
		counterHigherAgents++;
		neighbors.addNeighbor( counterHigherAgents % this.numAgents);			

	    }

	}

	int randomNeighborCount = 0;	
	Iterator randomIterator = RandomIterator.getRandomIterator( agents );		

	// assign the random neighbors

	while ( randomNeighborCount < numRandomNeighbors ) {

	    int neighborID = ((Agent)randomIterator.next()).getID();

	    if ( (! neighbors.contains(neighborID)) && (neighborID != agent.getID())) { 

		neighbors.addNeighbor( neighborID );
		randomNeighborCount++;
	    }

	}

	try {
	    agent.setNeighbors( neighbors );
	}

	catch ( Exception e ) {

	    e.printStackTrace();
	}

    }
	

    //========================================================================	

    /**
     * Instantiate and populate the agents using the temporary objects
     * that we parsed from the generation specification file.
     */

    public Agent[] createAgents() {

	// Create an array to store the agents

	agents = new Agent[ this.numAgents ];

	try {

	    // Verify that sum of the agents in all the groups with
	    // absolute size does *not* exceed the total number of agents 
	    // in the simulation.

	    int totalIndividuals = 0;
	    Iterator iteratorGroupsWithFixedSize = setGroupsWithFixedSize.iterator();
	    
	    while( iteratorGroupsWithFixedSize.hasNext() ) {

		Group g = (Group)iteratorGroupsWithFixedSize.next();
		totalIndividuals = totalIndividuals + g.getIndividuals();
	    }

	    if ( totalIndividuals > this.numAgents ) 

		ErrorHandler.exit( "The sum of the agents [" + totalIndividuals + "] in all the groups with absolute size exceeds the number of agents in the simulation [" + this.numAgents + "]" ); 

	    // Verify that the sum of the percentages specified in all the
	    // groups with proportional size is equal to 100%.

	    double totalPercentage = 0;
	    Iterator iteratorGroupsWithProportionalSize = setGroupsWithProportionalSize.iterator();

	    while( iteratorGroupsWithProportionalSize.hasNext() ) {

		Group g = (Group)iteratorGroupsWithProportionalSize.next();
		totalPercentage = totalPercentage + g.getPercentageOfPopulation();
	    }

	    if ( totalPercentage != 100 )

		ErrorHandler.exit( "The sum of the percentages [" + totalPercentage + "] in all the groups with proportional size is not equal to 100" );

	    // Populate the array with some empty agents

	    for ( int i=0; i < this.numAgents; i++ )
		//if (Constants.AGENT_TYPE != null && Constants.AGENT_TYPE.equals("REPUTATION"))
		//    agents[i] = new ReputationAgent( i );
		//else
		agents[i] = new Agent( i);

	    // Get a random iterator over the collection of agents.
	    // The iterator allows us to easily get a handle on the next
	    // agent which has not yet beeen assigned the properites of a 
	    // group, AND it allows us to get agents in a random sequence.
	    // This helps avoid having all the agents with consecutive IDs
	    // belonging to the same group.

	    Iterator randomIterator = RandomIterator.getRandomIterator( agents );

	    // Keep track of how many agents in the randomIterator still
	    // have not had their properties set

	    int unassignedAgents = this.numAgents;				

	    //----------- Process groups with Fixed sizes ---------------

	    iteratorGroupsWithFixedSize = setGroupsWithFixedSize.iterator();
	    
	    // loop over all groups with a fixed size
	    while ( iteratorGroupsWithFixedSize.hasNext() ) {

		Group g = (Group)iteratorGroupsWithFixedSize.next();				
		// repeat for the number of agents in the group
		for ( int k=0; k<g.getIndividuals(); k++ ) {
		    // randomly select an agent from the array and
		    // assign its properties based on the specification
		    // for this group
			unassignedAgents--;
			Agent agent = (Agent)randomIterator.next();
			setAgentProperties( agent, g );
		}
	    }


	    //----------- Process groups with Proportional sizes ---------------

	    int numAgentsForProportionalDistribution = unassignedAgents;				
	    iteratorGroupsWithProportionalSize = setGroupsWithProportionalSize.iterator();

	    // loop over all groups with a proportional size
	    while ( iteratorGroupsWithProportionalSize.hasNext() ) {

		Group g = (Group)iteratorGroupsWithProportionalSize.next();				

		// calculate how many agents should be in this group
		// by taking the specified percentage of (the total agent
		// population minus the number of agents who were assigned
		// to groups with a fixed size)				

		BigDecimal base = new BigDecimal( numAgentsForProportionalDistribution );
		BigDecimal percentage = new BigDecimal( g.getPercentageOfPopulation()/100 );				
		BigDecimal result = base.multiply( percentage );

		// always round down to the nearest integer so that we
		// don't attempt to set properties for more agents than
		// there are remaining

		int numAgentsToAssignToThisGroup = result.setScale( 0, BigDecimal.ROUND_FLOOR ).intValue();
		
		// repeat for the number of agents in the group
		for ( int k=0; k<numAgentsToAssignToThisGroup; k++ ) {

		    // randomly select an agent from the array and
		    // assign its properties based on the specification
		    // for this group					
		    unassignedAgents--;
		    Agent agent = (Agent)randomIterator.next();	
		    setAgentProperties( agent, g );
		}
	    }

	    //------------------------------------------------------------			
	    // Because we always rounded down when we were calculating how
	    // many agents should be belong to the groups with proportional
	    // size, it could be possible that there are some agents which
	    // were never assigned to any group at all.
	    // There are two steps to fixing this.

	    // First, we sort the groups with proportional size in
	    // descending order by size.

	    iteratorGroupsWithProportionalSize = setGroupsWithProportionalSize.iterator();			
	    Vector vecSortedGroups = new Vector();

	    while ( iteratorGroupsWithProportionalSize.hasNext() ) {

		Group g = (Group)iteratorGroupsWithProportionalSize.next();					
		boolean flgInserted = false;

		for ( int i=0; i<vecSortedGroups.size(); i++ ) {

		    if ( g.getPercentageOfPopulation() > ((Group)vecSortedGroups.elementAt( i )).getPercentageOfPopulation() )
			{
			    vecSortedGroups.insertElementAt( g, i );
			    flgInserted = true;
			    break;
			}
		}

		if ( ! flgInserted )
		    vecSortedGroups.add( g );
	    }


	    // Second, we loop over all the groups from largest to smallest
	    // assigning one agent to each group.
	    // Repeat until there are no more unassigned agents.

	    while ( unassignedAgents > 0 ) {

		Iterator iteratorSortedGroups = vecSortedGroups.iterator();

		// loop over all groups with a proportional size
		while ( iteratorSortedGroups.hasNext() ) {

		    Group g = (Group)iteratorSortedGroups.next();				
		    unassignedAgents--;

		    Agent agent = (Agent)randomIterator.next();	
		    setAgentProperties( agent, g );

		    if ( unassignedAgents == 0 ) {
			break;
		    }
		}	

	    }

		//--------------------------------------------------------			

	}

	catch ( Exception e ) {
	    ErrorHandler.exit( e, "The SpecificationReader could not create the agents specified in the file" );
	}
		
	return agents;

    }

    //========================================================================			

    public Agent[] getAgents() { 
	return this.agents; 
    }

    //========================================================================			

    public int getDimensions() { 
	return this.numDimensions; 
    }

    //========================================================================			

    public AgentModel getDefaultAgentModel() { 
	return this.defaultAgentModel; 
    }		

    //========================================================================			
    /**
     * Parses the data from an generation specification XML element, and 
     * temporarily stores the data in instance variables until the createAgents 
     * method uses the data. 
     */

    public void loadGenerationSpecification( IXMLElement xmlGenerationSpecification ) {

	try {					
	    Enumeration enumChildrenOfSimulationGeneration = xmlGenerationSpecification.getChildren().elements();

	    while( enumChildrenOfSimulationGeneration.hasMoreElements() ) {
		IXMLElement xmlChildOfSimulationGeneration = (IXMLElement)enumChildrenOfSimulationGeneration.nextElement();
		if ( xmlChildOfSimulationGeneration.getName().equals(Constants.XML_TAG_GROUP) ) {
		    
		    Group g = Group.parse( xmlChildOfSimulationGeneration, this.numDimensions );
		    if ( g.isProportionalToPopulation() )
			this.setGroupsWithProportionalSize.add( g );
		    else
			this.setGroupsWithFixedSize.add( g );
		}
		else if ( xmlChildOfSimulationGeneration.getName().equals(Constants.XML_TAG_DEFAULTS) ) {

		    Enumeration enumChildrenOfDefaults = xmlChildOfSimulationGeneration.getChildren().elements();	

		    while( enumChildrenOfDefaults.hasMoreElements() ) {

			IXMLElement xmlChildOfDefaults = (IXMLElement)enumChildrenOfDefaults.nextElement();
			// load the model of other ageents that every agent will start with
			if ( xmlChildOfDefaults.getName().equals(Constants.XML_TAG_AGENT_MODEL) )
			    defaultAgentModel = AgentModel.parse( xmlChildOfDefaults );

			// load the neighbor configuration that an agent will
			// use unless it is overridden at the group level
			else if ( xmlChildOfDefaults.getName().equals(Constants.XML_TAG_NEIGHBORS) ) {
			    Properties attributesOfNeighbors = xmlChildOfDefaults.getAttributes();

			    defaultNeighborPercentage = Double.parseDouble( attributesOfNeighbors.getProperty(Constants.XML_ATTRIBUTE_PERCENTAGE_OF_AGENTS) );
			    defaultClustering = Double.parseDouble( attributesOfNeighbors.getProperty(Constants.XML_ATTRIBUTE_CLUSTERING) );												
			}						
			// load the policies that an agent will use unless
			// they are overridden at the group level
			else if ( xmlChildOfDefaults.getName().equals(Constants.XML_TAG_POLICIES) ) {
			    Enumeration enumChildrenOfPolicies = xmlChildOfDefaults.getChildren().elements();					
			    while( enumChildrenOfPolicies.hasMoreElements() ) {

				IXMLElement xmlChildOfPolicies = (IXMLElement)enumChildrenOfPolicies.nextElement();
				if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_ASKING) ) 
				    classDefaultAskingPolicy = Class.forName( xmlChildOfPolicies.getContent() );

				else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_ANSWERING) )
				    classDefaultAnsweringPolicy = Class.forName( xmlChildOfPolicies.getContent() );

				else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_REFERRING) )
				    classDefaultReferringPolicy = Class.forName( xmlChildOfPolicies.getContent() );

				else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_LEARNING) )
				    classDefaultLearningPolicy = Class.forName( xmlChildOfPolicies.getContent() );

				else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_CHANGING_NEIGHBORS) )
				    classDefaultChangingNeighborsPolicy = Class.forName( xmlChildOfPolicies.getContent() );

				else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_CACHING) )
				    classDefaultCachingPolicy = Class.forName( xmlChildOfPolicies.getContent() );
				else if ( xmlChildOfPolicies.getName().equals(Constants.XML_TAG_POLICY_QUERY_GENERATION) )
				    classDefaultQueryGenerationPolicy = Class.forName( xmlChildOfPolicies.getContent() );
			    } //end while
			} //end else if					
		    } //end while

		} //end else if

	    } //end while

	} //end try

	catch ( Exception e ) {
	    ErrorHandler.exit( e, "SpecificationReader could not parse the generation specification XML element loaded from the file");
	}

    }	

	
    //========================================================================		
    /**
     * Sets all the properties of the agent based on the specification 
     * encapsulated in the group object.
     */

    private void setAgentProperties( Agent agent, Group g ) {

	try {

	    Expertise expertise = new DomainExpertise( this.numDimensions );
	    DomainInterests interests = new DomainInterests( this.numDimensions );

	    // loop over the specifications for each dimension for this group
	    // set the expertise and interest value for a dimension on each pass

	    for ( int d=0; d<this.numDimensions; d++ ) {

		Dimension dimension = g.getDimension( d );	
		if ( dimension.hasRangeOfExpertise() )
		    expertise.setDimensionValueWithinRange( d, dimension.getExpertiseRangeLow(), dimension.getExpertiseRangeHigh() );
		else
		    expertise.setDimensionValue( d, dimension.getExpertiseValue() );

		if ( dimension.hasRangeOfInterests() )
		    interests.setDimensionValueWithinRange( d, dimension.getInterestsRangeLow(), dimension.getInterestsRangeHigh() );
		else
		    interests.setDimensionValue( d, dimension.getInterestsValue() );
	    }


	    // set the agent's expertise and interests vectors
	    agent.setExpertise( expertise );					
	    agent.setInterests( interests );		
	
	    // set the agent's policies
	    agent.setAskingPolicy( (AskingPolicy)g.getAskingPolicyClass().newInstance() );
	    agent.setAnsweringPolicy( (AnsweringPolicy)g.getAnsweringPolicyClass().newInstance() );
	    agent.setReferringPolicy( (ReferringPolicy)g.getReferringPolicyClass().newInstance() );
	    agent.setLearningPolicy( (LearningPolicy)g.getLearningPolicyClass().newInstance() );
	    agent.setChangingNeighborsPolicy( (ChangingNeighborsPolicy)g.getChangingNeighborsPolicyClass().newInstance() );
	    agent.setCachingPolicy( (CachingPolicy)g.getCachingPolicyClass().newInstance() );
	    agent.setQueryGenerationPolicy((QueryGenerationPolicy)g.getQueryGenerationPolicyClass().newInstance() );

	    // determine how many ring neighbors and how many random neighbors
	    // this agent will have
	    BigDecimal totalAgents = new BigDecimal( this.numAgents ); 
	    BigDecimal percentageNeighborsWith = new BigDecimal( g.getNeighborPercentage() );
	    BigDecimal clustering = new BigDecimal( g.getClustering() );
	    BigDecimal numNeighbors = totalAgents.multiply( percentageNeighborsWith ).setScale(0,BigDecimal.ROUND_HALF_DOWN);

	    int numRingNeighbors = numNeighbors.multiply(clustering).setScale(0,BigDecimal.ROUND_HALF_DOWN).intValue();
	    int numRandomNeighbors = numNeighbors.intValue() - numRingNeighbors;
	    // set the agent's neighbors			
	    assignNeighbors( agent, numRingNeighbors, numRandomNeighbors );

	}

	catch( Exception e ) {
	    ErrorHandler.exit( "The SpecificationReader could not set the properties of the given agent using the given group object");
	}

    }	

}
