//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.io;

import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.agent.*;
import edu.ncsu.simulators.vectors.*;


import net.n3.nanoxml.*;

import java.lang.*;
import java.io.*;
import java.util.*;

import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.Enumeration;
import java.util.Vector;

/**
 * The StateReader is a SimulationGenerator which parses a collection of existing agents from a state file. 
 * The StateReader is designed to accept as input the name of a state file 
 * which was generated as input by the StateWriter.
 */
public final class StateReader implements SimulationGenerator
{	
	private Agent[] agents = null;	
	private int numDimensions = 0;
	private AgentModel defaultAgentModel = null;
	
	//========================================================================			
	
	public StateReader( String fileName )
	{		
		try
		{
			File fileState = new File( fileName );
			
			int i=0;
			IXMLParser parser = XMLParserFactory.createDefaultXMLParser();
			IXMLReader reader = new StdXMLReader(new FileInputStream(fileState));
			parser.setReader(reader);

			IXMLElement xmlAgents = (IXMLElement) parser.parse();
			
			Properties attributes = xmlAgents.getAttributes();
			
			// Set the number of dimensions
			this.numDimensions = Integer.parseInt( attributes.getProperty(Constants.XML_ATTRIBUTE_DIMENSIONS) );
			
			int numAgents = Integer.parseInt( attributes.getProperty(Constants.XML_ATTRIBUTE_POPULATION_SIZE) );		
			
			agents = new Agent[numAgents];
			int agentCounter = 0;
			
			Enumeration enumChildrenOfAgents = xmlAgents.getChildren().elements();
			
			while( enumChildrenOfAgents.hasMoreElements() )
			{
				IXMLElement xmlChildOfAgents = (IXMLElement)enumChildrenOfAgents.nextElement();
				
  	
				if ( xmlChildOfAgents.getName().equals(Constants.XML_TAG_AGENT) )
				{
					Agent agent = Agent.parse( xmlChildOfAgents, new DomainExpertise(), new DomainInterests() );
					agents[agentCounter++] = agent;
				}
				// Create an agent based on the tag.

				if ( xmlChildOfAgents.getName().equals(Constants.XML_TAG_SERVICE_AGENT) )
				    {   //AgentParser.parseServiceAgent
					Agent agent = (ServiceAgent)AgentParser.parseServiceAgent( xmlChildOfAgents, new LatticeExpertise(), new DomainInterests());
					agents[agentCounter++] = agent;
					
				}
				
				else if ( xmlChildOfAgents.getName().equals(Constants.XML_TAG_AGENT_MODEL) )
				{
					defaultAgentModel = AgentModel.parse( xmlChildOfAgents );				
				}
			}		
		}
		catch ( Exception e )
		{
			e.printStackTrace();
		}
	}	
	
	//========================================================================			

	public Agent[] getAgents()
	{
		return this.agents;
	}
	
	//========================================================================			

	public int getDimensions()
	{
		return this.numDimensions;
	}	
	
	//========================================================================			

	public AgentModel getDefaultAgentModel()
	{
		return this.defaultAgentModel;
	}			
	
	//========================================================================		
}
