//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.policies;

import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.agent.AgentModel;
import edu.ncsu.simulators.agent.AgentModels;
import edu.ncsu.simulators.agent.Neighbors;
import edu.ncsu.simulators.io.LogWriter;
import edu.ncsu.simulators.policies.Heuristics;
import edu.ncsu.simulators.util.ErrorHandler;
import edu.ncsu.simulators.vectors.Expertise;
import edu.ncsu.simulators.vectors.Interests;
import edu.ncsu.simulators.vectors.Sociability;

import java.util.*;

/**
 * This "standard" policy provides a very straight forward implementation of a
 * ChangingNeighborsPolicy.
 *
 * @author Mike Suralik
 */
public class ChangingNeighborsPolicy_Standard implements ChangingNeighborsPolicy
{	
	private LogWriter logWriter = null;
	
	//========================================================================	
	
	public ChangingNeighborsPolicy_Standard()
	{
	}
	
	//========================================================================	
	
	public void setLogWriter( LogWriter lw )
	{
		this.logWriter = lw;
	}		
	
	//========================================================================	
	/**
	 * Under this implementation of the policy:
	 *
	 * 1) the number of neighbors does not change (i.e. every addition of a 
	 * neighbor is accompanied by the removal of an neighbor).
	 * 2) after all the adds and removes, neighbors will contain those agents
	 * from the agentModels who have the highest relevance to the agent making
	 * the change
	 *
	 * @param agentModels the models of all the agents that the agent who is making the change is aware of
	 * @param expertise the expertise of agent who is making the change
	 * @param interests the expertise of the agent who is making the change.
	 */	
    public synchronized void changeNeighbors(Neighbors neighbors, AgentModels agentModels, Expertise expertise, Interests interests) {

	//	Iterator iteratorAgentModels = agentModels.getModels();
	Enumeration e = agentModels.getEnumeration();
	while(e.hasMoreElements()) {

	    AgentModel modelCandidate = (AgentModel)e.nextElement();
	    Integer candidateID = new Integer(modelCandidate.getID());

	    // if the model is *NOT* a neighbor
	    if (!neighbors.contains( candidateID)) {

		double relevanceOfCandidate = calculateRelevanceOfAgent( interests, modelCandidate);
		// find the neighbor with the minimum relevance				
		    
		Integer worstNeighborID = null;				
		double worstNeighborRelevance = Double.MAX_VALUE;
		Neighbors temp = neighbors.copy();
		Iterator iteratorNeighbors = temp.iterator();
				
		while(iteratorNeighbors.hasNext()) {

		    Integer neighborID = (Integer)iteratorNeighbors.next();
		    AgentModel modelNeighbor = agentModels.getModel( neighborID );
		    double relevanceOfNeighbor = calculateRelevanceOfAgent( interests, modelNeighbor );

		    // if the candidate has more relevance than the
		    // least relevant neighbor, then replace the neighbor
		    // with the candidate
		    if (relevanceOfNeighbor <= worstNeighborRelevance) {
			worstNeighborRelevance = relevanceOfNeighbor;
			worstNeighborID = neighborID;
		    }
		}

		// compare the nonNeighbor that we are currently evaluating to the worst neighbor
		if (relevanceOfCandidate > worstNeighborRelevance) {
		    log( "replacing Neighbor(" + worstNeighborID + ") with NON-Neighbor(" + modelCandidate.getID() + ")" );
		    // System.out.println("replacing Neighbor(" + worstNeighborID + ") with NON-Neighbor(" + modelCandidate.getID() + ")" );
		    neighbors.remove( worstNeighborID );
		    neighbors.addNeighbor( candidateID.intValue());
		}
	    }
	}
    }

    //=======================================================================	
    /**
     * Calculate the relevance of the compared agent to basis agent - based 
     * on the compared agent's expertise and sociability and basis agent's 
     * interests.
	 */
    private double calculateRelevanceOfAgent( Interests interestsOfBasisAgent, AgentModel modelComparedAgent) {
	Expertise expertiseOfComparedAgent = modelComparedAgent.getExpertise();
	Sociability sociabilityOfComparedAgent = modelComparedAgent.getSociability();
		
	double relevance = ((1-Constants.W) * Heuristics.calculateSimilarity(interestsOfBasisAgent, expertiseOfComparedAgent)) + (Constants.W * Heuristics.calculateSociabilitySimilarity(interestsOfBasisAgent, sociabilityOfComparedAgent));

	return relevance;
    }	

    //=======================================================================		
	
    private void log(String text) {

	if (logWriter != null)
	    logWriter.log( text );
    }
	
    //=======================================================================		
}
