//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.policies;

import edu.ncsu.simulators.agent.AgentModels;
import edu.ncsu.simulators.agent.AgentModel;
import edu.ncsu.simulators.dialogs.ReferralChain;
import edu.ncsu.simulators.vectors.Expertise;
import edu.ncsu.simulators.vectors.Interests;
import edu.ncsu.simulators.vectors.Query;

import java.io.FileWriter;

/**
 * Defines the interface between an agent and his policy regarding the way
 * that he learns things about the other agents that he is interacting with.
 */
public interface LearningPolicy
{	
	/**
	 * Modifies the model of the agent who asked the specified query.
	 *
	 * The query parameter is a copy of the actual object that it represents.  
	 * It may be altered in order to facilitate the modification of the 
	 * questionerModel. However, altering the query parameter will have no 
	 * effect outside this method.
	 *
	 * @param questionerModel the model of the agent who asked the query as 
	 * maintained by the agent who received the query
	 * @param query the query that was asked by the questioning agent
	 */
	public void learnFromQuestion( AgentModel questionerModel, Query query );	
	
	/**
	 * This method must update the agentModels collection based on the answer 
	 * that the answering agents gave in response to the query, and on the 
	 * referrals that led the questioning agents to ask the answering
	 * agents in the first place.
	 *
	 * @param referralChain if the answering agent was a referral, it will be in this referral chain; otherwise this parameter will be null
	 * @param agentModels all of the agent models tha the questioning agent maintains
	 * @param expertiseOfAskingAgent the expertise of the agent who asked this question
	 */ 
	public void learnFromResponses( ReferralChain referralChain, AgentModels agentModels, Expertise expertiseOfAskingAgent );	
	
	/**
	 * Sets the FileWriter that the policy can use to output the results
	 */
	public void setFileWriter( FileWriter fw );
}