//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================



package edu.ncsu.simulators.policies;



import edu.ncsu.simulators.Constants;

import edu.ncsu.simulators.agent.AgentModel;
import edu.ncsu.simulators.agent.AgentModels;

import edu.ncsu.simulators.dialogs.ReferralChain;
import edu.ncsu.simulators.dialogs.ReferralChainDAG;

import edu.ncsu.simulators.util.ErrorHandler;

import edu.ncsu.simulators.policies.Heuristics;

import edu.ncsu.simulators.vectors.*;

import java.io.FileWriter;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;



/**

 * This "standard" policy provides a very straight forward implementation of a
 * LearningPolicy.
 *
 * @author Mike Suralik

 */

public class LearningPolicy_Standard implements LearningPolicy
{	

    private static final int BAD_ANSWER = 1;
    private static final int GOOD_ANSWER = 2;
    private static final int NO_ANSWER = 3;

    private int goodAnswerCount;
    private int badAnswerCount;
    private int noAnswerCount;

    private FileWriter fwLog = null;

    //========================================================================	

    public LearningPolicy_Standard()

    {

    }
    //========================================================================	
    public void setFileWriter( FileWriter fw) {
	this.fwLog = fw;
    }	

    //========================================================================			
    /**
     * For the agent who received a query, update that agent's model of
     * the agent who asked the question.  Specifically, update the model
     * of the interests by taking a weighted average of the current model
     * of the interests and the query.
     */

    public void learnFromQuestion( AgentModel modelQuestioner, Query query ) {

	Interests currentEstimatedInterests = modelQuestioner.getInterests();

	for ( int i=0; i<currentEstimatedInterests.size(); i++ ) {

	    Double currentEstimatedInterestsComponent = (Double)currentEstimatedInterests.elementAt( i );
	    Double newlyObservedInterestsComponent = (Double)query.elementAt( i );
	    
	    Double newEstimatedInterestsComponent = new Double( (newlyObservedInterestsComponent.doubleValue() * Constants.RATE_OF_LEARNING_INTERESTS) + (currentEstimatedInterestsComponent.doubleValue() * (1-Constants.RATE_OF_LEARNING_INTERESTS)) ); 				

	    currentEstimatedInterests.setElementAt( newEstimatedInterestsComponent, i );

	}

    }	

    //========================================================================	

    public void learnFromResponses( ReferralChain referralChain, AgentModels agentModels, Expertise expertiseOfAskingAgent) {

	goodAnswerCount = 0;
	badAnswerCount = 0;
	noAnswerCount = 0;

	double bestSimilarity = Double.MIN_VALUE;
	Answer bestAnswer = null;

	Query query = referralChain.getQuery();
	Hashtable hashAnswers = referralChain.getAnswers();		

	Enumeration enumAgentIDs = hashAnswers.keys();
	Enumeration enumAnswers = hashAnswers.elements();

	while (enumAgentIDs.hasMoreElements()) {

	    Integer answeringAgentID = (Integer)enumAgentIDs.nextElement();		
	    Answer answer = (Answer)enumAnswers.nextElement();				
	    AgentModel modelAnsweringAgent = agentModels.getModel( answeringAgentID );

	    int grade = evaluateAnswer( query, answer );
	    
	    if ( grade == GOOD_ANSWER) {

		double similarity = Heuristics.calculateSimilarity( query, answer );
		//System.out.println( "Got a good answer with Similarity = " + similarity );
		if ( similarity > bestSimilarity) {
		    
		    bestSimilarity = similarity;
		    bestAnswer = answer;
		    /*Set the best answer in the referral chain... the best answer
		     *is cached... This class might not be the best place to do it. Evaluating
		     *answers should be part of an abstract class 'Learning Policy' perhaps
		     */
		    ((ReferralChainDAG)(referralChain)).setBestAnswer(bestAnswer);
		}
	    }

	    updateExpertise( modelAnsweringAgent, grade, answer ); 

	    int generationsRemoved = 0;
	    Enumeration enumGenerations = referralChain.getAncestorsByGeneration( answeringAgentID.intValue() );

	    while ( enumGenerations.hasMoreElements()) {

		generationsRemoved++;

		Enumeration enumAncestors = (Enumeration)enumGenerations.nextElement();
		while( enumAncestors.hasMoreElements()) {

		    Integer ancestorID = (Integer)enumAncestors.nextElement();
		    AgentModel modelAncestor = agentModels.getModel( ancestorID );					
		    //double oldSociability = modelAncestor.getSociability();
		    updateSociability(modelAncestor, query, grade, generationsRemoved );
		    //double newSociability = modelAncestor.getSociability();
		}
	    }

	}
		
	try {

	    if (fwLog != null) {
		fwLog.write( "GOOD " + goodAnswerCount + '\r' + '\n' );
		fwLog.write( "BAD " + badAnswerCount + '\r' + '\n' );
		fwLog.write( "NO " + noAnswerCount + '\r' + '\n' );
	    }

	}
	catch ( Exception e) { 
	    e.printStackTrace();
	}

	// update the expertise of the agent based on the best correct answer
	// that it received (if any).

	int i = 0;

	if (bestAnswer != null) {

	    Iterator iteratorExpertiseComponents = expertiseOfAskingAgent.iterator();
	    Iterator iteratorBestAnswerComponents = bestAnswer.iterator();

	    while( iteratorExpertiseComponents.hasNext()) {

		double bestAnswerComponent = ((Double)iteratorBestAnswerComponents.next()).doubleValue();
		double oldExpertiseComponent = ((Double)iteratorExpertiseComponents.next()).doubleValue();
		double newExpertiseComponent = Math.min( 1, oldExpertiseComponent + (Constants.RATE_OF_LEARNING_FROM_GOOD_ANSWERS * bestAnswerComponent) ); 
		//System.out.println( "oldExpertiseComponent = " + oldExpertiseComponent );
		//System.out.println( "bestAnswerComponent = " + bestAnswerComponent );			
		//System.out.println( "newExpertiseComponent = " + newExpertiseComponent );						
		expertiseOfAskingAgent.setElementAt( new Double(newExpertiseComponent), i++ );
	    }

	    //expertiseOfAskingAgent = newExpertise;
	    //System.out.println( "newExpertise = " + newExpertise );				
	    //System.out.println( "expertiseOfAskingAgent = " + expertiseOfAskingAgent );    
	}

    }

    //========================================================================	

    /**
     * Evaluates the specified answer to the specified question, and returns
     * a grade.
     */

    public int evaluateAnswer( Query query, Answer answer ) {

	int evaluation = -1;

	if ( answer == Constants.NULL_ANSWER )
	    evaluation = NO_ANSWER;

	else if ( Heuristics.areSufficientlySimilar(query,answer,Constants.SIMILARITY_THRESHOLD_FOR_EVALUATING_ANSWER) )
	    evaluation = GOOD_ANSWER;
	else
	    evaluation = BAD_ANSWER;


	return evaluation;

    }

	

    //========================================================================			

    /**
     * Updates the expertise of the specified agent model based on the 
     * specified answer that it provided to the specified query.
     */

    public  void updateExpertise( AgentModel answererModel, int grade, Answer answer )

    {

	Expertise currentExpertise = answererModel.getExpertise();
	Expertise newExpertise = new DomainExpertise();


	Double currentExpertiseComponent = null;
	Double answerComponent = null;
	Double newExpertiseComponent = null;		

	//	String stringGrade = "";

	if ( grade == GOOD_ANSWER) {

	    goodAnswerCount++;
	    //  stringGrade = "a GOOD answer";
	}
	else if ( grade == BAD_ANSWER) {
	    badAnswerCount++;
	    //stringGrade = "a BAD answer";
	}
	else if ( grade == NO_ANSWER) {
	    noAnswerCount++;
	    //stringGrade = "NO answer";
	}

	//System.out.println( "updating the expertise of agent(" + answererModel.getID() + ") who gave " + stringGrade );

	if ( (grade == GOOD_ANSWER) || (grade == BAD_ANSWER)) {

	    for ( int i=0; i<currentExpertise.size(); i++ ){
		currentExpertiseComponent = (Double)currentExpertise.elementAt( i );

		answerComponent = (Double)answer.elementAt( i );
		newExpertiseComponent = new Double( ((1-Constants.ALPHA) * currentExpertiseComponent.doubleValue()) + (Constants.ALPHA * answerComponent.doubleValue()) ); 

		newExpertise.addElement( newExpertiseComponent );

	    }

	}

	else if ( grade == NO_ANSWER) {

	    for ( int i=0; i<currentExpertise.size(); i++) {
		currentExpertiseComponent = (Double)currentExpertise.elementAt( i );

		newExpertiseComponent = new Double( (1-Constants.PERCENTAGE_DECAY_FOR_NO_ANSWER) * currentExpertiseComponent.doubleValue() );
		newExpertise.addElement( newExpertiseComponent );				
	    }				
	}

	else {
	    ErrorHandler.exit( "Attempting to update expertise based on an answer with an unknown grade type" );
	}		

	//System.out.println( "agent(" + answererModel.getID() + ")'s old expertise = " + currentExpertise + ", new expertise = " + newExpertise );
	answererModel.setExpertise( newExpertise );
    }				



    public  double updateSociabilityElement (double initial, int grade, int generationsRemoved, double weight) {

	double newSociability = 0;
	if ( grade == GOOD_ANSWER ) {
	    newSociability = initial + ((((1-initial)*Constants.BETA)/generationsRemoved)*weight);
	    //	    System.out.println("initial " + initial + " new " + newSociability);
	}
	else if ( (grade == BAD_ANSWER) || (grade == NO_ANSWER) )
	    newSociability = initial - (((initial*Constants.BETA)/generationsRemoved)*weight);			
	else
	    ErrorHandler.exit( "Attempting to update sociability based on an answer with an unknown grade type" );

	return newSociability;
    }

    //========================================================================			
    /**
     * Updates the sociability of the specified agent based on the assumption
     * that is was separated by <generationsRemoved> referrals from an agent
     * that gave an answer which evaluated to <grade>
     */

    public  void updateSociability( AgentModel modelAncestor, Query currentQuery, int grade, int generationsRemoved )
    {

	Sociability oldSociability = modelAncestor.getSociability();
	Sociability newSociability = null;
	double currentSociabilityComponent;
	double currentQueryComponent;
	double newSociabilityComponent;
	
	if ( oldSociability instanceof ScalarSociability) {
	    newSociability = new ScalarSociability(updateSociabilityElement((
			     (ScalarSociability)oldSociability).getSociability(), grade, generationsRemoved, 1));
	} // end of if ()
	else {

	    oldSociability = (DomainSociability)oldSociability;
	    newSociability = new DomainSociability();

	    for ( int i=0; i<oldSociability.size(); i++ ){
		currentSociabilityComponent = ((Double)oldSociability.elementAt(i)).doubleValue();
		currentQueryComponent = ((Double)currentQuery.elementAt(i)).doubleValue();
		
		if (currentQueryComponent > 0.2) {
		    
		    newSociabilityComponent = updateSociabilityElement(currentSociabilityComponent, grade, generationsRemoved, currentQueryComponent);
		} // end of if ()
		else {
		    
		    newSociabilityComponent = currentSociabilityComponent;
		} // end of else
		

		newSociability.addElement(new Double( newSociabilityComponent) );

	    }	   

	    
	} // end of else


	modelAncestor.setSociability( newSociability );		

    }

		

}
