//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at the North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	made by Pinar Yolum and Bin Yu, who were conducting research under the 
//	guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.util;

import edu.ncsu.simulators.vectors.SimulationVector;
import edu.ncsu.simulators.vectors.lattice.*;
import edu.ncsu.simulators.Constants;

import java.util.Enumeration;
import java.util.*;

public final class Heuristics
{
    //========================================================================			
	
    private Heuristics()
    {	
    }	
	

    /**
     *  Determines whether the similarity of the specified query and expertise
     * is greater than the specified threshold.
     *
     * @param basisVector a <code>SimulationVector</code> value
     * @param comparedVector a <code>SimulationVector</code> value
     * @param threshold a <code>double</code> value
     * @return a <code>boolean</code> value
     */
    public static boolean areSufficientlySimilar(SimulationVector basisVector, SimulationVector comparedVector, double threshold) {

	double similarity = calculateSimilarity( basisVector, comparedVector );
	return (similarity>threshold);
    }	
	
    /**
     * 
     * Determines whether the similarity of the specified query and
     * expertise is greater than the specified threshold.
     * 
     *
     * @param basisVector a <code>SimulationVector</code> value
     * @param comparedVector a <code>SimulationVector</code> value
     * @return a <code>double</code> value 
     */

    public static double calculateSimilarity( SimulationVector basisVector, SimulationVector comparedVector )
    {	
	Enumeration enumBasisComponents = basisVector.elements();
	Enumeration enumComparedComponents = comparedVector.elements();
		
	double similarityNumerator = 0;
	double similarityDenominator = 0;
	double similarity = 0;

	while (enumBasisComponents.hasMoreElements()) {

	    double basisComponent = ((Double)enumBasisComponents.nextElement()).doubleValue();			
	    double comparedComponent = ((Double)enumComparedComponents.nextElement()).doubleValue();
	    
	    similarityNumerator = similarityNumerator + comparedComponent * basisComponent;
	    similarityDenominator = similarityDenominator + basisComponent * basisComponent;
	}

	similarityDenominator = Math.sqrt( basisVector.size() * similarityDenominator );

	similarity = similarityNumerator / similarityDenominator;
		
	return similarity;
    }	
	

    /**
     * Determines whether a provider can be promoted to a new role
     *
     * @param provider a <code>ProviderListNode</code> has a
     * providerId and a list of ratings for the initial service
     * @param weight a <code>double</code> is the current weight
     * between the initial and the candiate service
     * @param threshold a <code>double</code> value
     * @return a <code>boolean</code> value 
     */

    public static boolean isPromotable(ProviderListNode provider, double weight, double threshold){
        return ( (provider.numberOfRatings() > Constants.NUMBER_OF_INTERACTIONS) && ((provider.avgRating() * weight) > threshold));
    }
    //========================================================================    
}


