package edu.ncsu.simulators.vectors;

/**
 * LatticeExpertise.java
 *
 *
 * Created: Tue May  7 18:24:00 2002
 *
 * @author <a href="mailto:pyolum@eos.ncsu.edu">Pinar Yolum</a>
 * @version
 */

import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.vectors.lattice.*;
import edu.ncsu.simulators.vectors.Expertise;
import net.n3.nanoxml.*;
import java.util.*;


public class LatticeExpertise extends Expertise {

    protected int dimension = -1;
    protected WeightMatrix weights;
    protected Hashtable providers = new Hashtable();
    protected Hashtable promotionTable = new Hashtable();

    public LatticeExpertise (){
    }

    /**
     * Creates a new <code>LatticeExpertise</code> instance.
     *
     * @param d an <code>int</code> Number of points in the lattice
     * @param w a <code>WeightMatrix</code> Keeps the similarity weights for the lattice
     * @param p a <code>Hashtable</code> Keeps a list of providers for each node
     */
    public LatticeExpertise (int d, WeightMatrix w, Hashtable p){
	this.dimension = d;
	this.weights = w;
	this.providers = p;
    }
        
    /**
     * Get the value of promotionTable.
     * @return value of promotionTable.
     */
    public Hashtable getPromotionTable() {
	return promotionTable;
    }
    
    /**
     * Set the value of promotionTable.
     * @param v  Value to assign to promotionTable.
     */
    public void setPromotionTable(Hashtable  v) {
	this.promotionTable = v;
    }
    

    /**
     * Returns true only if the provider has been promoted to roleId
     *
     * @param provider an <code>int</code> value
     * @param roleId an <code>int</code> value
     * @return a <code>boolean</code> value
     */
    public boolean isPromoted (int provider, int roleId) {

	Hashtable h = (Hashtable)(promotionTable.get(new Integer(provider)));

	return ((h != null) && (h.get(new Integer(roleId)) != null));

    }
    public void updatePromotionTable(int provider, Integer firstRole, Integer secondRole) {
	
	Hashtable twoRoles = null;

	// this returns all the promotions of the provider
	Object entry = promotionTable.get(new Integer(provider));
	
	if ( entry == null) {

	    twoRoles = new Hashtable();
	    
	} // end of if ()
	else { //the provider has other promotions

	    twoRoles = (Hashtable)entry;
	    
	} // end of if ()else

	//Provider has been promoted to the secondRole from the firstrole
	twoRoles.put(secondRole, firstRole);
	promotionTable.put(new Integer(provider), twoRoles);


    }

    /**
     *
     * @return an <code>Expertise</code> that is the duplicate of this one
     */
    public Expertise copy(){
	return (Expertise)copy (new LatticeExpertise());
    }

    /**
     * Describe <code>updateProviders</code> method here.
     *
     * @param roleNumber an <code>int</code> value
     * @param providerId an <code>int</code> value
     * @param rating a <code>double</code> value
     * @param promoted a <code>boolean</code> value
     */
    public void updateProviders (int roleNumber, int providerId, double rating, boolean promoted) {

	ProviderList p = (ProviderList)providers.get(new Integer(roleNumber));
	if (p == null) {
	    p = new ProviderList (new Integer(roleNumber));
	} // end of if ()
	p.update (providerId, rating, promoted);
	providers.put(new Integer(roleNumber), p);

    }

    /**
     * if the promoted bit is not specified assume false
     *
     * @param roleNumber an <code>int</code> value
     * @param providerId an <code>int</code> value
     * @param rating a <code>double</code> value
     */
    public void updateProviders (int roleNumber, int providerId, double rating) {

	updateProviders(roleNumber, providerId, rating, false);
    }


    /**
     * Returns true if the agent with agentId is providing the service
     * with roleId
     *
     * @param agentId an <code>int</code> value
     * @param roleId an <code>int</code> value
     * @return a <code>boolean</code> value
     */
    public boolean providesService (int agentId, int roleId){
	
	ProviderList pList =(ProviderList)providers.get(new Integer(roleId)); 
	boolean found = false;
	// if there are no providers for the role, return false.
	if ( pList != null) {
	    
	    Iterator i = pList.iterator();
	    ProviderListNode currentNode = null;
	    //Traverse through all providers for that role
	    //and check for matching agendId.
	    while ((i.hasNext()) && (!found)){
		currentNode = (ProviderListNode)i.next();
		if (currentNode.getProviderId() == agentId)
		    found = true;
	    }

	} // end of if ()
	

	return found;
	    
    }

    /* Returns the number of roles in the lattice */
    public int getSize(){
	return dimension;
    }

    public void assign (int providerId, int newRoleId){
	//Find the list of providers for the newRoleId
	ProviderList p = (ProviderList)providers.get(new Integer(newRoleId));
	//Create a new node for this provider
	ProviderListNode pNode = new ProviderListNode(providerId);
	//Add the pNode to the list of providers
	p.add(pNode);
    }

    public  Expertise parse( IXMLElement xmlExpertise ){

	int roles = -1;
	WeightMatrix w = new WeightMatrix();
	Hashtable v = new Hashtable(); //Vector of providerLists
	ProviderList p = new ProviderList();
	LatticeExpertise newLattice = new LatticeExpertise();

	//System.out.println("Number of roles " + roles);
	try {

	    Properties attributes = xmlExpertise.getAttributes();
	    //dimensions
	    roles = Integer.parseInt( attributes.getProperty(Constants.XML_ATTRIBUTE_DIMENSIONS));

	    // Parse all the child elements

	    Vector vecChildElements = xmlExpertise.getChildren();
	    Enumeration enumChildElements = vecChildElements.elements();

	    while( enumChildElements.hasMoreElements() ){

		IXMLElement xmlChildElement = (IXMLElement)enumChildElements.nextElement();
		
		//parse the weight matrix
		if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_WEIGHT_MATRIX) )
		    w  = WeightMatrix.parse( xmlChildElement);

		//parse the provider list
		if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_PROVIDER_LIST)){
		    //System.out.println("ProviderList");
		    p  = ProviderList.parse( xmlChildElement );
		    v.put(p.getRoleId(), p);
		}
	    }	
	    
	    newLattice.dimension = roles;
	    newLattice.weights = w;
	    newLattice.providers = v;
	}

	catch ( Exception e ){
	    e.printStackTrace();
	}
		
	return newLattice;
	
    }	

    public void toTruncatedXML( IXMLElement xmlExpertise){
	try {
	    // Set the number of dimensions

	    xmlExpertise.setAttribute( Constants.XML_ATTRIBUTE_DIMENSIONS, Integer.toString(this.dimension));

	// Set the Weights

	    IXMLElement xmlMatrix = xmlExpertise.createElement( Constants.XML_TAG_WEIGHT_MATRIX );

	    this.weights.toXML( xmlMatrix );			
	    xmlExpertise.addChild( xmlMatrix );

	    // Set the Providers

	    Enumeration e = this.providers.keys();
	    IXMLElement xmlProviders = null;
	    while (e.hasMoreElements()){
		xmlProviders = xmlExpertise.createElement(Constants.XML_TAG_PROVIDER_LIST);
		((ProviderList)providers.get(e.nextElement())).toXML(xmlProviders);
		xmlExpertise.addChild(xmlProviders);
	    }
	}

	    catch (Exception e) {
		e.printStackTrace();
	    }				

    }	

	
    
    
    /**
     * Gets the value of dimension
     *
     * @return the value of dimension
     */
    public int getDimension() {
	return this.dimension;
    }

    /**
     * Sets the value of dimension
     *
     * @param argDimension Value to assign to this.dimension
     */
    public void setDimension(int argDimension){
	this.dimension = argDimension;
    }

    /**
     * Gets the value of weights
     *
     * @return the value of weights
     */
    public WeightMatrix getWeights() {
	return this.weights;
    }

    /**
     * Sets the value of weights
     *
     * @param argWeights Value to assign to this.weights
     */
    public void setWeights(WeightMatrix argWeights){
	this.weights = argWeights;
    }

    /**
     * Get the value of providers.
     * @return value of providers.
     */
    public Hashtable getProviders() {
	return providers;
    }
    
    /**
     * Set the value of providers.
     * @param v  Value to assign to providers.
     */
    public void setProviders(Hashtable  v) {
	this.providers = v;
    }
    
 
	
}// LatticeExpertise
