//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.vectors;

import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.util.ErrorHandler;
import edu.ncsu.simulators.util.Statistics;
import edu.ncsu.simulators.io.StateWriter;

import net.n3.nanoxml.*;

import java.lang.*;
import java.io.*;
import java.util.*;

import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;


/**
 * Agents use the SimulationVector to represent knowledge of the world that
 * they live in.  Each dimension of the SimulationVector represents a different
 * domain of knowledge.  Every fact about their world can be expressed as a 
 * vector of values between 0 and 1, with the number of dimensions equal to the
 * number of domains of knowledge.  The set of all possible vectors comprises
 * the agents' "Universe of Discourse".
 * <p>
 * SimulationVectors can be used in a variety of different ways.  They can
 * be used to represent knowledge that an agent has (expertise), knowledge that 
 * an agent wants (interests), knowledge that an agent is requesting (query), 
 * or knowledge that an agent is offering (answer).  In all of these cases,
 * SimulationVectors look the same.  The only differences between them is in
 * how they are <i>used</i>.
 * <p>
 * Despite these similarities, we found that it was very usefull to develop a 
 * separate subclass for each possible use of a SimulationVector.  Doing so 
 * made the code much easier to read, and greatly reduced the risk that a 
 * developer might accidentally mix up objects (e.g. by manipulating a query 
 * that he mistakenly believes is an answer).
 */
public class SimulationVector extends Vector
{
    private static final char DELIMITER = ' ';
	
    //========================================================================			
	
    public SimulationVector()
    { 
    }
	

    //========================================================================			
    /**
     * Returns a deep copy this SimulationVector
     */
    public SimulationVector copy( SimulationVector copy )
    {	
	for ( int i=0; i<this.size(); i++ )
	    {
		Double copiedElement = new Double( ((Double)this.elementAt(i)).doubleValue() );
		copy.add( i, copiedElement );
	    }
		
	return copy;
    }
	
    //========================================================================			
    /**
     * Creates a new randomly generated Expertise vector with the specified
     * number of dimensions
     */
    public void setDimensionValueWithinRange( int dimension, double low, double high )
    {
	if ( high < low )
	    {
		ErrorHandler.exit( "SimulationVector.setDimentionValueWithinRange: Specified a range where the high value in the range was lower than the low value" );
	    }
		
	double range = high - low;
	Random randomNumberGenerator = Statistics.getRandomNumberGenerator();
	double random = low + (randomNumberGenerator.nextDouble() * range);
		
	setElementAt( new Double(random), dimension );
    }
	
    //========================================================================			
    /**
     *
     */
    public void setDimensionValue( int dimension, double value )
    {	
	setElementAt( new Double(value), dimension );
    }	

    //========================================================================	

    public Double at (int index){
	return (Double) elementAt(index);
    }
	
    public final static SimulationVector parseFromString( String string, SimulationVector simulationVector )
    {				
	StreamTokenizer tokenizer = new StreamTokenizer( new StringReader( string ) );
		
	tokenizer.whitespaceChars( DELIMITER, DELIMITER );
		
	Double number = new Double(-1);
	int type = 0;
		
	try
	    {
		while ( tokenizer.nextToken() != StreamTokenizer.TT_EOF )
		    {	
			number = new Double( tokenizer.nval );
			simulationVector.add( number );			
		    }
	    }
	catch( Exception e )
	    {
		e.printStackTrace();
	    }		
		
	return simulationVector;
    }
	
    //========================================================================	
	
    public static SimulationVector parseFromXML( IXMLElement xmlVector, SimulationVector simulationVector )
    {	
	StringReader reader = new StringReader( xmlVector.getContent() );
	StreamTokenizer tokenizer = new StreamTokenizer( reader );
	tokenizer.whitespaceChars( DELIMITER, DELIMITER );			

	try
	    {			
		while ( tokenizer.nextToken() != StreamTokenizer.TT_EOF )
		    {									
			Double number = new Double(tokenizer.nval);
			simulationVector.add( number );
		    }	
	    }
	catch ( Exception e )
	    {
		e.printStackTrace();
	    }
		
	return simulationVector;
    }	
	
    //========================================================================			
	
    public final String toString()
    {
	// we don't allow this method to be used because we want the 
	// developer to consciously choose between the 
	// toMessageString() method which preserves the precision of the vector elements, AND
	// toTruncatedString() method which truncates the vector elements to make them easier to read.
		
	ErrorHandler.exit( "SimulationVector.toString() is a forbidden method!" );

	return null;
    }
	
    //========================================================================			
	
    public final String toPreciseString()
    {	
	StringBuffer sbuf = new StringBuffer();

	Enumeration enum = elements();

	while ( enum.hasMoreElements() )
	    {	
		Double element = (Double)enum.nextElement();
		sbuf.append( element.toString() );

		if ( enum.hasMoreElements() )
		    {
			sbuf.append( DELIMITER );
		    }
	    }
		
	return sbuf.toString();
    }
		
    //========================================================================			
	
    public final String toTruncatedString()
    {	
	StringBuffer sbuf = new StringBuffer();

	Enumeration enum = elements();

	while ( enum.hasMoreElements() )
	    {	
		Double element = (Double)enum.nextElement();
		sbuf.append( element.toString().substring(0,Math.min(element.toString().length(),Constants.DECIMAL_STRING_LENGTH)) );

		if ( enum.hasMoreElements() )
		    {
			sbuf.append( DELIMITER );
		    }
	    }
		
	return sbuf.toString();
    }	
	
    //========================================================================	
	
    public final void toPreciseXML( IXMLElement xmlVector )
    {
	xmlVector.setContent( this.toPreciseString() );
    }	
    	
    //========================================================================	
	
    public void toTruncatedXML( IXMLElement xmlVector )
    {		
	xmlVector.setContent( this.toTruncatedString() );
    }		
}
