package edu.ncsu.simulators.vectors.lattice;

/**
 * ProviderListNode.java
 *
 *
 * Created: Tue May  7 18:58:14 2002
 *
 * @author <a href="mailto:pyolum@eos.ncsu.edu">Pinar Yolum</a>
 * @version
 */

import java.util.*;
import java.io.*;

import edu.ncsu.simulators.util.Statistics;
import edu.ncsu.simulators.Constants;
import net.n3.nanoxml.*;

public class ProviderListNode {
    

    private static final char DELIMITER = ' ';

    protected int providerId = -1;
    protected Vector ratings = new Vector();
    protected boolean promoted = false;

    public ProviderListNode () {
	
    }
    

    /**
     * Creates a new <code>ProviderListNode</code> instance.
     *
     * @param id an <code>int</code> value
     * @param r a <code>Vector</code> value
     * @param p a <code>boolean</code> value
     */
    public ProviderListNode(int id, Vector r, boolean p){
	providerId = id;
	ratings = r;
	promoted = p;
    }

    /**
     * Creates a new <code>ProviderListNode</code> instance.
     * + promoted is set to false
     *
     * @param id an <code>int</code> value
     * @param r a <code>Vector</code> value
     */
    public ProviderListNode(int id, Vector r){
	this(id, r, false);

    }


    /**
     * Creates a new <code>ProviderListNode</code> instance.
     *
     * @param id an <code>int</code> value
     * @param r a <code>Double</code> value
     * @param p a <code>boolean</code> value
     */
    public ProviderListNode(int id, Double r, boolean p){
	providerId = id;
	ratings = new Vector();
	ratings.add(r);
	promoted = p;
    }

    public ProviderListNode(int id, Double r){
	this(id, r, false);
    }

    /**
     * Creates a new <code>ProviderListNode</code> instance.
     *
     * @param id an <code>int</code> value
     */
    public ProviderListNode(int id){
	this(id, new Vector());

    }

    /**
     * For testing. A default node with one rating of 0.5
     */
    public static ProviderListNode defaultNode(int id){
	Vector t = new Vector();
	t.add(new Double(0.5));
	return new ProviderListNode(id, t);
    }

    /** 
     * @return average of the ratings so far.
     * if there are no ratings, then return 0.
     */
    public double avgRating(){
	double d = 0.0;

	// if there are some ratings so far.
	if (ratings != null) {
	    d = Statistics.average(ratings); 
	} // end of if ()
	
	return d;
    }

    /**
     * Returns the number of available ratings.
     *
     * @return a <code>double</code> value
     */
    public double numberOfRatings(){
        return ratings.size();
    }

    /**
     * Parses a ProviderListNode.
     *
     * @param xmlNode an <code>IXMLElement</code> value
     * @return a <code>ProviderListNode</code> value
     */
    public static ProviderListNode parse( IXMLElement xmlNode ){	

	Vector ratings = null;
	int agentID = -1;
	Double number = null;
	String temp = xmlNode.getContent();
	ProviderListNode p = null;

	Properties attributes = xmlNode.getAttributes();
	agentID = Integer.parseInt( attributes.getProperty(Constants.XML_ATTRIBUTE_ID) );	 


	if ( temp != null) {
	    
	    StringReader reader = new StringReader( temp );
	    StreamTokenizer tokenizer = new StreamTokenizer( reader );
	    tokenizer.whitespaceChars( DELIMITER, DELIMITER );			
	    ratings = new Vector();
	    try{			
	    
		while ( tokenizer.nextToken() != StreamTokenizer.TT_EOF ){
		    number = new Double( tokenizer.nval );
		    ratings.add( number );			
		}

	    }
	
	    catch ( Exception e ){
		e.printStackTrace();
	    }
	} // end of if ()

		
	return new ProviderListNode(agentID, ratings);
    }	
	
    //========================================================================			

    /**
     * Puts the rating vector into a string seperated by spaces.
     *
     * @return a <code>String</code> value
     */
    public final String toSpacedString()
    {	
	StringBuffer sbuf = new StringBuffer();

	if ( ratings != null) {

	    Enumeration enum = ratings.elements();

	    while ( enum.hasMoreElements() ) {
		Double element = (Double)enum.nextElement();
		sbuf.append( element.toString() );

		if ( enum.hasMoreElements() )
		    {
			sbuf.append( DELIMITER );
		    }
	    }

	} // end of if ()		
	return sbuf.toString();
    }

    //========================================================================	
	
    /**
     * Creates an IXMLElement that keeps the node as an XML element.
     *
     * @param xmlNode an <code>IXMLElement</code> value
     */
    public void toXML( IXMLElement xmlNode ) {
	xmlNode.setAttribute( Constants.XML_ATTRIBUTE_ID, Integer.toString(this.providerId) );
	xmlNode.setContent( this.toSpacedString() );
    }	

    /**
     * Gets the value of providerId
     *
     * @return the value of providerId
     */
    public int getProviderId() {
	return this.providerId;
    }

    /**
     * Sets the value of providerId
     *
     * @param argProviderId Value to assign to this.providerId
     */
    public void setProviderId(int argProviderId){
	this.providerId = argProviderId;
    }

    /**
     * Gets the value of ratings
     *
     * @return the value of ratings
     */
    public Vector getRatings() {
	return this.ratings;
    }

    /**
     * Sets the value of ratings
     *
     * @param argRatings Value to assign to this.ratings
     */
    public void setRatings(Vector argRatings){
	this.ratings = argRatings;
    }

    /**
     * Adds a new rating
     *
     * @param rating Value to add to this.ratings
     */
    public void addRating(Double rating){
	this.ratings.add(rating);
    }

        
    /**
     * Get the value of promoted.
     * @return value of promoted.
     */
    public boolean isPromoted() {
	return promoted;
    }
    
    /**
     * Set the value of promoted.
     * @param v  Value to assign to promoted.
     */
    public void setPromoted(boolean  v) {
	this.promoted = v;
    }
    
    
}// ProviderListNode
