//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.agent;

import edu.ncsu.simulators.Simulator;
import edu.ncsu.simulators.io.StateWriter;
import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.vectors.*;

import net.n3.nanoxml.*;

import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;

/**
 * The AgentModel is a representation of one agent's perception of
 * another agent that it is aware of.
 */
public class AgentModel
{
    public int id = -1;
    private Expertise expertise = new DomainExpertise();
    private Interests interests = new DomainInterests();
    private Sociability sociability = new DomainSociability();
    //    private double sociability = -1;
	
    //========================================================================
    /**
     * Agent models should only be created through the createDefaultAgentModel()
     * method and the parse() method.
     */
    public AgentModel()
    {
    }
	
    //========================================================================
    /**
     * Returns a deep copy of this AgentModel object.
     */	
    public AgentModel copy()
    {
	AgentModel copy = new AgentModel();
	copy.id = this.id;
	copy.expertise = this.expertise.copy();
	copy.interests = this.interests.copy();		
	copy.sociability = this.sociability;
		
	return copy;
    }	
	
    //========================================================================

    public static AgentModel createDefaultAgentModel( int _id )
    {
	AgentModel defaultModel = Simulator.getDefaultAgentModel();
		
	// an easy way of making a copy is to convert it XML and then parse it
	// this may not be the most efficient
		
	IXMLElement xmlDefaultModel = new XMLElement();
	defaultModel.toXML( xmlDefaultModel );
	AgentModel newModel = AgentModel.parse( xmlDefaultModel );
		
	newModel.setID( _id );
		
	return newModel;
    }	
	
    //========================================================================
	
    public Expertise getExpertise()
    {
	return expertise;
    }
	
    //========================================================================
	
    public int getID()
    {
	return id;
    }
	
    //========================================================================	

    public Interests getInterests()
    {
	return interests;
    }
	
    //========================================================================		

    public Sociability  getSociability()
    {
	return sociability;
    }		
	
    //=======================================================================
    /**
     * Parses the agent represented by the specified xml element.
	 */
    public static AgentModel parse( IXMLElement xmlAgentModel )
    {
	AgentModel model = new AgentModel();
		
	Agent agent = null;
	int agentID = -1;				
	Expertise expertise = new DomainExpertise();
	//Interests interests = null;
	Interests interests = (Interests) new DomainInterests();		//Changed from previous 
	//to the current line
	//by Raghuram on Apr 4
	//	double sociability = -1;
	Sociability sociability = null;//new DomainSociability();
	double temp;
	Double number = null;		
		
	try
	    {	
		// Parse all the attributes
		Properties attributes = xmlAgentModel.getAttributes();
			
		agentID = Integer.parseInt( attributes.getProperty(Constants.XML_ATTRIBUTE_ID) );
		temp = Double.parseDouble( attributes.getProperty(Constants.XML_ATTRIBUTE_SOCIABILITY, Constants.XML_ATTRIBUTE_DEFAULT) );
		if (temp!=-1) { //scalar sociability
		    sociability = new ScalarSociability(temp);
		    
		} // end of if ()
		else {
		    sociability = new DomainSociability();
		} // end of else
		

	
		// Parse all the child elements
		Vector vecChildElements = xmlAgentModel.getChildren();
		Enumeration enumChildElements = vecChildElements.elements();
			
		while( enumChildElements.hasMoreElements() )
		    {
			IXMLElement xmlChildElement = (IXMLElement)enumChildElements.nextElement();
			
				// parse Expertise
			if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_EXPERTISE) )
			    {				
				expertise = expertise.parse( xmlChildElement );

			    }
					
				// parse Interests
			else if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_INTERESTS) )
			    {
				// interests = Interests.parse( xmlChildElement );
				interests = interests.parse( xmlChildElement ); // changed from the previous line
				// to the current line by Raghuram.
			    }

			else if ( xmlChildElement.getName().equalsIgnoreCase(Constants.XML_TAG_SOCIABILITY) )
			    {
				// interests = Interests.parse( xmlChildElement );
				sociability = sociability.parse( xmlChildElement ); 

			    }
		    }
			
		model.setID( agentID );
		model.setSociability( sociability );
		model.setExpertise( expertise );
		model.setInterests( interests );			
	    }
	catch (Exception e) {
		e.printStackTrace();
	}
		
	return model;
    }
	
    //=======================================================================
    /**
     * 
	 */
    public void toXML( IXMLElement xmlAgentModel )
    {		
	try
	    {	
		// Set all the attributes
		xmlAgentModel.setAttribute( Constants.XML_ATTRIBUTE_ID, Integer.toString(this.id) );

		if ( this.sociability instanceof ScalarSociability) {
		    
		    String stringSociability = new Double(((ScalarSociability)this.sociability).getSociability() ).toString();
		    String truncatedSociability = stringSociability.substring(0,Math.min(stringSociability.length(),Constants.DECIMAL_STRING_LENGTH));
			
		    xmlAgentModel.setAttribute( Constants.XML_ATTRIBUTE_SOCIABILITY, truncatedSociability );
		} // end of if ()
		else {
		    IXMLElement xmlSociability = xmlAgentModel.createElement( Constants.XML_TAG_SOCIABILITY );
		
		    this.sociability.toTruncatedXML( xmlSociability );	
		    xmlAgentModel.addChild( xmlSociability );		    

		} // end of else
		
						
		// Set all the child elements
		IXMLElement xmlExpertise = xmlAgentModel.createElement( Constants.XML_TAG_EXPERTISE );
		
		this.expertise.toTruncatedXML( xmlExpertise );	
		xmlAgentModel.addChild( xmlExpertise );

		IXMLElement xmlInterests = xmlAgentModel.createElement( Constants.XML_TAG_INTERESTS );
		this.interests.toTruncatedXML( xmlInterests );			
		xmlAgentModel.addChild( xmlInterests );
	    }
	catch( Exception e )
	    {
		e.printStackTrace();
	    }				
    }	
	
    //=======================================================================
	
    public void setExpertise( Expertise expertise )
    {
	this.expertise = expertise;
    }
	
    //=======================================================================
	
    public void setID( int id )
    {
	this.id = id;
    }
	
    //=======================================================================
	
    public void setInterests( Interests interests )
    {
	this.interests = interests;
    }
	
    //=======================================================================
	
    public void setSociability( Sociability sociability )
    {
	this.sociability = sociability;
    }	
	
    //========================================================================				
}
