//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================
package edu.ncsu.simulators.dialogs;

import edu.ncsu.simulators.vectors.Answer;
import edu.ncsu.simulators.vectors.Query;

import java.util.Enumeration;
import java.util.Hashtable;

/**
 * The ReferralChain interface is the definition an interface which allows
 * agents to track a chain of referrals which leads an agent from the 
 * intitial asking of a question to one or more good answers, bad answers, 
 * or dead-ends (i.e. non-answers).
 */
public interface ReferralChain
{	
	//========================================================================			
	
	/**
	 * Adds the specified referring agent's referral to the specified referred
	 * agent.
	 * Some implementations may choose not to add certain referrals, such as
	 * those that lead to cycles.
	 */
	public void addReferral( int referringAgentID, int referredAgentID );
	
	/**
	 * Returns an Enumeration of Enumerations containing the (Integer)ids of the
	 * specified agent's ancestors.
	 * An ancestor is an agent that either refers directly to the referred agent,
	 * or to another ancestor of the referred agent.
	 * The 1st item in the enumeration contains an enumeration of the agent's parents.
	 * The 2nd item contains an enumeration of the agent's grandparents.
	 * the 3rd item contains an enumeration of the agent's great-grandparents.
	 * etc.
	 */
	public Enumeration getAncestorsByGeneration( int agentID );
	
	/**
	 * Returns a Hashtable containing all the answers (including non-answers)
	 * in the referral chain that are to be evaluated.
	 * keys - (Integer) id of the answering agent
	 * element - (Answer) answer given by the answering agent
	 */
	public Hashtable getAnswers();
	
	/**
	 * Returns the length of the referral chain from the questioning agent
	 * to the specified agent
	 */
	public int getChainLength( int agentID );
	
	/**
	 * Returns an Enumeration containing the (Integer)ids of the specified 
	 * agent's descendants, plus the agentID parameter.
	 * A descendant is an agent that the specified agent either directly refers
	 * to, or is referred to by another descendant of the specified agent.
	 * MJS-task
	 * Currently, this method *MUST* also return the agentID parameter in 
	 * the Enumeration (this should probably be changed).
	 */
	public Enumeration getDescendants( int agentID );	
	
	/**
	 * Returns the query that is associated with this referral chain
	 */
	public Query getQuery();	
	
	/**
	 * Records the answer given by the specified agent to the query associated
	 * with this referral chain.
	 */
	public void recordAnswer( int answeringAgentID, Answer answer );	
	
	//========================================================================	
}