//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.io;

import edu.ncsu.simulators.agent.Agent;
import edu.ncsu.simulators.agent.AgentModel;
import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.Simulator;

import net.n3.nanoxml.*;

import java.io.*;

/**
 * The StateWriter writes a collection of existing agents to a state file.
 * As long as the OPTION_WRITE_READABLE_STATE_FILE parameter was set to TRUE
 * in the parameter file, the StateWriter saves all the information that is 
 * necessary to reconstruct the agents.  If the option was set to FALSE, then
 * the StateWriter will only save the most fundamental data about the state of 
 * the agent - which would save a lot file space buy would render the file
 * unreadable by the StateReader. 
 */
public final class StateWriter extends FileWriter
{	
	private int tabs = 0;
	
	//========================================================================			
	
	public StateWriter(String fileName) throws java.io.IOException
	{
		super( fileName );		
	}
	
	//========================================================================				
	
	public void save( Agent[] agents, int numDimensions, AgentModel defaultAgentModel )
	{				
		try
		{	
			IXMLElement xmlAgents = new XMLElement( Constants.XML_TAG_AGENTS );

			// ------- save the simulation attributes --------------
			xmlAgents.setAttribute( Constants.XML_ATTRIBUTE_POPULATION_SIZE, Integer.toString(agents.length) );			
			xmlAgents.setAttribute( Constants.XML_ATTRIBUTE_DIMENSIONS, Integer.toString(numDimensions) );
			
			// Only do the following if the parameters indicate that 
			// the state file must contain enough information to be 
			// read in as an agent input file.
			if ( Constants.OPTION_WRITE_READABLE_STATE_FILES )
			{
				// ----- save the default agent model used by each agent ---
				IXMLElement xmlAgentModel = xmlAgents.createElement( Constants.XML_TAG_AGENT_MODEL );
				defaultAgentModel.toXML( xmlAgentModel ); 
				xmlAgents.addChild( xmlAgentModel );			
			}

			// ------- save every agent --------------							
			for ( int i=0; i<Simulator.getAgentCount(); i++ )
			{
				Agent agent = agents[i];

				IXMLElement xmlAgent = xmlAgents.createElement( agent.getXMLTag() );
				agent.toXML( xmlAgent ); 
				xmlAgents.addChild( xmlAgent );
			}
			
			XMLWriter writer = new XMLWriter( this );
			writer.write( xmlAgents, true, 5 );							
		}
		catch( Exception e )
		{
			e.printStackTrace();
		}			
	}	
}
