//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.policies;

import edu.ncsu.simulators.Constants;
import edu.ncsu.simulators.policies.Heuristics;
import edu.ncsu.simulators.util.Statistics;
import edu.ncsu.simulators.agent.AgentModel;
import edu.ncsu.simulators.agent.AgentModels;
import edu.ncsu.simulators.vectors.*;
import edu.ncsu.simulators.agent.Neighbors;

import java.util.Iterator;
import java.util.Random;
import java.util.Vector;

/**
 * This "standard" policy provides a very straight forward implementation of an
 * AnsweringPolicy.
 *
 * @author Mike Suralik
 */
public class AnsweringPolicy_Standard implements AnsweringPolicy
{
    // the random number generator used by this agent
    private Random random = null;	
	
	//=======================================================================
	
    public AnsweringPolicy_Standard()
    {
	random = Statistics.getRandomNumberGenerator();	
    }


	
    public Answer generateAnswer( AgentModel modelQuestioner, Query query, Expertise expertise, Interests interest, int myId, Neighbors neighbors) {
	
	return generateAnswerIfExpert(modelQuestioner, query,  expertise);
    }
	

    //=======================================================================
    /**
     * In this policy implementation we ignore the model of the questioning
     * agent, and treat all agents equally.  The answering agent will generate
     * the best answer that it can, as long as its expertise is similiar to the
     * query.  Otherwise, it will not generate any answer at all.
	 */
    public Answer generateAnswerIfExpert( AgentModel modelQuestioner, Query query, Expertise expertise )
    {
	Answer answer = null;
		
	double expertiseComponent = 0;
	double answerComponent = 0;
	double percentage = 0;
	double sign = 0;
		
	// if the answering agent's expertise is sufficiently similar to the query being made,
	// then answering agent will make its best effort to answer the query
	if ( Heuristics.areSufficientlySimilar(query, expertise, Constants.SIMILARITY_THRESHOLD_FOR_ANSWERING_QUESTIONER) )
	    {
		answer = new Answer();
			
		Iterator iteratorExpertiseComponents = expertise.iterator();

		while( iteratorExpertiseComponents.hasNext() )
		    {
			percentage = random.nextDouble() * Constants.MAXIMUM_PERCENTAGE_PERTURBATION_FOR_GENERATING_ANSWER;			
			sign = random.nextDouble();

			if ( sign < 0.5 )
			    {
				sign = -1;
			    }
			else
			    {
				sign = 1;
			    }

			expertiseComponent = ((Double)iteratorExpertiseComponents.next()).doubleValue();			
			answerComponent = expertiseComponent + (sign * percentage * expertiseComponent);

				// the answer component can not be less than 0, or greater than 1
			answerComponent = Math.max( 0, answerComponent );
			answerComponent = Math.min( 1, answerComponent );

			answer.add( new Double(answerComponent) );
		    }
	    }
	// otherwise, the answering agent will not answer the question
	else
	    {
		answer = null;
	    }

	return answer;	
    }
	
    //=======================================================================
}
