//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.policies;

import edu.ncsu.simulators.Constants;

import edu.ncsu.simulators.agent.AgentModel;
import edu.ncsu.simulators.agent.AgentModels;
import edu.ncsu.simulators.agent.Neighbors;
import edu.ncsu.simulators.policies.Heuristics;
import edu.ncsu.simulators.util.RandomIterator;
import edu.ncsu.simulators.vectors.Query;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

/**
 * This "standard" policy provides a very straight forward implementation of an
 * AskingPolicy.
 *
 * @author Mike Suralik
 */
public class AskingPolicy_Standard implements AskingPolicy
{
	//=======================================================================
	
	public AskingPolicy_Standard()
	{
	}
	
	//=======================================================================
	/**
	 * @param query the query that the questioning agent wants to send
	 * @param neighbors all of the neighbors whom the questioning agent can consider sending the query to
	 * @param agentModels all of the agent models that the questioning agent maintains
	 *
	 * @return a collection of neighbors whom the questioning agent should send
	 * the query to (this will be a subset of the agents specified in the neighbors parameter)
	 */ 
	public Iterator getNeighborsToAsk( Query query, Neighbors neighbors, AgentModels agentModels )
	{
		return getNeighborsWithSufficientSimilarity( query, neighbors, agentModels, Constants.SIMILARITY_THRESHOLD_FOR_ASKING_NEIGHBOR );
	}
	
	//=======================================================================
	/**
	 * Returns an enumeration of the ID's of those neighbors who are 
	 * thought to be knowledgable about this query
	 */
	private Iterator getNeighborsWithSufficientSimilarity( Query query, Neighbors neighbors, AgentModels agentModels, double similarityThreshold )
	{	
		Iterator iteratorKnowledgeableNeighbors = null;
		
		Iterator iteratorNeighbors = neighbors.iterator();
		Vector vecKnowledgeableNeighbors = new Vector();

		while( iteratorNeighbors.hasNext() )
		{
			Integer neighborID = (Integer)iteratorNeighbors.next();		
			AgentModel neighborModel = agentModels.getModel( neighborID );			
			
			if ( Heuristics.areSufficientlySimilar( query, neighborModel.getExpertise(), similarityThreshold ) )
			{
				vecKnowledgeableNeighbors.addElement( neighborID );
			}
		}
		
		// Create an Iterator that will iterate over the knowledgeable neighbors in some random order.
		iteratorKnowledgeableNeighbors = RandomIterator.getRandomIterator( vecKnowledgeableNeighbors );
		
		return iteratorKnowledgeableNeighbors;
	}	
}