//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.policies;

import edu.ncsu.simulators.vectors.*;
import java.util.Enumeration;

/**
 * Provides heuristic functions that are commonly used by implementations of
 * various policies.  All of this classes methods should be static.
 */
public final class Heuristics
{
    //========================================================================			
	
    private Heuristics()
    {	
    }	
	
    //=======================================================================	
    /**
     * Determines whether the similarity of the specified query and expertise
     * is greater than the specified threshold.
	 */

    public static boolean areSufficientlySimilar( SimulationVector basisVector, SimulationVector comparedVector, double threshold) {
	
	double similarity = calculateSimilarity( basisVector, comparedVector );
		
	return (similarity > threshold);
    }	
	
    //=======================================================================	
    /**
     * Determines whether the similarity of the specified query and expertise
     * is greater than the specified threshold.
     */
    public static double calculateSimilarity( SimulationVector basisVector, SimulationVector comparedVector )
    {	
	Enumeration enumBasisComponents = basisVector.elements();
	Enumeration enumComparedComponents = comparedVector.elements();
		
	double similarityNumerator = 0;
	double similarityDenominator = 0;
	double similarity = 0;

	while( enumBasisComponents.hasMoreElements() )
	    {
		double basisComponent = ((Double)enumBasisComponents.nextElement()).doubleValue();			
		double comparedComponent = ((Double)enumComparedComponents.nextElement()).doubleValue();

		similarityNumerator = similarityNumerator + comparedComponent * basisComponent;
		similarityDenominator = similarityDenominator + basisComponent * basisComponent;
	    }

	similarityDenominator = Math.sqrt( basisVector.size() * similarityDenominator );

	similarity = similarityNumerator / similarityDenominator;
		
	return similarity;
    }	


    /**
     * if the sociability is scalar, just return it. Otherwise,
     * calculate the similarity between the interest vector and the
     * sociability vector, as capability.
     *
     * @param basisVector a <code>SimulationVector</code> value
     * @param compared a <code>Sociability</code> value
     * @return a <code>double</code> value 
     */

    public static double calculateSociabilitySimilarity(SimulationVector basisVector, Sociability compared){
	
	double result = 0.0;
	if (compared instanceof ScalarSociability ) {
	    result = ((ScalarSociability)compared).getSociability();
	    
	} 
	else {
	    result = calculateSimilarity(basisVector, (DomainSociability)compared);
	} // end of else
	
	return result;
	
    }	
		  
 }


    
