//==========================================================================
//	This source code is part of the Agent Simulation Test Bed system which
//	was developed for simulating interaction among agents who are attempting 
//	to acquire information from each other and learn about each other. 
//
//	The system was created at North Carolina State Univeristy, 
//	Department of Computer Science between September and December 2001. 
//	It was implemented by Michael J. Suralik based on the specifications 
//	developed by Pinar Yolum and Bin Yu, who were conducting research under 
//   the guidance of Dr. Munindar Singh. 
//==========================================================================

package edu.ncsu.simulators.util;

import java.util.Random;
import java.util.Iterator;
import java.util.Vector;

/**
 * The Statistics class provides methods related to statistics.
 * All methods in this utility class should be public.
 */
public final class Statistics
{
    // this random number generator is only to be used interally as an aid
    // for generating the seeds that will be used to seed the creation of
    // other random number generators.
    private static Random seedGenerator = null;
	
	// the seed for the seed generator
    private static Long primalSeed = null;
	
    //========================================================================			
	
    private Statistics()
    {	
    }	

    //=======================================================================		
    /**
     * Instantiates and returns a new random number generator using the
     * seed value that has been set for this class.  If no seed has been
     * set, it will make a new random seed for every generator.
	 */
    public static Random getRandomNumberGenerator()
    {
	Random randomNumberGenerator = null;
		
	if ( seedGenerator == null )
	    {	
		if ( primalSeed == null )
		    {
			initializeSeed();
		    }
			
		//System.out.println( "primal seed = " + primalSeed );			
		seedGenerator = new Random( primalSeed.longValue() );
	    }

	long seed = seedGenerator.nextLong();
	//System.out.println( "seed = " + seed );
	randomNumberGenerator = new Random( seed );
		
	return randomNumberGenerator;
    }

    //=======================================================================
    /**
     * Returns the primal seed. If a primal seed has not already been set,
     * it will create one.
     */
    public static Long getSeed()
    {
	if ( primalSeed == null )
	    {
		initializeSeed();
	    }
		
	return primalSeed;
    }
	
    //=======================================================================	
    /**
     * Initializes the primal seed with a random numnber.  The primal seed
     * can never be changed once it is set.
	 */
    public synchronized static void initializeSeed()
    {
	// if a primal seed has not been set yet, 
	// then generate a random number between Long.MIN_VALUE and Long.MAX_VALUE
	if ( primalSeed == null )
	    {		
		// Get 3 of the least significant digits of the current time,
		// as measured in milliseconds elapsed since some arbitrary date.
		long longCurrentTime = System.currentTimeMillis();
		//System.out.println( "time = " + longCurrentTime );
		String strCurrentTime = String.valueOf( longCurrentTime );
		String strLeastSignificantDigits = strCurrentTime.substring( strCurrentTime.length()-3, strCurrentTime.length() );
		double leastSignificantDigits = Double.parseDouble( strLeastSignificantDigits );

		// Divide that 3 digit number by 1,000 to produce
		// a random number between 0 and 1.
		double percentage = leastSignificantDigits / 1000; 

			// Now multiply the number between 0 and 1 by the maximum possible seed value.
		primalSeed = new Long ((long)(percentage * (new Long(Long.MAX_VALUE)).doubleValue()));
	    }		
	else
	    {
		ErrorHandler.exit( "Attempting to set the primal seed for the Statistics class AFTER it has already been set" );
	    }		
    }
	
    //=======================================================================	
    /**
     * Initializes the primal seed with a specified number.  The primal seed
     * can never be changed once it is set.
	 */
    public synchronized static void initializeSeed( Long seed )
    {
	if ( primalSeed == null )
	    {		
		primalSeed = seed;
	    }		
	else
	    {
		ErrorHandler.exit( "Attempting to set the primal seed for the Statistics class AFTER it has already been set" );
	    }
    }
	
    //========================================================================

    public static double average (Vector v){

	double d = 0.0;
	Iterator i= v.iterator();

	while (i.hasNext())
	    d += ((Double)i.next()).doubleValue();
	return (d/v.size());
	    
    }

    
}



